#!/usr/bin/perl
#
# Copyright 2011, Julian Catchen <jcatchen@uoregon.edu>
#
# This file is part of Stacks.
#
# Stacks is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Stacks is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Stacks.  If not, see <http://www.gnu.org/licenses/>.
#

use strict;
use Bio::SeqIO;

use constant true    => 1;
use constant false   => 0;

my $debug       = false;
my $in_path     = "";
my $out_path    = "";
my $fasta_path  = "";
my $pct_aln_lim = 95;
my $pct_id_lim  = 95;

parse_command_line();

my (%hits, %best_aln, %genes);

load_blat_alns($in_path, \%hits);

calc_best_alns(\%hits, \%best_aln, \%genes);

write_gene_coord($out_path, \%genes);

translate_fasta($fasta_path, \%best_aln);

sub calc_best_alns {
    my ($hits, $best, $genes) = @_;

    my ($best_hit, $max_pct_aln, $hit, $num_hits, $i, $max_len);

    my $tot_genes = 0;
    my $tot_trans = 0;

    foreach my $gene (keys %{$hits}) {
	$tot_genes++;

	foreach my $tran (keys %{$hits->{$gene}}) {
	    $tot_trans++;

	    $best_hit  = undef;
	    $max_pct_aln = 0;

	    foreach $hit (@{$hits->{$gene}->{$tran}}) {
		$hit->{'pct_id'}  = 100 * ($hit->{'match'} - $hit->{'mmatch'}) / $hit->{'match'};
		$hit->{'pct_aln'} = 100 * $hit->{'match'} / $hit->{'len'};

		if ($hit->{'pct_aln'} > $max_pct_aln) {
		    $best->{$gene}->{$tran} = $hit;
		    $max_pct_aln = $hit->{'pct_aln'};
		}
	    }
	}
    }

    print STDERR "Total genes $tot_genes, total transcripts $tot_trans\n";

    foreach my $gene (keys %{$best}) {
	foreach my $tran (keys %{$best->{$gene}}) {
	    $hit = $best->{$gene}->{$tran};
	    # print STDERR 
	    # 	"  ", $gene, "; ", $tran, 
	    # 	sprintf(" pct id: %.1f; pct aln: %.1f", $hit->{'pct_id'}, $hit->{'pct_aln'}), "; ",
	    # 	"Length: ", $hit->{'len'}, "\n";

	    if ($hit->{'pct_id'} > $pct_id_lim && $hit->{'pct_aln'} > $pct_aln_lim) {
		$hit->{'status'} = "keep";
	    } else {
		$hit->{'status'} = "discard";
	    }
	}
    }

    foreach my $gene (keys %{$best}) {
	$max_len  = 0;
	$best_hit = undef;
	foreach my $tran (keys %{$best->{$gene}}) {
	    $hit = $best->{$gene}->{$tran};

	    next if ($hit->{'status'} eq "discard");

	    if ($hit->{'len'} > $max_len) {
		$best_hit = $hit;
	    }
	}

	if (defined($best_hit)) {
	    $genes->{$gene} = $best_hit;
	}
    }
}

sub write_gene_coord {
    my ($out_path, $genes) = @_;

    my ($fh, $gene, $path, $gene_id);

    $path = $out_path . "xma_summary.tsv";
    open($fh, ">" . $path) or die($!);

    foreach $gene (keys %{$genes}) {
	($gene_id) = ($gene =~ /^Locus_(\d+)$/);
	$gene_id = sprintf("G%07d", $gene_id);
	print $fh
	    $gene_id, "\t",
	    "\t", # External ID
	    $genes->{$gene}->{'chr'},   "\t",
	    $genes->{$gene}->{'start'}, "\t",
	    $genes->{$gene}->{'end'},   "\t",
	    $genes->{$gene}->{'strand'} eq "+" ? "1" : "-1",   "\t",
	    "protein_coding\n";
    }

    close($fh);
}

sub translate_fasta {
    my ($fasta_path, $best) = @_;

    my ($gene, $tran, $id, $gene_id, $tran_id, $path, $seq);

    $path = $out_path . "xma_cdna.fa";
    my $contigs = Bio::SeqIO->new(-file => "$fasta_path", -format => "fasta");
    my $out_ctg = Bio::SeqIO->new(-file => ">$path", -format => "fasta");

    my $transcript_id = 1;
    my $skipped = 0;
    my $tot     = 0;
    my $wrote   = 0;

    my %gene_cnt;
    my %skipped_gene_cnt;
    my %wrote_gene_cnt;

    while ($seq = $contigs->next_seq()) {

	($gene, $tran) = ($seq->id() =~ /^(Locus_\d+)_(Transcript_\d+\/\d+)_Confidence.+$/);

	$tot++;
	$gene_cnt{$gene}++;

	if (!defined($best->{$gene}->{$tran}) ||
	    $best->{$gene}->{$tran}->{'status'} eq "discard") {
	    #print STDERR "Skipping $gene / $tran\n";
	    $skipped++;
	    $skipped_gene_cnt{$gene}++;
	    next;
	}

	($gene_id) = ($gene =~ /^Locus_(\d+)$/);
	$gene_id = sprintf("G%07d", $gene_id);
	$tran_id = sprintf("T%07d", $transcript_id);

	$id = $tran_id . "|" . $gene_id;
	
	$seq->desc($seq->id());
	$seq->id($id);

	$out_ctg->write_seq($seq);
	$wrote++;
	$wrote_gene_cnt{$gene}++;
	$transcript_id++;
    }

    print STDERR 
	"Wrote $wrote of $tot transcripts, skipping $skipped transcripts due to insufficient alignments.\n",
	"Viewed ", 
	scalar(keys %gene_cnt), " genes, wrote ", scalar(keys %wrote_gene_cnt), ", skipped ", scalar(keys %skipped_gene_cnt), ".\n";
}

sub load_blat_alns {
    my ($in_path, $genes) = @_;

    my ($fh, @parts, $line, $href);

    open($fh, "<" . $in_path) or die("Unable to open BLAT file '$in_path' $!\n");

    #
    # Read off the five line header.
    #
    $line = <$fh>;
    $line = <$fh>;
    $line = <$fh>;
    $line = <$fh>;
    $line = <$fh>;

    while ($line = <$fh>) {
        chomp $line;
        @parts = split(/\t/, $line);

	$href = {};
	$href->{'match'}  = $parts[0];
	$href->{'mmatch'} = $parts[1];
	$href->{'strand'} = $parts[8];
	$href->{'id'}     = $parts[9];
	$href->{'len'}    = $parts[10];
	$href->{'chr'}    = $parts[13];
	$href->{'start'}  = $parts[15];
	$href->{'end'}    = $parts[16];

	#
	# Pattern match to extract the gene ID / transcript ID
	#
	# Locus_21_Transcript_2/8_Confidence_0.542_Length_4664|ENSGACG00000013959|ENSORLG00000019467|ENSTRUG00000000711
	#
	($href->{'gene'}, $href->{'tran'}) = ($parts[9] =~ /^(Locus_\d+)_(Transcript_\d+\/\d+)_Confidence.+$/);

	push(@{$genes->{$href->{'gene'}}->{$href->{'tran'}}}, $href);
    }

    close($fh);
}

sub parse_command_line {
    while (@ARGV) {
	$_ = shift @ARGV;
	if    ($_ =~ /^-p$/) { $in_path    = shift @ARGV; }
	elsif ($_ =~ /^-m$/) { $pct_id_lim = shift @ARGV; }
	elsif ($_ =~ /^-o$/) { $out_path   = shift @ARGV; }
	elsif ($_ =~ /^-f$/) { $fasta_path = shift @ARGV; }
	elsif ($_ =~ /^-d$/) { $debug++; }
	elsif ($_ =~ /^-h$/) { usage(); }
	else {
	    print STDERR "Unknown command line option: '$_'\n";
	    usage();
	}
    }
    if (length($in_path) == 0) {
	print STDERR "You must specify the BLAT alignments file.\n";
	usage();
    }
    if (length($fasta_path) == 0) {
	print STDERR "You must specify the FASTA file containing gene models.\n";
	usage();
    }
    if (length($out_path) == 0 || !-e $out_path) {
	print STDERR "You must specify a directory to output files.\n";
	usage();
    }
    $out_path .= "/" if (substr($out_path, -1) ne "/");
}

sub usage {
    print STDERR <<EOQ; 
annotate.pl -p path -o path -f fasta [-d] [-h]
    p: path to BLAT alignment file.
    o: path to output files.
    f: path to FASTA file containing gene models.
    h: display this help message.
    d: turn on debug output.

EOQ

exit(0);
}
