#!/usr/bin/perl
#
# Copyright 2011-2015, Julian Catchen <jcatchen@illinois.edu>
#

use strict;
use Statistics::Descriptive;

use constant true    => 1;
use constant false   => 0;
use constant unknown => -1;
use constant forw    => 0;
use constant reve    => 1;

my $debug     = false;
my $write_php = false;
my $php_path  = "";
my $in_path   = "";
my $sam_hits  = "";
my $agp_path  = "";
my $doc_path  = "";
my $wl_path   = "";
my $bl_path   = "";
my $min_mark  = 1;

my $org     = "ssc";
my $version = "0.2";
my $scaffold_prefix = "SSC";

my (%lgs, %maps, %ctg_maps, %hits, %ctg_lens, %ctg_conflicts, %agp, %markers_in_map, %conflict_markers);
my (%fh, $map_fh, $con_fh);

parse_command_line();

load_agp_file(\%agp);

load_sam_aln(\%ctg_lens, \%hits);

my $scaffold_id = 1;

load_lgs(\%lgs, \%hits, \%agp, \%conflict_markers);

open($map_fh, ">" . $in_path . "/scaffold_map.tsv") or die($!);

resolve_interlg_conflicts(\%lgs, \%hits, \%ctg_conflicts, \%agp, \%markers_in_map, \%conflict_markers, \$scaffold_id, $map_fh);

update_lg_marker_positions(\%lgs, \%hits, \%conflict_markers, \%agp);

update_contig_lengths(\%agp, \%ctg_lens);

my ($lg, $log_fh);

foreach $lg (keys %lgs) {
    $maps{$lg}    = [];
    $ctg_maps{$lg} = {};

    ###
    # next if ($lg ne "1");

    print STDERR "Processing LG$lg...\n";

    open($fh{$lg}, ">" . $in_path . "/lg" . $lg . "_scaffold.log") or die($!);

    build_map($lgs{$lg}, $maps{$lg}, \%ctg_lens);

    build_contig_map($maps{$lg}, $ctg_maps{$lg});

    print {$fh{$lg}} "LG$lg unmodified:\n";
    traverse($maps{$lg}, $fh{$lg}, true);

    write_php($php_path, $doc_path, $lg, $maps{$lg}, 'before') if ($write_php);

    prune_markers($lg, $ctg_maps{$lg}, \%conflict_markers);
}

%lgs            = ();
%maps           = ();
%ctg_maps       = ();
%markers_in_map = ();

load_lgs(\%lgs, \%hits, \%agp, \%markers_in_map, \%conflict_markers);

print STDERR "Rebuilding map structure...\n";

foreach $lg (keys %lgs) {
    $maps{$lg}    = [];
    $ctg_maps{$lg} = {};

    ###
    # next if ($lg ne "1");

    print STDERR "Processing LG$lg...\n";

    build_map($lgs{$lg}, $maps{$lg}, \%ctg_lens);

    build_contig_map($maps{$lg}, $ctg_maps{$lg});

    resolve_conflicts($lg, $ctg_maps{$lg}, \%hits, \%agp, \$scaffold_id, $map_fh, \%conflict_markers);
}

# exit(1);

#
# After resolving conflicting scaffolds, rebuild map structure.
#
%lgs            = ();
%maps           = ();
%ctg_maps       = ();
%markers_in_map = ();

update_contig_lengths(\%agp, \%ctg_lens);

load_lgs(\%lgs, \%hits, \%agp, \%markers_in_map, \%conflict_markers);

print STDERR "Rebuilding map structure...\n";

print STDERR scalar(keys %conflict_markers), " conflict markers will be excluded.\n";

open($con_fh, ">" . $in_path . "/conflict_markers.tsv") or die($!);
foreach my $key (keys %conflict_markers) {
    print $con_fh 
	$key, "\t", 
	"LG", $conflict_markers{$key}->{'lg'}, "\t", 
	$conflict_markers{$key}->{'cm'}, "\t", 
	$conflict_markers{$key}->{'contig'}, "\n";
}
close($con_fh);

$min_mark = 1;

foreach $lg (keys %lgs) {
    $maps{$lg}    = [];
    $ctg_maps{$lg} = {};

    ###
    # next if ($lg ne "1");

    build_map($lgs{$lg}, $maps{$lg}, \%ctg_lens);

    build_contig_map($maps{$lg}, $ctg_maps{$lg});

    detect_conflicts($maps{$lg}, $lg, $ctg_maps{$lg}, $fh{$lg});
}


foreach $lg (keys %lgs) {
    ###
    # next if ($lg ne "1");

    print {$fh{$lg}} "\n\nLG$lg final:\n";
    traverse($maps{$lg}, $fh{$lg}, false);

    write_php($php_path, $doc_path, $lg, $maps{$lg}, 'after') if ($write_php);

    write_map($maps{$lg}, $lg);

    close($fh{$lg});
}

close($map_fh);

summarize(\%conflict_markers, \%maps, \%ctg_lens);

#
# Write the AGP file back out.
#
write_agp(\%agp, $agp_path, $in_path);

sub write_agp {
    my ($agp, $agp_path, $in_path) = @_;

    my ($agp_fh, $key, $i, $cnt);

    my ($agp_file) = ($agp_path =~ /^.+\/(.+)$/);
    $agp_file  = $in_path . "/" . $agp_file;
    $agp_file .= ".new" if ($agp_file eq $agp_path);

    open($agp_fh, ">" . $agp_file) or die($!);

    print $agp_fh 
	"# Unplaced Scaffolds-from-components\n",
	"# Format: object object_beg object_end part_number component_type component_id component_beg component_end  orientation\n",
	"#   Gaps: object object_beg object_end part_number     N          gap_length   gap_type      linkage        -empty-\n";

    foreach $key (keys %{$agp}) {
	$cnt = scalar(@{$agp->{$key}});

	for ($i = 0; $i < $cnt; $i++) {
	    print $agp_fh
		$agp->{$key}->[$i]->{'scaf'},  "\t",
		$agp->{$key}->[$i]->{'start'}, "\t",
		$agp->{$key}->[$i]->{'end'},   "\t",
		$agp->{$key}->[$i]->{'part'},  "\t",
		$agp->{$key}->[$i]->{'type'},  "\t",
		$agp->{$key}->[$i]->{'5'},     "\t",
		$agp->{$key}->[$i]->{'6'},     "\t",
		$agp->{$key}->[$i]->{'7'},     "\t",
		$agp->{$key}->[$i]->{'8'},     "\n";
	}
    }

    close($agp_fh);
}

sub summarize {
    my ($conflict_markers, $maps, $ctg_lens) = @_;

    my ($map_fh, $line, @parts, %map, $scaf, $s);

    open($map_fh, "<" . $in_path . "/scaffold_map.tsv") or die($!);
    while ($line = <$map_fh>) {
        chomp $line;
        @parts = split(/\t/, $line);
	if (!defined($map{$parts[0]})) {
	    $map{$parts[0]} = {};
	}
	$map{$parts[0]}->{$parts[3]}++;
    }    
    close($map_fh);

    my $tot = 0;
    my %dist;

    foreach $scaf (keys %map) {
	my @frags;

	next if ($scaf =~ /JHP/);

	$tot++;
	# print STDERR $scaf, "\n";
	foreach $s (keys %{$map{$scaf}}) {
	    if (defined($map{$s})) {
		push(@frags, keys %{$map{$s}});
	    } else {
		push(@frags, $s);
	    }
	}
	# print STDERR "  ", join("; ", @frags), "\n"
	$dist{scalar(@frags)}++;
    }

    print STDERR $tot, " scaffolds were split:\n";
    foreach $s (sort keys %dist) {
	print STDERR "  $s fragments: ", $dist{$s}, " cases\n";
    }

    my ($node, $ctg, $lg, $map, $cnt);
    my $tot_final = 0;
    my $tot_scaf  = 0;
    my $tot_len   = 0;
    my $tot_orig  = 0;
    my $tot_mrks  = 0;
    my $tot_ori_known       = 0;
    my $tot_ori_unknown     = 0;
    my $tot_ori_known_len   = 0;
    my $tot_ori_unknown_len = 0;
    my %ctg_marker_dist;

    my $known_lens   = Statistics::Descriptive::Full->new();
    my $unknown_lens = Statistics::Descriptive::Full->new();

    foreach $map (keys %{$maps}) {

	my $lg_ori_known       = 0;
	my $lg_ori_unknown     = 0;
	my $lg_ori_known_len   = 0;
	my $lg_ori_unknown_len = 0;

	foreach $node (@{$maps{$map}}) {
	    foreach $ctg (@{$node->{'ctgs'}}) {
		$tot_len += $ctg_lens->{$ctg->{'id'}};
		$tot_final++;
		if (!defined($ctg->{'old'})) {
		    $tot_scaf++;
		}

		if ($ctg->{'ori'} == forw || $ctg->{'ori'} == reve) {
		    $lg_ori_known++;
		    $lg_ori_known_len += $ctg_lens->{$ctg->{'id'}};
		    $known_lens->add_data($ctg_lens->{$ctg->{'id'}});

		} else {
		    $lg_ori_unknown++;
		    $lg_ori_unknown_len += $ctg_lens->{$ctg->{'id'}};
		    $unknown_lens->add_data($ctg_lens->{$ctg->{'id'}});

		}
		$ctg_marker_dist{scalar(@{$ctg->{'mrk'}})}++;
		$tot_mrks += scalar(@{$ctg->{'mrk'}});
	    }
	}

	$tot_ori_known       += $lg_ori_known;
	$tot_ori_unknown     += $lg_ori_unknown;
	$tot_ori_known_len   += $lg_ori_known_len;
	$tot_ori_unknown_len += $lg_ori_unknown_len;

	print STDERR 
	    "LG ", $map, ": ",
	    "contig orientation known: $lg_ori_known (", commify($lg_ori_known_len), "bp); unknown: $lg_ori_unknown (", commify($lg_ori_unknown_len), "bp)\n";
    }

    $tot_orig = $tot_scaf + $tot;
    print STDERR 
	"$tot_final total scaffolds mapped from $tot_orig original scaffolds. ", commify($tot_len), "bp total length.\n",
	"$tot_mrks total markers mapped.\n",
	"Contig orientations known: $tot_ori_known (", commify($tot_ori_known_len), "bp); unknown: $tot_ori_unknown (", commify($tot_ori_unknown_len), "bp)\n";

    my $ctg_multiple_marker = 0;
    print STDERR "Contig marker distribution:\n", "Markers per contig --> number of contigs\n";
    foreach $cnt (sort {$a <=> $b} keys %ctg_marker_dist) {
	print STDERR "$cnt\t$ctg_marker_dist{$cnt}\n";
	if ($cnt > 1) { $ctg_multiple_marker += $ctg_marker_dist{$cnt}; }
    }
    print STDERR "Total contigs with a single marker: $ctg_marker_dist{1}; contigs with multiple markers: $ctg_multiple_marker\n";

    print STDERR 
	sprintf("Contigs with known orientation, mean: %0.2f; standard deviation: %0.2f\nContigs with unknown orientation, mean: %0.2f; standard deviation: %0.2f\n", 
		$known_lens->mean(), $known_lens->standard_deviation(), $unknown_lens->mean(), $unknown_lens->standard_deviation());
}

sub mean {
    my ($aref) = @_;

    my $i;
    my $tot = 0.0;

    foreach $i (@{$aref}) {
	$tot += $i;
    }

    return $tot / scalar(@{$aref});
}

sub resolve_interlg_conflicts {
    my ($lgs, $hits, $conflicts, $agp, $markers_in_map, $blacklist, $scaffold_cnt, $map_fh) = @_;

    my ($lg, $loc, $lref, $contig, $mrk, $doc_fh, $path, $res);
    my %markers;

    #
    # Sort through all the markers to record which markers are mapped
    # to which contigs.
    #
    foreach $mrk (keys %{$hits}) {
	#
	# Exclude markers in the SAM alignment file that are not in the genetic map.
	#
	next if (!defined($markers_in_map->{$mrk}));

	if (!defined($markers{$hits->{$mrk}->{'chr'}})) {
	    $markers{$hits->{$mrk}->{'chr'}} = [];
	}
	push(@{$markers{$hits->{$mrk}->{'chr'}}}, $mrk);
    }

    #
    # For each LG, check which contigs occur on that LG. Then, store all the contigs
    # along with which LGs the contigs appear on. We can then check to see which contigs
    # are mapped to multiple LGs.
    #
    foreach $lg (keys %{$lgs}) {
	foreach $loc (keys %{$lgs->{$lg}}) {

	    foreach $contig (keys %{$lgs->{$lg}->{$loc}}) {

		if (!defined($conflicts->{$contig})) {
		    $conflicts->{$contig} = {};
		}
		$conflicts->{$contig}->{$lg}++;
	    }
	}
    }

    #
    # Check if any contigs are mapped to multiple LGs
    #
    my @errors;
    my $cnt = 0;
    foreach $contig (keys %{$conflicts}) {
	my @to_sort;
	my %ctg_span;

	#next if ($contig ne "JH556677.1");
	#next if ($contig ne "JH557217.1"); # Overlapping LG segments
	#next if ($contig ne "JH557165.1");

	if (scalar(keys %{$conflicts->{$contig}}) > 1) {
	    $cnt++;
	    print STDERR "$contig is mapped to multiple linkage groups: ", join(" ", keys %{$conflicts->{$contig}}), "\n";

	    #
	    # Open a log file to document changes made to this scaffold.
	    #
	    $path = $doc_path . $contig . ".log";
	    open($doc_fh, ">$path") or die("Unable to open scaffold documentation '$path', $!\n");
	    print $doc_fh "Scaffold $contig is mapped to multiple linkage groups: ",  join(" ", keys %{$conflicts->{$contig}}), "\n";

	    #
	    # Obtain all the markers that map to this contig.
	    #
	    foreach $mrk (@{$markers{$contig}}) {
		push(@to_sort, $hits->{$mrk});
	    }

	    #
	    # If a contig maps to more than one LG, add the base pair positions
	    # for the different LGs, so we can detect how much of the contig is spanned
	    # by the set of markers from each LG.
	    #
	    @to_sort = sort {$a->{'bp'} <=> $b->{'bp'}} @to_sort;
	    my $n   = scalar(@to_sort);
	    my $cur = $to_sort[0]->{'lg'};
	    my $segment = 1;
	    while ($n > 0) {
		$mrk = shift @to_sort;
		if ($mrk->{'lg'} ne $cur) {
		    $segment++;
		    $cur = $mrk->{'lg'};
		}
		$mrk->{'seg'}    = $segment;
		$mrk->{'contig'} = $contig;
		push(@to_sort, $mrk);
		$n--;
	    }

	    my $i = 1;
	    foreach $mrk (@to_sort) {
		push(@{$ctg_span{$mrk->{'seg'}}}, $mrk->{'bp'});
		print $doc_fh
		    sprintf("%03d", $i), "\t", sprintf("% 6s", $mrk->{'marker'}), "\t", 
		    sprintf("% 4s", "LG" . $mrk->{'lg'}), "\t",
		    sprintf("% 6s", $mrk->{'cm'}), "cM <=> ", sprintf("% 7s", $mrk->{'bp'}), "bp => segment ", $mrk->{'seg'}, " [$mrk->{'cgr'}]\n";
		# print STDERR
		#     $mrk->{'seg'}, "\t",
		#     sprintf("%03d", $i), "\t", sprintf("% 6s", $mrk->{'marker'}), "\t", 
		#     sprintf("% 4s", "LG" . $mrk->{'lg'}), "\t", 
		#     sprintf("%03.3f", $mrk->{'cm'}), "cM <=> ", sprintf("% 7s", $mrk->{'bp'}), "bp => segment ", $mrk->{'seg'}, "\n";
		$i++;
	    }

	    my @ctg_keys = keys %ctg_span;
	    foreach $lg (@ctg_keys) {
		@{$ctg_span{$lg}} = sort {$a <=> $b} @{$ctg_span{$lg}};

		print $doc_fh 
		    "     ", 
		    sprintf("% 4s", "Segment " . $lg), ": ", 
		    sprintf("% 9s", commify($ctg_span{$lg}->[0])), "bp - ", 
		    sprintf("% 9s", commify($ctg_span{$lg}->[-1])), "bp\n";
	    }

	    $res = split_scaffold($contig, \%ctg_span, $hits, $agp, $markers{$contig}, $scaffold_cnt, $map_fh, $doc_fh);

	    #
	    # Delete the existing scaffold representation, since it has been split out onto new contigs.
	    #
	    if ($res) {
		delete($agp->{$contig});
	    } else {
		# 
		# If the split failed, remove all markers except those pointing to the majority node in the linkage map.
		#
		my $majority = 0;
		$segment = "";
		my ($aref);
		foreach $lg (@ctg_keys) {
		    $segment  = scalar(@{$ctg_span{$lg}}) > $majority ? $lg : $segment;
		    #print $doc_fh "Segment: ", $lg, " marker count: ", scalar(@{$ctg_span{$lg}}), "; majority: ", $majority, "; segment: ", $segment, "\n";
		    $majority = scalar(@{$ctg_span{$lg}}) > $majority ? scalar(@{$ctg_span{$lg}}) : $majority;
		}
		#
		# Remove annotations.
		#
		$aref = $agp->{$contig};
		for (my $i = 0; $i < scalar(@{$aref}); $i++) {
		    if (defined($aref->[$i]->{'m'})) {
			delete $aref->[$i]->{'m'};
		    }
		}
		#
		# Blacklist markers
		#
		foreach $mrk (@to_sort) {
		    if ($mrk->{'seg'} ne $segment) {
			print $doc_fh "  Removing marker ", sprintf("% 5s", $mrk->{'marker'}), "; marker segment: ", $mrk->{'seg'}, "; majority segment: ", $segment, "\n";
			$blacklist->{$mrk->{'marker'}} = $mrk;
		    }
		}
	    }
	    print $doc_fh "\n";
	}
    }
    print STDERR "$cnt total contigs mapped to multiple linkage groups.\n";
}

sub split_scaffold {
    my ($scaffold, $ctg_span, $hits, $agp, $markers, $scaffold_id, $map_fh, $doc_fh) = @_;

    my ($start, $end, $cnt, $s, $size, $lg, $mrk);

    if (!defined($agp->{$scaffold}) || scalar(@{$agp->{$scaffold}}) == 0) {
	print STDERR "Unable to process $scaffold, can't find AGP reference.\n";
	return;
    }

    $cnt  = scalar(@{$agp->{$scaffold}}) - 1;
    $s    = $agp->{$scaffold};
    $size = scalar(@{$s}) - 1;

    print $doc_fh "Splitting scaffold $scaffold, ", $cnt + 1, " components\n";

    #
    # For each LG on this scaffold, attach the begining markers to the nearest gap sequence.
    #
    foreach $lg (keys %{$ctg_span}) {
	# print STDERR "Looking at segment $lg from $scaffold\n";

	# foreach my $j (@{$ctg_span->{$lg}}) {
	#     print STDERR "   ", $j, "\n";
	# }

	my $cur = 0;
	#
	# Place the marker that defines the starting position of this LG.
	#
	while ($cur <= $size && $s->[$cur]->{'end'} < $ctg_span->{$lg}->[0]) {
	    $cur++;
	}
	#
	# Move it to the nearest gap.
	#
	while ($cur > 0 && $s->[$cur]->{'type'} ne "N") {
	    $cur--;
	}
	push(@{$s->[$cur]->{'m'}}, {'lg' => $lg, 'lg_type' => "start"});
    }
    #
    # For each LG on this scaffold, attach the ending markers to the nearest gap sequence.
    #
    foreach $lg (keys %{$ctg_span}) {
	my $cur = $size;
	#
	# Place the marker that defines the ending position of this LG.
	#
	while ($cur > 0 && $ctg_span->{$lg}->[-1] < $s->[$cur]->{'end'}) {
	    $cur--;
	}
	$cur++ if ($cur < $size);
	#
	# Move it to the nearest gap.
	#
	while ($cur < $size && $s->[$cur]->{'type'} ne "N") {
	    $cur++;
	}
	push(@{$s->[$cur]->{'m'}}, {'lg' => $lg, 'lg_type' => "end"});
    }

    print_scaffold_map($s, $doc_fh, true);

    #
    # If the end marker of one linkage group occurs before the starting marker of the next linkage group,
    # then interpolate where to break the scaffold by finding the largest gap and moving the end/start 
    # markers to both point at this gap.
    #
    my $next  = 0;
    my $cur   = 0;
    my $found = false;
    while ($next < $cnt && $cur < $cnt) {

	while ($cur < $cnt) {
	    while (!defined($s->[$cur]->{'m'}) && $cur < $cnt) { $cur++; }
	    last if ($cur >= $cnt);
	    last if(scalar(@{$s->[$cur]->{'m'}}) == 1 &&
		    $s->[$cur]->{'m'}->[0]->{'lg_type'} eq "end");
	    $cur++;
	}

    	$next = $cur + 1;

	while ($next < $cnt) {
	    while (!defined($s->[$next]->{'m'}) && $next < $cnt) { $next++; }
	    last if ($next >= $cnt);
	    last if (scalar(@{$s->[$next]->{'m'}}) == 1 &&
		     $s->[$next]->{'m'}->[0]->{'lg_type'} eq "start");
	    $next++;
	}

	last if ($cur >= $cnt || $next >= $cnt);

	#
	# Interpolate between these breakpoints -- find the largest gap and set the breakpoint there.
	#
	print $doc_fh 
	    "Interpolating between markers: ",
	    "Segment ", $s->[$cur]->{'m'}->[0]->{'lg'}, " ", $s->[$cur]->{'m'}->[0]->{'lg_type'}, " ($cur) -> ",
	    "Segment ", $s->[$next]->{'m'}->[0]->{'lg'}, " ", $s->[$next]->{'m'}->[0]->{'lg_type'}, " ($next)\n";

	my $max_gap = $s->[$cur]->{'len'};
	my $max_pos = $cur;
	for (my $i = $cur; $i <= $next; $i++) {
	    if ($s->[$i]->{'type'} eq "N" && $s->[$i]->{'len'} > $max_gap) {
		$max_gap = $s->[$i]->{'len'};
		$max_pos = $i;
	    }
	}
	print $doc_fh "  Maximum gap at position $max_pos (", $max_gap, "bp)\n";

	#
	# Move the markers to the maximal gap position.
	#
 	if ($cur != $max_pos) {
	    push(@{$s->[$max_pos]->{'m'}}, shift(@{$s->[$cur]->{'m'}}));
	    delete $s->[$cur]->{'m'};
	}
	if ($next != $max_pos) {
	    push(@{$s->[$max_pos]->{'m'}}, shift(@{$s->[$next]->{'m'}}));
	    delete $s->[$next]->{'m'};
	}

	$cur = $next + 1;
    }
    print $doc_fh "\n";

    print_scaffold_map($s, $doc_fh, true);

    #
    # Check the integrity of the marker placements, make sure they do not overlap.
    #
    # After interpolation, all markers should be in order: strart->end/start->end/start->end, 
    # so $segment_cnt should equal 0 at the end of traversing the scaffold if everything lines up.
    #
    my @marker_order;
    for ($cur = 0; $cur <= $size; $cur++) {
	if (defined($s->[$cur]->{'m'})) {
	    if (scalar(@{$s->[$cur]->{'m'}}) > 2) {
		print STDERR  "  Unable to split $scaffold, too many markers present at node $cur\n";
		print $doc_fh "Unable to split $scaffold, too many markers present at node $cur\n";
		return 0;
	    } elsif (scalar(@{$s->[$cur]->{'m'}}) == 2) {
		if (($s->[$cur]->{'m'}->[0]->{'lg_type'} eq "start" &&
		     $s->[$cur]->{'m'}->[1]->{'lg_type'} eq "start") ||
		    ($s->[$cur]->{'m'}->[0]->{'lg_type'} eq "end" &&
		     $s->[$cur]->{'m'}->[1]->{'lg_type'} eq "end")) {
		    print STDERR  "  Unable to split $scaffold, too many 'start' or 'end' markers present at node $cur\n";
		    print $doc_fh "Unable to split $scaffold, too many 'start' or 'end' markers present at node $cur\n";
		    return 0;
		}

		if ($s->[$cur]->{'m'}->[0]->{'lg_type'} eq "end") {
		    push(@marker_order, $s->[$cur]->{'m'}->[0]->{'lg_type'});
		} elsif ($s->[$cur]->{'m'}->[1]->{'lg_type'} eq "end") {
		    push(@marker_order, $s->[$cur]->{'m'}->[1]->{'lg_type'});
		}
		if ($s->[$cur]->{'m'}->[0]->{'lg_type'} eq "start") {
		    push(@marker_order, $s->[$cur]->{'m'}->[0]->{'lg_type'});
		} elsif ($s->[$cur]->{'m'}->[1]->{'lg_type'} eq "start") {
		    push(@marker_order, $s->[$cur]->{'m'}->[1]->{'lg_type'});
		}
	    } else {
		push(@marker_order, $s->[$cur]->{'m'}->[0]->{'lg_type'});
	    }
	}
    }

    #
    # Serialize the start/end annotations to make sure they occur in the proper order (out of order segments
    # is caused when there is not enough gaps between markers and the scaffold cannot be split).
    #
    my @anns = ("start", "end");
    my $i = 0;
    foreach my $ann (@marker_order) {
	if ($ann ne $anns[$i % 2]) {
	    print STDERR  "  Unable to split $scaffold, markers overlap, abandoning split attempt.\n";
	    print $doc_fh "Unable to split $scaffold, markers overlap, abandoning split attempt.\n";
	    return 0;
	}
	$i++;
    }

    #
    # Traverse the scaffold and split it into new contigs.
    #
    my @stack;
    my @new_scaffolds;
    my $start = 0;
    my $end   = 0;
    my $last  = 0;
    my ($aref, $new_scaf, $adj_end, $adj_bp);

    #
    # Find the last marker on the scaffold, record it so we can adjust the end point of the last scaffold later on.
    #
    while ($end <= $size) {
	if (defined($s->[$end]->{'m'})) {
	    $last = $end;
	}
	$end++;
    }
    $end = 0;

    #
    # Find the first starting marker, add the corresponding LG to the stack.
    #
    while ($end < $size && !defined($s->[$end]->{'m'})) {
	$end++;
    }
    unshift(@stack, $s->[$end]->{'m'}->[0]->{'lg'});
    # print STDERR "Pushing LG", $s->[$end]->{'m'}->[0]->{'lg'}, " onto the stack, ", scalar(@stack), " elements remain.\n";
    $end++;

    #
    # Loop until the stack is empty and we have reached the end of the scaffold.
    #
    while (scalar(@stack) > 0) {
	#
	# Find the next boundary.
	#
	while ($end < $size && !defined($s->[$end]->{'m'})) {
	    $end++;
	}
	$end = $size if ($end == $last);

	# print STDERR "Setting start to $start; end to $end\n";

	#
	# Create a new scaffold.
	#
	$new_scaf = sprintf("%s%05d.0", $scaffold_prefix, $$scaffold_id);
	push(@new_scaffolds, $new_scaf);
	$$scaffold_id++;

	#
	# Adjust the endpoint to not include the ending gap. Calculate the basepair adjustment.
	#
	$adj_end = $end == $size ? $end : $end - 1;
	$adj_bp  = $s->[$start]->{'start'};

	#
	# Adjust markers from the old segment of the scaffold.
	#
	foreach $mrk (@{$markers}) {
	    if ($hits->{$mrk}->{'bp'} >= $s->[$start]->{'start'} &&
		$hits->{$mrk}->{'bp'} <= $s->[$adj_end]->{'end'}) {

		#
		# We go through two rounds of scaffold splitting, first for inter-LG splits, then 
		# intra-LG splits, so, we want to keep the original scaffold/bp recorded for the 
		# scaffold map.
		#
		if (!defined($hits->{$mrk}->{'old_chr'})) {
		    $hits->{$mrk}->{'old_chr'} = $hits->{$mrk}->{'chr'};
		}
		if (!defined($hits->{$mrk}->{'old_bp'})) {
		    $hits->{$mrk}->{'old_bp'}  = $hits->{$mrk}->{'bp'};
		}
		$hits->{$mrk}->{'chr'}     = $new_scaf;
		$hits->{$mrk}->{'bp'}      = $hits->{$mrk}->{'bp'} - $adj_bp + 1;
		# print STDERR 
		#     "Starting region: ", $s->[$start]->{'start'}, "; Ending region: ", $s->[$adj_end]->{'end'}, "\n",
		#     "  Adjusting position of marker $mrk: old scaffold: ",
		#     $hits->{$mrk}->{'old_chr'}, ", old bp: ", $hits->{$mrk}->{'old_bp'},
		#     "; new scaffold: ", $hits->{$mrk}->{'chr'}, ", new bp: ", $hits->{$mrk}->{'bp'}, "\n";
	    }
	}

	#
	# Extract out the contigs.
	#
	$aref = [];
	foreach my $i ($start..$adj_end) {
	    push(@{$aref}, $s->[$i]);
	    #
	    # Rename and renumber.
	    #
	    $s->[$i]->{'old_scaf'}  = $s->[$i]->{'scaf'};# if (!defined($s->[$i]->{'old_scaf'}));
	    $s->[$i]->{'scaf'}      = $new_scaf;
	    $s->[$i]->{'old_start'} = $s->[$i]->{'start'};
	    $s->[$i]->{'old_end'}   = $s->[$i]->{'end'};
	    $s->[$i]->{'start'}     = $s->[$i]->{'start'} - $adj_bp + 1;
	    $s->[$i]->{'end'}       = $s->[$i]->{'end'}   - $adj_bp + 1;
	}

	# print STDERR "  Adjusting the end point to $adj_end\n";
	$agp->{$new_scaf} = $aref;
	print $doc_fh "Creating scaffold $new_scaf with ", scalar(@{$aref}), " elements.\n";

	#
	# Remove marker annotations from the newly created scaffold.
	#
	for (my $i = 0; $i < scalar(@{$aref}); $i++) {
	    if (defined($aref->[$i]->{'m'})) {
		delete $aref->[$i]->{'m'};
	    }
	}

	#
	# Adjust the stack.
	#   If the final marker in this segment was an 'end' marker for a linkage group, 
	#   pop the linkage group off the stack. If it was a 'start' marker, push the new linkage group on.
	#
	if ($end == $size) {
	    shift(@stack);
	    # print STDERR "Poping off the stack, ", scalar(@stack), " elements remain.\n";
	} else {  
	    foreach my $cur (@{$s->[$end]->{'m'}}) {
		if ($cur->{'lg_type'} eq "end") {
		    shift(@stack);
		    # print STDERR "Poping LG$cur->{'lg'} off the stack, ", scalar(@stack), " elements remain.\n";
		} elsif ($cur->{'lg_type'} eq "start") {
		    unshift(@stack, $cur->{'lg'});
		    # print STDERR "Pushing LG$cur->{'lg'} onto the stack, ", scalar(@stack), " elements remain.\n";
		}
	    }
	}

	#
	# Reset the start/end pointers.
	#
	$start = $end + 1;
	$end   = $start;
	# print STDERR "Reseting start to $start, end to $end.\n";
    }

    #
    # Record a map describing how the old scaffold maps to the new scaffolds.
    #
    foreach $new_scaf (@new_scaffolds) {
	$s   = $agp->{$new_scaf};
	$cnt = @{$agp->{$new_scaf}};

	# print STDERR "Scaffold $new_scaf\n";
	for (my $i = 0; $i < $cnt; $i++) {
	    print $map_fh 
		$s->[$i]->{'old_scaf'}, "\t",
		$s->[$i]->{'old_start'}, "\t",
		$s->[$i]->{'old_end'}, "\t",
		$s->[$i]->{'scaf'}, "\t",
		$s->[$i]->{'start'}, "\t",
		$s->[$i]->{'end'}, "\t",
		$s->[$i]->{'type'}, "\t",
		$s->[$i]->{'part'}, "\n";
	}
    }

    return 1;
}

sub print_scaffold_map { 
    my ($s, $doc_fh, $compress) = @_;

    my $cnt = scalar(@{$s});
    print $doc_fh "Scaffold $s->[0]->{'scaf'}, $cnt components.\n";

    #
    # Find components that contain marker annotations.
    #
    my (@pattern, @indexes);
    if ($compress == true) {
	for (my $i = 0; $i < $cnt; $i++) {
	    if (defined($s->[$i]->{'m'})) {
		push(@indexes, $i);
	    }
	}

	my ($start, $end, $i);
	while (scalar(@indexes) > 0) {
	    $i = shift @indexes;
	    $start = $i - 3 < 0 ? 0 : $i - 3;
	    $end   = $i + 3 > $cnt - 1 ? $cnt - 1 : $i + 3;

	    for (my $j = $start; $j <= $end; $j++) {
		$pattern[$j] = 1;
	    }
	}
    } else {
	for (my $i = 0; $i < $cnt; $i++) {
	    push(@pattern, 1);
	}
    }

    my $dots = 0;
    for (my $i = 0; $i < $cnt; $i++) {
	print $doc_fh sprintf("%03d", $i), "   " if ($i == 0 && $pattern[0] == 0);

	if ($pattern[$i] == 0) {
	    print $doc_fh ".";
	    $dots++;
	    if ($dots >= 60) {
		print $doc_fh "\n      ";
		$dots = 0;
	    }
	    if ($i < $cnt - 1 && $pattern[$i + 1] == 1) {
		print $doc_fh "\n";
	    }
	} else {
	    $dots = 0;
	    if ($s->[$i]->{'type'} eq "N") {
		print $doc_fh 
		    sprintf("%03d", $i),
		    "      Gap",
		    " start: ",  sprintf("% 7d", $s->[$i]->{'start'}), 
		    ", end: ",   sprintf("% 7d", $s->[$i]->{'end'}), 
		    ", size: ",  sprintf("% 7d", $s->[$i]->{'len'}), "\n";
	    } else {
		print $doc_fh 
		    sprintf("%03d", $i),
		"   ",
		"Contig start: ", sprintf("% 7d", $s->[$i]->{'start'}), 
		", end: ",        sprintf("% 7d", $s->[$i]->{'end'}), 
		", size: ",       sprintf("% 7d", $s->[$i]->{'len'}), "\n";
	    }

	    if (defined($s->[$i]->{'m'})) {
		foreach my $cur (@{$s->[$i]->{'m'}}) {
		    print $doc_fh "         Segment ", $cur->{'lg'}, " ", $cur->{'lg_type'}, "\n";
		}
	    }

	    if ($i < $cnt - 1 && $pattern[$i + 1] == 0) {
		print $doc_fh sprintf("%03d", $i + 1), "   ";
	    }
	}
    }

    print $doc_fh "\n";
}

sub update_lg_marker_positions {
    my ($lgs, $hits, $blacklist, $agp) = @_;

    my @markers;
    my ($lg, $cm, $ctg, $m);

    foreach $lg (keys %{$lgs}) {
	foreach $cm (keys %{$lgs->{$lg}}) {
	    foreach $ctg (keys %{$lgs->{$lg}->{$cm}}) {
		foreach $m (@{$lgs->{$lg}->{$cm}->{$ctg}}) {
		    $m->{'contig'} = $hits->{$m->{'marker'}}->{'chr'};
		    $m->{'bp'}     = $hits->{$m->{'marker'}}->{'bp'};
		    push(@markers, $m);
		}
	    }
	}
    }

    #
    # Delete the LG structure and rebuild it to account for newly created scaffolds.
    #
    %{$lgs} = ();

    foreach $m (@markers) {
	if (!defined($blacklist->{$m->{'marker'}})) {
	    push(@{$lgs->{$m->{'lg'}}->{$m->{'cm'}}->{$m->{'contig'}}}, $m);

	    if (defined($agp->{$m->{'contig'}}->[0]->{'old_scaf'})) {
		$m->{'old'} = $agp->{$m->{'contig'}}->[0]->{'old_scaf'};
	    }
	}
    }
}

sub update_contig_lengths {
    my ($agp, $ctg_lens) = @_;

    my ($scaf, $c);

    foreach $scaf (keys %{$agp}) {
	$c = $agp->{$scaf}->[-1]->{'end'};
	# print STDERR "Updating $scaf from a length of ", $ctg_lens->{$scaf}, " to ", $c, "\n";
	$ctg_lens->{$scaf} = $c;
    }
}

sub sort_lgs {
    $a->{'lg'} <=> $b->{'lg'} or $b->{'cm'} <=> $a->{'cm'};
}

sub write_map {
    my ($map, $lg) = @_;

    my $out_file = $in_path . "/lg" . $lg . "_scaffold.tsv";

    my ($fh, $node, $ctg);

    open($fh, ">$out_file") or die("Unable to open '$out_file' $!\n");

    my $master_bp = 1;

    foreach $node (@{$map}) {
	foreach $ctg (@{$node->{'ctgs'}}) {
	    print $fh 
		$master_bp, "\t", 
		$ctg->{'id'}, "\t", 
		($ctg->{'ori'} == reve) ? "-" : "+", "\t",
		($ctg->{'ori'} == reve) ? 
		"$ctg->{'ebp'}\t$ctg->{'sbp'}" :
		"$ctg->{'sbp'}\t$ctg->{'ebp'}";
	    if (defined($ctg->{'old'})) {
		print $fh "\t[", $ctg->{'old'}, "]\n";
	    } else { 
		print $fh "\n";
	    }
	    $master_bp += abs($ctg->{'ebp'} - $ctg->{'sbp'} + 1);
	}
    }

    close($fh);
}

sub detect_conflicts {
    my ($map, $lg, $contigs, $log_fh) = @_;

    my ($contig, $node, $ctg, $cur, $next, $mrk_cnt);

    #
    # Record conflicts among contigs that appear in multiple, disparate places in this linkage group.
    #
    foreach $contig (keys %{$contigs}) {
    	if (scalar(@{$contigs->{$contig}}) > 1) {
	    #next if ($contig ne "JH556864.1");
    	    print STDERR "Conflict: $contig exists on LG$lg in ", scalar(@{$contigs->{$contig}}), " places.\n";

    	    foreach $node (@{$contigs->{$contig}}) {
    		foreach $ctg (@{$node->{'ctgs'}}) {
    		    if ($ctg->{'id'} eq $contig) {
    			print STDERR 
    			    "  $node->{'id'} [", $ctg->{'mrk'}->[0]->{'marker'}, " and ", $ctg->{'cnt'} - 1, " other markers]\n"; 
    			#$ctg->{'cnt'} > 1 ? print STDERR " markers]\n" : print STDERR " marker]\n";
    		    }
    		}
    	    }
    	}
    }

    my $tot_conflicts = 0;
    #
    # Detect markers within a contig that are out of order
    #
    foreach $node (@{$map}) {
	my @conflicts;

	foreach $ctg (@{$node->{'ctgs'}}) {
	    @{$ctg->{'mrk'}} = $ctg->{'ori'} == reve ? 
		sort sort_markers_rev @{$ctg->{'mrk'}} :
		sort sort_markers @{$ctg->{'mrk'}};

	    next if (scalar(@{$ctg->{'mrk'}}) <= 1);

	    $cur     = shift @{$ctg->{'mrk'}};
	    $mrk_cnt = scalar(@{$ctg->{'mrk'}});
	    do {
		$next = shift @{$ctg->{'mrk'}};
		$mrk_cnt--;

		if ($ctg->{'ori'} == forw && $cur->{'bp'} > $next->{'bp'} ||
		    $ctg->{'ori'} == reve && $cur->{'bp'} < $next->{'bp'}) {
		    push(@conflicts, $cur);
		    $tot_conflicts++;
		}

		push(@{$ctg->{'mrk'}}, $cur);
		$cur = $next;
	    } while ($mrk_cnt > 0);

	    push(@{$ctg->{'mrk'}}, $next);

	    if (scalar(@conflicts) > 0) {
		print $log_fh "$ctg->{'id'} has conflicts:\n";
		foreach $cur (@conflicts) {
		    printf($log_fh "  M%05d\n", $cur->{'marker'});
		}
	    }
	}
    }

    print $log_fh "$tot_conflicts conflicting markers identified on LG$lg.\n";
    print STDERR "$tot_conflicts conflicting markers identified on LG$lg.\n";
}

sub sort_markers {
    $a->{'cm'} <=> $b->{'cm'} or $a->{'bp'} <=> $b->{'bp'};
}

sub sort_markers_rev {
    $a->{'cm'} <=> $b->{'cm'} or $b->{'bp'} <=> $a->{'bp'};
}

sub sort_markers_cm_rev {
    $b->{'cm'} <=> $a->{'cm'} or $a->{'bp'} <=> $b->{'bp'};
}

sub sort_nodes {
    if (defined($a->{'old'}) && defined($b->{'old'})) {
	$a->{'old'} cmp $b->{'old'};

    } elsif (defined($a->{'old'}) && !defined($b->{'old'})) {
	$a->{'old'} cmp $b->{'id'};

    } elsif (!defined($a->{'old'}) && defined($b->{'old'})) {
	$a->{'id'} cmp $b->{'old'};
    } else {
	$a->{'id'} cmp $b->{'id'};
    }
}

sub prune_markers {
    my ($lg, $contigs, $conflict_markers) = @_;

    my ($contig, $path, $doc_fh, $node, $ctg, $m, $i, $or);

    my %orient = (-1 => "Unknown", 0 => "Forward", 1 => "Reverse");

    #
    # For each contig, gather all the components which may be split across nodes
    # in the map, determine orientation and order the cM and bp of the markers. 
    # Prune out markers that are inconsistent in their ordering and record them
    # as conflict markers (which will be excluded when the map is rebuilt).
    #

    foreach $contig (keys %{$contigs}) {
	#
	# Skip contigs that don't span more than one node on the map.
	#
	next if (scalar(@{$contigs->{$contig}}) <= 1);

	#
	# Open a log file to document changes made to this scaffold.
	#
	$path = $doc_path . $contig . ".log";
	open($doc_fh, ">>$path") or die("Unable to open scaffold documentation '$path', $!\n");

	print $doc_fh 
	    "\nPruning out of order markers.\n", 
	    "Linkage group $lg; Scaffold $contig\n";

	my @ordering;

	#
	# Determine orientation of scaffold with respect to map, if possible.
	#
	$or = determine_orientation($contig, $contigs->{$contig});

	print $doc_fh "Orientation of $contig is ", $orient{$or}, "\n";

	print $doc_fh "Unmodified Marker ordering:\n";
	$i = 1;
	foreach $node (@{$contigs->{$contig}}) {
	    foreach $ctg (@{$node->{'ctgs'}}) {
		if ($ctg->{'id'} eq $contig) {
		    foreach $m (@{$ctg->{'mrk'}}) {
			print $doc_fh
			    sprintf("%03d", $i), "\t", sprintf("% 6s", $m->{'marker'}), "\t", "LG", $m->{'lg'}, "\t",
			    sprintf("%03.3f", $m->{'cm'}), "cM <=> ", sprintf("% 7s", $m->{'bp'}), "bp", "\t", $m->{'contig'}, "  [$m->{'cgr'}]\n";
			#
			# Gather the markers from this contig that are scattered across map nodes.
			#
			push(@ordering, {'marker' => $m->{'marker'},
					 'bp'     => $m->{'bp'},
					 'lg'     => $m->{'lg'},
					 'cm'     => $m->{'cm'},
					 'contig' => $m->{'contig'},
					 'index'  => $i});
			$i++;
		    }
		}
	    }
	}

	if ($or == forw) {
	    @ordering = sort sort_markers @ordering;
	} else {
	    @ordering = sort sort_markers_cm_rev @ordering;
	}

	print $doc_fh "\n";

	find_outoforder_markers(\@ordering, $conflict_markers, $or, $doc_fh);

	#
	# We need to blacklist all single marker map nodes since they are likely 
	# out of order and will interfere when we rebuild the linkage map.
	#
	# But if all nodes for this scaffold consist of a single marker, just keep 
	# the first one, so the scaffold remains in the assembly.
	#
	my @mrk_cnts;
	#
	# Count up the number of markers present in each map node for this scaffold.
	#
	foreach $node (@{$contigs->{$contig}}) {
	    foreach $ctg (@{$node->{'ctgs'}}) {
		if ($ctg->{'id'} eq $contig) {
		    push(@mrk_cnts, scalar(@{$ctg->{'mrk'}}));
		}
	    }
	}
	#
	# Sort the number of markers on each contig, largest to smallest.
	#
	@mrk_cnts = sort {$b <=> $a} @mrk_cnts;

	my $first = 0;
	foreach $node (@{$contigs->{$contig}}) {
	    foreach $ctg (@{$node->{'ctgs'}}) {
		if ($ctg->{'id'} eq $contig) {
		    #
		    # If this is the only node that this scaffold appears in, and there is only one
		    # marker keeping this scaffold in the assembly, keep the marker.
		    #
		    if (scalar(@{$ctg->{'mrk'}}) == 1) {
			if ($mrk_cnts[0] == 1 && $first == 0) {
			    $first++;
			    next;
			}
			print $doc_fh "Removing single node marker ", $ctg->{'mrk'}->[0]->{'marker'}, "\n";
			$conflict_markers->{$ctg->{'mrk'}->[0]->{'marker'}} = $ctg->{'mrk'}->[0];
		    }
		}
	    }
	}

	print $doc_fh "Pruned marker ordering:\n";

	$i = 1;
	foreach my $href (@ordering) {
	    $href->{'index'} = $i;
	    print $doc_fh
		sprintf("%03d", $i), "\t", sprintf("% 6s", $href->{'marker'}), "\t", 
		sprintf("%03.3f", $href->{'cm'}), "cM <=> ", sprintf("% 7s", $href->{'bp'}), "bp => segment ", $href->{'seg'}, "\n";
	    $i++;
	}
	print $doc_fh "\n";

	close($doc_fh);
    }
}

sub resolve_conflicts {
    my ($lg, $contigs, $hits, $agp, $scaffold_cnt, $map_fh, $conflict_markers) = @_;

    my ($contig, $ctg, $m, $node, $seg, $doc_fh, $path, $i, $or, $res);

    my %orient = (-1 => "Unknown", 0 => "Forward", 1 => "Reverse");

    #
    # Resolve conflicts by removing out of order markers and breaking scaffolds.
    #
    foreach $contig (keys %{$contigs}) {
	next if (scalar(@{$contigs->{$contig}}) <= 1);
	#next if ($contig ne "JH556864.1"); # Reverse orientation
	#next if ($contig ne "JH556732.1");

	#
	# Open a log file to document changes made to this scaffold.
	#
	$path = $doc_path . $contig . ".log";
	open($doc_fh, ">>$path") or die("Unable to open scaffold documentation '$path', $!\n");

	print $doc_fh "Linkage group $lg; Scaffold $contig\n";

	my @marker_ids;
	my @ordering;

	#
	# Determine orientation of scaffold with respect to map, if possible.
	#
	$or = determine_orientation($contig, $contigs->{$contig});

	print $doc_fh "Orientation of $contig is ", $orient{$or}, "\n";

	my $segment = 1;
	#
	# Group the markers that occur in distinct linkage map nodes. We will then interpolate 
	# the gaps between these markers and split the scaffold.
	#
	# $node represent one node on the linkage map and contains all the markers within that 
	# node arranged within contigs. So, we want to break the contig/scaffold up so that 
	# groups of markers exist only on separate nodes.
	#
	foreach $node (@{$contigs->{$contig}}) {

	    foreach $ctg (@{$node->{'ctgs'}}) {
		if ($ctg->{'id'} eq $contig) {
		    foreach $m (@{$ctg->{'mrk'}}) {
			#
			# Skip conflicting markers; these markers will be officially removed when the map is rebuilt.
			#
			next if (defined($conflict_markers->{$m->{'marker'}}));

			push(@ordering, {'marker' => $m->{'marker'},
					 'bp'     => $m->{'bp'},
					 'lg'     => $m->{'lg'},
					 'cm'     => $m->{'cm'},
					 'contig' => $m->{'contig'},
					 'seg'    => $segment});
			push(@marker_ids, $m->{'marker'});
		    }
		}
	    }

	    $segment++;
	}

	print $doc_fh "\nMarker ordering after pruning single-marker nodes, prior to spliting scaffold:\n";

	if ($or == forw) {
	    @ordering = sort sort_markers @ordering;
	} else {
	    @ordering = sort sort_markers_cm_rev @ordering;
	}

	$i = 1;
	foreach my $href (@ordering) {
	    $href->{'index'} = $i;
	    print $doc_fh
		sprintf("%03d", $i), "\t", sprintf("% 6s", $href->{'marker'}), "\t", 
		sprintf("%03.3f", $href->{'cm'}), "cM <=> ", sprintf("% 7s", $href->{'bp'}), "bp => segment ", $href->{'seg'}, "\n";
	    $i++;
	}
	print $doc_fh "\n";

	# find_outoforder_markers(\@ordering, $conflict_markers, $or, $doc_fh);

	# calc_regression(\@ordering, \@blist);

	my %ctg_span;
	foreach my $href (@ordering) {
	    push(@{$ctg_span{$href->{'seg'}}}, $href->{'bp'});
	}

	my @ctg_keys = keys %ctg_span;
	if (scalar(@ctg_keys) == 1) {
	    print $doc_fh "Scaffold $contig does not need to be split after removing out of order markers.\n";
	} else {
	    print $doc_fh "Scaffold $contig will be split into ", scalar(@ctg_keys), " pieces.\n";
	}
	foreach $seg (@ctg_keys) {
	    @{$ctg_span{$seg}} = sort {$a <=> $b} @{$ctg_span{$seg}};

	    print $doc_fh 
		"     ", 
		sprintf("% 4s", "segment " . $seg), ": ", 
		sprintf("% 9s", commify($ctg_span{$seg}->[0])), "bp - ", 
		sprintf("% 9s", commify($ctg_span{$seg}->[-1])), "bp\n";
	}
	print $doc_fh "\n";

	if (scalar(@ctg_keys) > 1) {
	    $res = split_scaffold($contig, \%ctg_span, $hits, $agp, \@marker_ids, $scaffold_cnt, $map_fh, $doc_fh);

	    #
	    # Delete the existing scaffold representation, since it has been split out onto new contigs.
	    #
	    if ($res) {
		delete($agp->{$contig});
	    } else {
		# 
		# If the split failed, remove all markers except those pointing to the majority node in the linkage map.
		#
		my $majority = 0;
		$segment = "";
		my ($aref);
		foreach $seg (@ctg_keys) {
		    $segment  = scalar(@{$ctg_span{$seg}}) > $majority ? $seg : $segment;
		    #print $doc_fh "Segment: ", $seg, " marker count: ", scalar(@{$ctg_span{$lg}}), "; majority: ", $majority, "; segment: ", $segment, "\n";
		    $majority = scalar(@{$ctg_span{$seg}}) > $majority ? scalar(@{$ctg_span{$seg}}) : $majority;
		}
		#
		# Remove annotations.
		#
		$aref = $agp->{$contig};
		for (my $i = 0; $i < scalar(@{$aref}); $i++) {
		    if (defined($aref->[$i]->{'m'})) {
			delete $aref->[$i]->{'m'};
		    }
		}
		#
		# Blacklist markers
		#
		foreach $m (@ordering) {
		    if ($m->{'seg'} ne $segment) {
			print $doc_fh "  Removing marker ", sprintf("% 5s", $m->{'marker'}), "; marker segment: ", $m->{'seg'}, "; majority segment: ", $segment, "\n";
			$conflict_markers->{$m->{'marker'}} = $m;
		    }
		}
	    }
	}

	close($doc_fh);
    }
}

sub determine_orientation {
    my ($contig, $contigs) = @_;

    my ($node, $ctg, $mrk, $cnt, $orientation);

    #
    # Determine the number of markers at each node in the linkage map. We
    # require at least two markers in a node to use it to determine orientation
    # (many time single markers will be out of order in a separate map node due 
    # to genotyping errors).
    #
    my (%cms, @mrks);

    foreach $node (@{$contigs}) {
    	foreach $ctg (@{$node->{'ctgs'}}) {
    	    if ($ctg->{'id'} eq $contig) {
    		if (scalar(@{$ctg->{'mrk'}}) > 1) {
    		    push(@mrks, @{$ctg->{'mrk'}});
    		}
    	    }
    	}
    }

    #
    # Make sure we have at least two linkage map nodes from which to judge orientation.
    #
    foreach $mrk (@mrks) {
    	$cms{$mrk->{'cm'}}++;
    }

    return unknown if (scalar(keys %cms) <= 1);

    @mrks = sort sort_markers @mrks;

    #
    # Check three positions to be sure we have the orientation correct, majority rules.
    #
    $cnt = scalar(@mrks);
    if ($cnt >= 6) {
	my ($or_1, $or_2, $or_3);
	$or_1 = ($mrks[0]->{'bp'} < $mrks[$cnt - 1]->{'bp'}) ? forw : reve;
	$or_2 = ($mrks[1]->{'bp'} < $mrks[$cnt - 2]->{'bp'}) ? forw : reve;
	$or_3 = ($mrks[2]->{'bp'} < $mrks[$cnt - 3]->{'bp'}) ? forw : reve;

	my %ori = (0 => 0, 1 => 0);
	$ori{$or_1}++;
	$ori{$or_2}++;
	$ori{$or_3}++;

	$orientation = ($ori{0} > $ori{1}) ? forw : reve;

	# print STDERR join(" ", keys %ori), "; Forward: ", $ori{0}, "; Reverse: ", $ori{1}, "\n";
    } else {
	$orientation = ($mrks[0]->{'bp'} < $mrks[-1]->{'bp'}) ? forw : reve;
    }

    return $orientation;
}

sub find_outoforder_markers {
    my ($markers, $blacklist, $orientation, $doc_fh) = @_;

    my ($mrk, $n, $cur, $next, $i, $first, $last);

    if (scalar(@{$markers}) == 1) {
	print $doc_fh "Not searching for out of order markers as there is only one marker.\n";
	return;
    }

    my $complete = false;

    print $doc_fh "Searching for out of order markers:\n";

    #
    # We will scan down the list of markers looking for markers out of order. However, before
    # we can do that we must have properly ordered endpoints.
    #
    while ($complete == false) {

    	@{$markers} = $orientation == forw ? 
	    sort sort_markers @{$markers} : sort sort_markers_cm_rev @{$markers};

    	$n     = scalar(@{$markers});
	$first = $markers->[0]->{'marker'};
	$last  = $markers->[$n - 1]->{'marker'};

	@{$markers} = sort {$a->{'bp'} <=> $b->{'bp'}} @{$markers};

	if ($markers->[0]->{'marker'} ne $first) {
	    print $doc_fh "  001: Removing head marker $first\n";
	    while ($n > 0) {
		$cur = shift @{$markers};
		if ($cur->{'marker'} eq $first) {
		    $blacklist->{$cur->{'marker'}} = $cur;
		} else {
		    push(@{$markers}, $cur);
		}
		$n--;
	    }
	}

    	$n = scalar(@{$markers});
	if ($markers->[-1]->{'marker'} ne $last) {
	    print $doc_fh "  ", sprintf("%03s", $n), ": Removing tail marker $last\n";
	    while ($n > 0) {
		$cur = shift @{$markers};
		if ($cur->{'marker'} eq $last) {
		    $blacklist->{$cur->{'marker'}} = $cur;
		} else {
		    push(@{$markers}, $cur);
		}
		$n--;
	    }
	}

	if ($markers->[0]->{'marker'}  eq $first &&
	    $markers->[-1]->{'marker'} eq $last) {
	    $complete = true;
	}

	if (scalar(@{$markers}) < 2) {
	    $complete = true;
	}
    }

    $complete = false;

    #
    # Scan the list of markers searching for out of order markers.
    #
    while ($complete == false) {

    	@{$markers} = $orientation == forw ? sort sort_markers @{$markers} : sort sort_markers_cm_rev @{$markers};
    	$n          = scalar(@{$markers});
    	$cur        = undef;

    	#print STDERR "$n markers remaining.\n";

    	while ($n > 1) {
    	    $cur  = shift @{$markers} if (!defined($cur));
    	    $next = shift @{$markers};

    	    #print STDERR "  $n: Cur marker: $cur->{'marker'}, $cur->{'bp'}\n";
    	    #print STDERR "  $n: Next marker: $next->{'marker'}, $next->{'bp'}\n";

    	    if ($next->{'bp'} < $cur->{'bp'}) {
    		print $doc_fh 
		    "  ", sprintf("%03d", $next->{'index'}), ": ", 
		    "Removing marker ", sprintf("% 5s", $next->{'marker'}), ", $next->{'bp'}bp ",
		    "[Rule 1; $next->{'bp'}bp < $cur->{'bp'}bp]\n";
    		$blacklist->{$next->{'marker'}} = $next;
    		push(@{$markers}, $cur);
    		$cur = undef;
    		last;
    	    } else {
    		if ($next->{'bp'} > $cur->{'bp'} && $n > 2) {
		    $i = 0;
		    while ($markers->[$i]->{'bp'} < $cur->{'bp'} && $i < $n - 1) { $i++; }
		    if ($markers->[$i]->{'bp'} < $next->{'bp'}) {
			print $doc_fh 
			    "  ", sprintf("%03d", $next->{'index'}), ": ", 
			    "Removing marker ", sprintf("% 5s", $next->{'marker'}), ", $next->{'bp'}bp ", 
			    "[Rule 2; $markers->[0]->{'bp'}bp < $next->{'bp'}bp]\n";
			$blacklist->{$next->{'marker'}} = $next;
			push(@{$markers}, $cur);
			last;
		    } else {
			push(@{$markers}, $cur);
			$cur = $next;
		    }
    		} else {
    		    push(@{$markers}, $cur);
    		    $cur = $next;
    		}
    	    }

    	    $n--;
    	}

    	if ($n == 1) {
    	    $complete = true;
    	}
    }
    push(@{$markers}, $cur);

    print $doc_fh "\n";
}

sub calc_regression {
    my ($markers, $blacklist) = @_;

    my ($mrk, $m, $b, $n, $avg_x, $avg_y, $avg_xy, $avg_x2, $obs, $exp);

    #
    # Plot cenitMorgans on x-axis, basepair on Y
    #
    $n      = scalar(@{$markers});
    $avg_x  = 0;
    $avg_y  = 0;
    $avg_xy = 0;
    $avg_x2 = 0;

    my $i = 1;
    foreach $mrk (@{$markers}) {
	$mrk->{'cm'} += 0;
	$mrk->{'bp'} += 0;
	$avg_x  +=  $i; #$mrk->{'cm'};
	$avg_y  +=  $mrk->{'bp'};
	$avg_x2 +=  $i**2;
	$avg_xy += ($i * $mrk->{'bp'});

	print STDERR "$i\t$mrk->{'bp'} # $mrk->{'cm'}\n";
	$i++;
    }

    $avg_x  = $avg_x  / $n;
    $avg_y  = $avg_y  / $n;
    $avg_x2 = $avg_x2 / $n;
    $avg_xy = $avg_xy / $n;

    print STDERR "Avg x: $avg_x; Avg y: $avg_y; Avg x^2: $avg_x2; Avg XY: $avg_xy\n";

    $m = (($avg_x * $avg_y) - $avg_xy) / ($avg_x**2 - $avg_x2);
    $b = $avg_y - ($m * $avg_x);

    print STDERR "Regression, Slope (m): $m, Y-intercept (b): $b\n";

    my $stat = Statistics::Descriptive::Full->new();

    $i   = 1;
    my $diff = 0;
    foreach $mrk (@{$markers}) {
	$obs = $mrk->{'bp'};
	$exp = ($m * $i) + $b;

	#$diff += abs($obs - $exp);

	$stat->add_data(abs($obs - $exp));
	$i++;
    }

    my $mean  = $stat->mean();
    my $stdev = $stat->standard_deviation();

    #$diff = $diff / $n;
    print STDERR "Average distance from mean: $mean\n";
    $i = 1;
    while ($n > 0) {
	$mrk = shift(@{$markers});

	$obs = $mrk->{'bp'};
	$exp = ($m * $i) + $b;

	$diff = abs($obs - $exp);

	if ($diff > $stdev) {
	    print STDERR "Distance to mean: $i ", abs($obs - $exp), "; final diff: ", abs(abs(($obs - $exp)) - $diff), "\n";
	    push(@{$blacklist}, $mrk);
	    print STDERR "Adding $mrk->{'marker'} to blacklist.\n";
	} else {
	    push(@{$markers}, $mrk);
	}
	$i++;
	$n--;
    }


}

sub build_contig_map {
    my ($map, $contigs) = @_;

    my ($node, $ctg, %mrks, %discards, %readd);

    #
    # For each contig, record the nodes from the linkage map (and the markers 
    # that exist at those nodes) that occur on each contig. If a contig is referenced
    # in multiple nodes in the linkage map, then all those nodes are recorded in $contigs.
    # 
    foreach $node (@{$map}) {
	foreach $ctg (@{$node->{'ctgs'}}) {

	    if (scalar(@{$ctg->{'mrk'}}) < $min_mark) {
		my $markers = "";
		foreach my $m (@{$ctg->{'mrk'}}) {
		    $markers .= $m->{'marker'} . ",";
		}
		print STDERR "Discarding marker ", $markers, " from scaffold ", $ctg->{'id'}, "\n";
		$discards{$ctg->{'id'}}++;
		next;
	    }

	    if (!defined($contigs->{$ctg->{'id'}})) {
		$contigs->{$ctg->{'id'}} = [];
		$mrks{$ctg->{'id'}} = [];
	    }
	    push(@{$contigs->{$ctg->{'id'}}}, $node);
	    push(@{$mrks{$ctg->{'id'}}}, @{$ctg->{'mrk'}});
	}
    }

    #
    # If a contig only had a single marker keeping it in the map, re-add it.
    #
    foreach $ctg (keys %discards) {
	$readd{$ctg}++ if (!defined($contigs->{$ctg}));
    }
    foreach $node (@{$map}) {
	foreach $ctg (@{$node->{'ctgs'}}) {
	    if (defined($readd{$ctg->{'id'}}) && !defined($contigs->{$ctg->{'id'}})) {
		$contigs->{$ctg->{'id'}} = [];
		$mrks{$ctg->{'id'}} = [];
		push(@{$contigs->{$ctg->{'id'}}}, $node);
		push(@{$mrks{$ctg->{'id'}}}, @{$ctg->{'mrk'}});
	    }
	}
    }

    #
    # Determine contig orientation
    #
    my ($f, $l, $c, $ori, $markers);

    my %o = (-1 => "Unknown", 0 => "Forward", 1 => "Reverse");

    foreach $ctg (keys %mrks) {
	$markers = $mrks{$ctg};

	@{$markers} = sort sort_markers @{$markers};
	$f = $markers->[0];
	$l = $markers->[-1];

	if ($f->{'cm'} == $l->{'cm'}) {
	    $ori = unknown;
	} elsif ($f->{'bp'} < $l->{'bp'}) {
	    $ori = forw;
	} elsif ($f->{'bp'} >= $l->{'bp'}) {
	    $ori = reve;
	} else {
	    die("Unable to set orientation for $ctg\n");
	}
	print STDERR 
	    "Setting $ctg to orientation $ori; ",
	    "First: $f->{'cm'} <=> Last: $l->{'cm'}cM; First: $f->{'bp'} <=> Last: $l->{'bp'}bp.\n" if ($debug);
	#
	# Record the orientation in each instance of this contig
	#
	foreach $node (@{$contigs->{$ctg}}) {
	    foreach $c (@{$node->{'ctgs'}}) {
		$c->{'ori'} = $ori if ($c->{'id'} eq $ctg);
	    }
	}
    }
}

sub build_map {
    my ($lg, $merged_map, $ctg_lens) = @_;

    my (@map, @new);

    #
    # Create a list of nodes, representing each mapping location
    #
    my ($node, $loc, $ctg, $c);

    foreach $loc (sort {$a <=> $b} keys %{$lg}) {
	$node = {};
	$node->{'id'}   = $loc;
	$node->{'ctgs'} = [];

	foreach $c (keys %{$lg->{$loc}}) {
	    $ctg = {};
	    $ctg->{'id'}  = $c;
	    $ctg->{'sbp'} = 1;
	    $ctg->{'ebp'} = $ctg_lens->{$c};
	    $ctg->{'ori'} = unknown;
	    $ctg->{'cnt'} = scalar(@{$lg->{$loc}->{$c}});
	    $ctg->{'mrk'} = [];
	    push(@{$ctg->{'mrk'}}, @{$lg->{$loc}->{$c}});
	    push(@{$node->{'ctgs'}}, $ctg);

	    if (defined($lg->{$loc}->{$c}->[0]->{'old'})) {
		$ctg->{'old'} = $lg->{$loc}->{$c}->[0]->{'old'};
	    }
	}

	push(@map, $node);
    }

    #
    # Create nodes in the map to represent contigs that safely span more than one map position.
    #
    my ($cur, $next, $n, $cmn);

    @new = ();
    $n   = undef;
    $cur = shift @map;

    do {
	$next = shift @map if (!defined($next));

	$cmn = find_common_ctgs($cur, $next);

	if (defined($cmn)) {
	    print STDERR "$cmn is common between $cur->{'id'} and $next->{'id'}\n" if ($debug);
	    $n = extract_node($cur, $next, $cmn);
	    push(@new, $cur) if (scalar(@{$cur->{'ctgs'}}) > 0);
	    $cur = $n;
	    #
	    # If we removed the only contig at that node, discard.
	    #
	    $next = undef if (scalar(@{$next->{'ctgs'}}) == 0);

	    if (scalar(@map) == 0 && defined($next)) {
		push(@new, $cur);
		$cur = $next;
	    }
	} else {
	    push(@new, $cur);
	    $cur  = $next;
	    $next = undef;
	}
    } while (scalar(@map) > 0);

    push(@new, $cur);

    @{$merged_map} = @new;
}

sub extract_node {
    my ($cur, $next, $ctg_id) = @_;

    my ($node, $n, $cnt);

    #
    # Create a new node that will hold the contig that is common to more than one mapping
    # node along with the subset of markers from each genetic map node that map to this 
    # contig.
    #
    $node = {};
    $node->{'id'}   = $cur->{'id'} . "|" . $next->{'id'};
    $node->{'ctgs'} = [];

    $cnt = scalar(@{$cur->{'ctgs'}});
    while ($cnt > 0) {
	$n = shift @{$cur->{'ctgs'}};
	$cnt--;

	if ($n->{'id'} eq $ctg_id) {
	    push(@{$node->{'ctgs'}}, $n);
	} else {
	    push(@{$cur->{'ctgs'}}, $n);
	}
    }

    $cnt = scalar(@{$next->{'ctgs'}});
    while ($cnt > 0) {
	$n = shift @{$next->{'ctgs'}};
	$cnt--;

	if ($n->{'id'} eq $ctg_id) {
	    $node->{'ctgs'}->[0]->{'cnt'} += $n->{'cnt'};
	    push(@{$node->{'ctgs'}->[0]->{'mrk'}}, @{$n->{'mrk'}});

	} else {
	    push(@{$next->{'ctgs'}}, $n);
	}
    }

    return $node;
}

sub find_common_ctgs {
    my ($cur, $next) = @_;

    my (%common, $ctg, $cmn_ctg);

    $cmn_ctg = undef;

    foreach $ctg (@{$cur->{'ctgs'}}, @{$next->{'ctgs'}}) {
	$common{$ctg->{'id'}}++;
    }

    foreach $ctg (keys %common) {
	if ($common{$ctg} > 1) {
	    return undef if (defined($cmn_ctg));
	    $cmn_ctg = $ctg;
	}
    }

    return $cmn_ctg;
}

sub traverse {
    my ($map, $fh, $write_markers) = @_;

    my ($node);

    foreach $node (@{$map}) {
	dump_node($node, $write_markers, $fh);
    }
}

sub dump_node {
    my ($node, $write_markers, $fh) = @_;

    $fh = \*STDERR if (!defined($fh));

    my ($ctg, $mrk);

    my %o = (-1 => "Unknown", 0 => "Forward", 1 => "Reverse");

    print $fh "$node->{'id'}\n";

    foreach $ctg (@{$node->{'ctgs'}}) {
	($ctg->{'ori'} == reve) ?
	    print $fh 
	    "  $ctg->{'id'} $ctg->{'ebp'}-$ctg->{'sbp'}bp; orientation: '", $o{$ctg->{'ori'}}, "'" :
	    print $fh 
	    "  $ctg->{'id'} $ctg->{'sbp'}-$ctg->{'ebp'}bp; orientation: '", $o{$ctg->{'ori'}}, "'";

	print $fh " [$ctg->{'cnt'} markers] " if ($write_markers == true);

	if (defined($ctg->{'old'})) {
	    print $fh "[$ctg->{'old'}]";
	}
	print $fh "\n";
	    
	@{$ctg->{'mrk'}} = $ctg->{'ori'} == reve ? 
	    sort sort_markers_rev @{$ctg->{'mrk'}} :
	    sort sort_markers @{$ctg->{'mrk'}};

	foreach $mrk (@{$ctg->{'mrk'}}) {
	    printf($fh "    M%05d: %07dbp, %.03fcM\n", $mrk->{'marker'}, $mrk->{'bp'}, $mrk->{'cm'}) if ($write_markers == true);
	}
    }
}

sub write_php {
    my ($php_path, $doc_path, $lg, $map, $suffix) = @_;

    my ($node, $path, $fh);

    $path = $php_path . "/lg" . $lg . "_" . $suffix . ".php";
    open($fh, ">" . $path) 
	or die("Unable to open PHP LG mapping file '$path' $!\n");

    print $fh 
	"<?php\n",
	"include(\"../header.php\");\n",
	"write_compact_header(\"LG$lg\");\n\n";

    my $i = 1;
    foreach $node (@{$map}) {
	dump_php_node($node, $doc_path, \$i, $fh);
    }

    print $fh 
	"\n\ninclude(\"../footer.php\");\n",
	"?>\n";

    close($fh);
}

sub dump_php_node {
    my ($node, $doc_path, $i, $fh) = @_;

    my ($ctg, $mrk, $str, $path);

    my %o = (-1 => "Unknown", 0 => "Forward", 1 => "Reverse");

    print $fh 
	"echo <<< EOQ\n",
	"<h2 class=\"node\">+ $node->{'id'}</h2>\n";

    foreach $ctg (sort sort_nodes @{$node->{'ctgs'}}) {
	#
	# Check if a log file exists for this scaffold
	#
	$path = $doc_path . $ctg->{'id'} . ".log";
	if (-e $path) {
	    $str = "<a target=\"blank\" href=\"/$org/view_log.php?v=$version&log=$ctg->{'id'}.log\"><span class=\"scaffold\">$ctg->{'id'}</span></a>";
	} else { 
	    $str = "<span class=\"scaffold\">$ctg->{'id'}</span>";
	}

	my $id = $ctg->{'id'} . "_" . $$i;

	print $fh 
	    "<div class=\"node\">\n",
	    "<h3 class=\"scaffold\">",
	    "<a onclick=\"toggle_markers('$id', '\$img_path');\">\n",
	    "<img id=\"$id", "_img\" src=\"\$img_path/caret-u.png\" /></a>\n";

	if ($ctg->{'ori'} == reve) {
	    print $fh 
		$str,
		"<a onclick=\"toggle_markers('$id', '\$img_path');\">",
		 " <span class=\"range\">", commify($ctg->{'ebp'}), "-", commify($ctg->{'sbp'}), "bp</span>; ori: <span class=\"rev\">", 
		$o{$ctg->{'ori'}}, "</span>";
	} elsif ($ctg->{'ori'} == forw) {
	    print $fh 
		$str,
		"<a onclick=\"toggle_markers('$id', '\$img_path');\">",
		" <span class=\"range\">", commify($ctg->{'sbp'}), "-", commify($ctg->{'ebp'}), "bp</span>; ori: <span class=\"for\">", 
		$o{$ctg->{'ori'}}, "</span>";
	} else {
	    print $fh 
		$str,
		"<a onclick=\"toggle_markers('$id', '\$img_path');\">",
		" <span class=\"range\">", commify($ctg->{'sbp'}), "-", commify($ctg->{'ebp'}), "bp</span>; ori: <span class=\"unk\">", 
		$o{$ctg->{'ori'}}, "</span>";
	}

	print $fh " [$ctg->{'cnt'} markers]</a> ";

	if (defined($ctg->{'old'})) {
	    $path = $doc_path . $ctg->{'old'} . ".log";
	    if (-e $path) {
		$str = "<a target=\"blank\" href=\"/$org/view_log.php?v=$version&log=$ctg->{'old'}.log\"><span class=\"scaffold\">$ctg->{'old'}</span></a>";
	    } else { 
		$str = "<span class=\"scaffold\">$ctg->{'old'}.log</span>";
	    }
	    print $fh "[$str]";
	}
	print $fh "</h3>\n";
	    
	@{$ctg->{'mrk'}} = $ctg->{'ori'} == reve ? 
	    sort sort_markers_rev @{$ctg->{'mrk'}} :
	    sort sort_markers @{$ctg->{'mrk'}};

	print $fh 
	    "<div id=\"$id\" style=\"display: none;\">\n",
	    "<ul class=\"marker_list\">\n";
	foreach $mrk (@{$ctg->{'mrk'}}) {
	    printf($fh 
		   "    <li><a target=\"blank\" href=\"\${stacks_url}%d\"><span class=\"marker\">M%05d</span></a>: ".
		   "<span class=\"bp\">%07dbp</span>, <span class=\"cm\">%.03fcM</span></li>\n", 
		   $mrk->{'marker'}, $mrk->{'marker'}, $mrk->{'bp'}, $mrk->{'cm'});
	}
	print $fh 
	    "</ul>\n",
	    "</div>\n",
	    "</div>\n";
	$$i++;
    }

    print $fh "EOQ;\n";
}

sub load_lgs {
    my ($lgs, $hits, $agp, $markers_in_map, $conflict_markers) = @_;

    my (@files, $file, $lg, $line, $href, $cnt);

    build_file_list(\@files);

    foreach $file (@files) {
	#
	# Load a file of the markers for a particular linkage group, ordered by JoinMap
	#
	open(MAP, "<" . $in_path . "/" . $file) 
	    or die("Unable to open LG mapping file '$in_path/$file' $!\n");

	$cnt  = 0;
	($lg) = ($file =~ /lg(\d+)\.tsv/);
	$lgs->{$lg} = {};

	while ($line = <MAP>) {
	    $href = parse_record($line);
	    # print STDERR "Read marker $href->{'marker'} at position $href->{'cm'}\n";
	    next if (!defined($href));

	    #
	    # If this marker doesn't have a contig associated with it, it isn't informative.
	    #
	    if (!defined($hits->{$href->{'marker'}})) {
		# print STDERR "No alignment for marker $href->{'marker'}\n";
		next;
	    }

	    #
	    # This marker has already been defined as conflicting, don't include it.
	    #
	    if (defined($conflict_markers->{$href->{'marker'}})) {
		# print STDERR "Marker has been blacklisted $href->{'marker'}\n";
		next;
	    }

	    $markers_in_map{$href->{'marker'}}++;

 	    if (!defined($lgs->{$lg}->{$href->{'cm'}})) {
		$lgs->{$lg}->{$href->{'cm'}} = {};
	    }

	    #
	    # Look up the position of this marker based on its alignment
	    #
	    $href->{'lg'}     = $lg;
	    $href->{'contig'} = $hits->{$href->{'marker'}}->{'chr'};
	    $href->{'bp'}     = $hits->{$href->{'marker'}}->{'bp'};
	    $href->{'ori'}    = $hits->{$href->{'marker'}}->{'ori'};
	    $href->{'cgr'}    = $hits->{$href->{'marker'}}->{'cgr'};

	    #
	    # Check if this contig has been renamed, record the old name if so.
	    #
	    if (defined($agp->{$href->{'contig'}}->[0]->{'old_scaf'})) {
		$href->{'old'} = $agp->{$href->{'contig'}}->[0]->{'old_scaf'};
	    }

	    #
	    # Record the LG that this marker is mapped to in the genetic map.
	    #
	    $hits->{$href->{'marker'}}->{'lg'} = $lg;
	    $hits->{$href->{'marker'}}->{'cm'} = $href->{'cm'};

	    if (!defined($lgs->{$lg}->{$href->{'cm'}}->{$href->{'contig'}})) {
		$lgs->{$lg}->{$href->{'cm'}}->{$href->{'contig'}} = [];
	    }
	    push(@{$lgs->{$lg}->{$href->{'cm'}}->{$href->{'contig'}}}, $href);

	    $cnt++;
	}

	print STDERR "Loaded ", scalar(keys %{$lgs->{$lg}}), " map locations from '$file' consisting of $cnt markers.\n";
	close(MAP);
    }
}

sub parse_record {
    my ($line) = @_;

    my (@parts, @marker, $href);

    chomp $line;

    return undef if (length($line) == 0);

    @parts = split(/\t/, $line);
    @marker = split(/\|/, $parts[1]);

    $href = {};
    $href->{'marker'} = $marker[0];
    $href->{'cm'}     = $parts[2];

    return $href;
}

sub load_agp_file {
    my ($agp) = @_;

    my ($fh, @parts, $line, $href);

    #
    # Load an AGP file defining gaps in the assembly.
    #  Format: object     object_beg object_end part_number component_type component_id component_beg component_end  orientation
    #  Gaps:   object     object_beg object_end part_number N              gap_length   gap_type      linkage        -empty-
    #          JH556661.1 1          70736 1 W AGAJ01000001.1 1        70736    +
    #          JH556661.1 70737      70782 2 N 46             fragment yes
    #
    open($fh, "<" . $agp_path) or die("Unable to open AGP file '$agp_path' $!\n");

    while ($line = <$fh>) {
        chomp $line;
        @parts = split(/\t/, $line);

	next if (substr($parts[0], 0, 1) eq "#");

	$href = {};
	$href->{'scaf'}  = $parts[0];
	$href->{'start'} = $parts[1];
	$href->{'end'}   = $parts[2];
	$href->{'part'}  = $parts[3];
	$href->{'type'}  = $parts[4];
	# $href->{'len'}   = $href->{'type'} eq "N" ? $parts[5] : $parts[7] - $parts[6] + 1;
	$href->{'len'}   = $href->{'type'} eq "N" ? $parts[5] : $parts[6] - $parts[5] + 1;
	$href->{'5'}     = $parts[5];
	$href->{'6'}     = $parts[6];
	$href->{'7'}     = $parts[7];
	$href->{'8'}     = $parts[8] if (scalar(@parts) >= 9);

	push(@{$agp->{$href->{'scaf'}}}, $href);
    }

    print STDERR "Loaded gap information for ", scalar(keys %{$agp}), " scaffolds.\n";

    close($fh);
}

sub load_sam_aln {
    my ($ctg_lens, $hits, $markers) = @_;

    my ($fh, @parts, $line, $ctg, $len, $flag, $bl_fh, $scaffold);
    my %blacklist;

    if (length($bl_path) > 0) {
        open($bl_fh, "<$bl_path") or die("Unable to open blacklist file '$bl_path'; $!\n");

        while (<$bl_fh>) {
            chomp $_;
            $blacklist{$_}++;
        }

        print STDERR "Read ", scalar(keys %blacklist), " blacklisted markers.\n";

        close($bl_fh);
    }

    #
    # Load a SAM file of alignments linking sequences to markers in the catalog.
    #
    open($fh, "<" . $sam_hits) or die("Unable to open SAM file '$sam_hits' $!\n");

    while ($line = <$fh>) {
	last if ($line !~ /^\@/);
	next if ($line !~ /^\@SQ\t/);
        chomp $line;

	#($ctg, $len) = ($line =~ /^\@SQ\tSN:gi\|\d+\|gb\|(.+)\|\tLN:(.*)$/);
	($ctg, $len) = ($line =~ /^\@SQ\tSN:(.+)\tLN:(.*)$/);
	$ctg_lens->{$ctg} = $len;
    }
    print STDERR "Read ", scalar(keys %{$ctg_lens}), " contigs from SAM header.\n";

    do {
        chomp $line;
        @parts = split(/\t/, $line);

	#
	# If chromosome field is a '*' then read is not aligned.
	#
	if ($parts[2] ne "*" && !defined($blacklist{$parts[0]})) {

	    #
	    # Check which strand this is aligned to: 
	    #   SAM reference: FLAG bit 0x10 - sequence is reverse complemented
	    #
	    $flag = $parts[1] & 16;
	    $flag = $flag >> 4;

	    #
	    # Parse the chromosome name.
	    #
	    #($scaffold) = ($parts[2] =~ /^gi\|\d+\|gb\|(.+)\|$/);
	    $scaffold = $parts[2];

	    $hits->{$parts[0]} = {
		'marker' => $parts[0],
		'chr'    => $scaffold, 
		'bp'     => int($parts[3]), 
		'ori'    => $flag > 0 ? "+" : "-", 
		'len'    => length($parts[9]),
		'cgr'    => $parts[5]};
	}
    } while ($line = <$fh>);

    print STDERR "Loaded ", scalar(keys %{$hits}), " SAM alignments.\n";

    close($fh);
}

sub commify {
    local $_ = shift;
    1 while s/^([-+]?\d+)(\d{3})/$1,$2/;
    return $_;
}

sub build_file_list {
    my ($files) = @_;

    my (@ls, $line, $prefix, $wl_keys, $wl_fh);
    my %whitelist;

    if (length($wl_path) > 0) {
        open($wl_fh, "<$wl_path") or die("Unable to open whitelist file '$wl_path'; $!\n");

        while (<$wl_fh>) {
            chomp $_;
            $whitelist{$_}++;
        }

        $wl_keys = scalar(keys %whitelist);
        print STDERR "Read ", $wl_keys, " whitelisted linkage groups.\n";

        close($wl_fh);
    }

    @ls = `ls -1 $in_path/lg*.tsv 2> /dev/null`;

    if (scalar(@ls) == 0) {
	print STDERR "Unable to locate any input files to process within '$in_path'\n";
	usage();
    }

    foreach $line (@ls) {
	chomp $line;

	($prefix) = ($line =~ /$in_path\/(lg\d+\.tsv)/);

	next if ($wl_keys && !defined($whitelist{$prefix}));

	push(@{$files}, $prefix) if (length($prefix) > 0);
    }
}

sub parse_command_line {
    while (@ARGV) {
	$_ = shift @ARGV;
	if    ($_ =~ /^-p$/) { $in_path  = shift @ARGV; }
	elsif ($_ =~ /^-s$/) { $sam_hits = shift @ARGV; }
	elsif ($_ =~ /^-w$/) { $wl_path  = shift @ARGV; }
	elsif ($_ =~ /^-b$/) { $bl_path  = shift @ARGV; }
	elsif ($_ =~ /^-a$/) { $agp_path = shift @ARGV; }
	elsif ($_ =~ /^-D$/) { $doc_path = shift @ARGV; }
	elsif ($_ =~ /^-m$/) { $min_mark = shift @ARGV; }
	elsif ($_ =~ /^-P$/) { $php_path = shift @ARGV; $write_php++; }
	elsif ($_ =~ /^-d$/) { $debug++; }
	elsif ($_ =~ /^-h$/) { usage(); }
	else {
	    print STDERR "Unknown command line option: '$_'\n";
	    usage();
	}
    }

    if (length($agp_path) == 0) {
	print STDERR "You must specify the AGP file defining assembly gaps.\n";
	usage();
    }

    if (length($sam_hits) == 0) {
	print STDERR "You must specify the SAM alignments file.\n";
	usage();
    }

    if (length($doc_path) == 0 || !-e $doc_path) {
	print STDERR "You must specify a directory to store scaffold assembly documentation.\n";
	usage();
    }
    $doc_path .= "/" if (substr($doc_path, -1) ne "/");
}

sub usage {
    print STDERR <<EOQ; 
order_contigs.pl -p path -s sam -a agp -D doc [-P path] [-w path] [-b path] [-d] [-h]
    p: path to directory containing LG map files.
    s: path to SAM formatted alignments linking map markers to genome contigs.
    a: path to the AGP formatted file defining assembly gaps.
    m: minimum number of markers required to consider a scaffold as attached to a genetic map node.
    D: path to write scaffold documentation data.
    P: output PHP files to this directory.
    w: whitelist file of marker IDs.
    b: blacklist file of marker IDs.
    h: display this help message.
    d: turn on debug output.

EOQ

    exit(0);
}
