// -*-mode:c++; c-style:k&r; c-basic-offset:4;-*-
//
// Copyright 2015-2018, Julian Catchen <jcatchen@illinois.edu>
//
// This file is part of Chromonomer.
//
// Chromonomer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Chromonomer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Chromonomer.  If not, see <http://www.gnu.org/licenses/>.
//

#include "chromonomer.h"

//
// Global variables to hold command-line options.
//
string lg_path;
string sam_path;
string agp_path;
string out_path;
string fasta_path;
string scaffold_wl_path;

bool   verbose          = false;
bool   split_scaffolds  = true;
bool   parse_fasta      = false;
bool   allpaths_agp     = false;
uint   min_node_markers = 2;
uint   join_gap_size    = 100;

uint   fieldw           = 3;
string scaffold_prefix  = "CHRR";
string data_version     = "";
string description      = "";

double max_clipped      = 0.15;
int    min_mapping_qual = 10;
bool   keep_sec_alns    = false;

set<string> scaffold_wl;
set<string> logged_scaffolds;

int 
main (int argc, char* argv[]) 
{
    //
    // Get comma separators for printing out basepair positions.
    //
    setlocale(LC_NUMERIC, "");

    parse_command_line(argc, argv);

    map<string, Marker>           markers;
    map<string, Marker>           conflict_markers;
    map<string, vector<ScafComp>> agp;
    map<string, Scaffold>         scaffolds;

    if (scaffold_wl_path.length() > 0) {
	load_marker_list(scaffold_wl_path, scaffold_wl);
	cerr << "Loaded " << scaffold_wl.size() << " scaffolds on the whitelist to process.\n";
    }

    cerr << "Reading the linkage map, marker alignments, and the assembly definition...\n";

    if (!parse_lg_definitions(lg_path, markers)) {
	cerr << "Unable to parse linkage group definition file.\n";
	exit(1);
    }

    if (!parse_sam_alignments(sam_path, markers)) {
	cerr << "Unable to parse SAM alignment file.\n";
	exit(1);
    }

    prune_misaligned_markers(markers);

    prune_unaligned_markers(markers);

    if (!parse_agp_file(agp_path, agp)) {
	cerr << "Unable to parse AGP assembly definition file.\n";
	exit(1);
    }

    //
    // Scaffold ID is used to number newly created scaffolds when splitting conflicting scaffolds.
    //
    int scaffold_id        = 1;
    int scaffold_splits    = 0;
    int scaffold_cnt       = 0;
    int scaffold_lg_splits = 0;
    int scaffold_lg_cnt    = 0;
    int scaf_splits, scaf_tot;

    cerr << "\nIdentifying and splitting scaffolds that are mapped to more than one linkage group...\n";

    //
    // Find and split scaffolds that map to multiple linkage groups.
    //
    resolve_interlg_conflicts(markers, agp, conflict_markers, scaffold_id, scaffold_cnt, scaffold_splits);

    //
    // Remove conflicting markers generated from scaffolds split across linkage groups
    // from the main marker archive.
    //
    remove_conflict_markers(markers, conflict_markers);

    //
    // Create a list of scaffold objects to store scaffold properties.
    //
    generate_scaffold_list(markers, agp, scaffolds);

    cerr << scaffold_cnt << " scaffolds examined; " << scaffold_splits << " were split resulting in " << scaffolds.size() << " total scaffolds.\n\n";

    //
    // Determine scaffold orientation.
    //
    determine_scaffold_orientation(scaffolds);

    map<string, list<Node>> temp_linkage_map;
    build_linkage_map(scaffolds, temp_linkage_map);
    for (auto it = temp_linkage_map.begin(); it != temp_linkage_map.end(); it++) {
	string lg = it->first;
	write_lg_php(out_path, lg, it->second, "before");
    	write_lg_json(out_path, lg, it->second, "before");
    }

    //
    // On each scaffold, prune out of order markers. Each scaffold must be internally consistent
    // prior to considering splitting it.
    //
    prune_outoforder_markers(scaffolds, conflict_markers);

    //
    // Remove conflicting markers generated from ordering markers on scaffolds.
    //
    remove_conflict_markers(markers, conflict_markers);

    //
    // Construct the linkage map.
    //
    map<string, list<Node>> linkage_map;
    build_linkage_map(scaffolds, linkage_map);

    cerr << "Checking for scaffolds that need to be split on each linkage group.\n";

    for (auto it = linkage_map.begin(); it != linkage_map.end(); it++) {
	string lg = it->first;

	cerr << "Processing linkage group '" << lg << "'...\n";

	string path = out_path + lg + "_scaffold.log";
	ofstream lg_fh(path.c_str());
	if (lg_fh.fail()) {
	    cerr << "Error: unable to open linkage group log file '" << path << "'.\n";
	    exit(1);
	}

	lg_fh << lg << " unmodified:\n";
	traverse(it->second, lg_fh, true);

	//
	// Remove a scaffold from a map node if it has less than min_node_markers supporting it at that node.
	//
	prune_map_nodes(lg, it->second, conflict_markers);

	resolve_conflicts(lg, it->second, agp, markers, conflict_markers, scaffold_id, scaf_tot, scaf_splits);
	scaffold_lg_cnt    += scaf_tot;
	scaffold_lg_splits += scaf_splits;

	write_lg_php(out_path, lg, it->second, "after");
	write_lg_json(out_path, lg, it->second, "after");

	write_map(out_path, lg, it->second);
	cerr << "\n";

	lg_fh.close();
    }

    //
    // Remove conflicting markers generated from failed scaffold splits.
    //
    remove_conflict_markers(markers, conflict_markers);

    cerr << scaffold_lg_cnt << " scaffolds examined on " << linkage_map.size() << " linkage groups; " << scaffold_lg_splits << " scaffolds were split.\n\n";

    summarize(linkage_map, conflict_markers, argc, argv);
    cerr << "\n";

    //
    // Write a map of all the scaffold transformations.
    //
    write_scaffold_map(out_path, linkage_map);

    //
    // Write the final linkage map.
    //
    write_final_linkage_map(out_path, linkage_map);

    //
    // Write the AGP files back out.
    //
    write_scaffold_agp(out_path, agp);

    write_lg_agp(out_path, linkage_map, agp);

    //
    // If requested, translate the FASTA and output a new file.
    //
    if (parse_fasta)
        translate_fasta(out_path, fasta_path, linkage_map, agp);
    
    return 0;
}

int
summarize(map<string, list<Node>> &linkage_map, map<string, Marker> &conflict_markers, int argc, char* argv[])
{
    map<string, ProbScaf> prob_scaf_map;
    map<string, int>      split_dist;
    map<string, long>     lg_lens;

    char doc_str[max_len];
    int  cnt        = 0;
    long tot_len    = 0;
    uint marker_cnt = 0;

    for (auto it = linkage_map.begin(); it != linkage_map.end(); it++) {
	string lg         = it->first;
	list<Node> &nodes = it->second;
        long lg_len = 0;

	for (auto node = nodes.begin(); node != nodes.end(); node++) {

	    for (auto scaf = node->scaffolds.begin(); scaf != node->scaffolds.end(); scaf++) {
		cnt++;
		Scaffold &scaffold = scaf->second;
		marker_cnt += scaffold.markers.size();
                lg_len     += scaffold.len;
                tot_len    += scaffold.len;
                
		string id = scaffold.old.size() == 0 ? scaffold.id : scaffold.old.front().scaffold;

                if (prob_scaf_map.count(id) == 0) {
                    prob_scaf_map[id] = ProbScaf(id);
                    prob_scaf_map[id].lg = lg;
                    prob_scaf_map[id].split_fail = scaffold.split_failure;
                } else {
                    prob_scaf_map[id].split_fail = scaffold.split_failure == true ? scaffold.split_failure : prob_scaf_map[id].split_fail;
                    prob_scaf_map[id].lg = lg == prob_scaf_map[id].lg ? lg : "Multi";
                }
                
		if (split_dist.count(id) == 0)
		    split_dist[id] = 0;
		else
		    split_dist[id]++;
	    }
	}
        lg_lens[lg] = lg_len;
    }

    string   path = out_path + "chromonomer_description.log";
    ofstream doc_fh(path.c_str());
    if (doc_fh.fail()) {
	cerr << "Error: unable to open description '" << path << "'.\n";
	exit(1);
    }
    doc_fh << description << "\n";
    doc_fh.close();

    path = out_path + "chromonomer_summary.log";
    doc_fh.open(path.c_str());
    if (doc_fh.fail()) {
	cerr << "Error: unable to open summary '" << path << "'.\n";
	exit(1);
    }

    init_log(doc_fh, argc, argv);
    
    cerr   << cnt << " scaffolds in the linkage map, from " << split_dist.size() << " original scaffolds.\n";
    doc_fh << cnt << " scaffolds in the linkage map, from " << split_dist.size() << " original scaffolds.\n\n";

    map<int, int> summary;
    for (auto it = split_dist.begin(); it != split_dist.end(); it++) {
	prob_scaf_map[it->first].splits = it->second;
	summary[it->second]++;
    }

    cerr   << "Splits\tNum Scaffolds\n";
    doc_fh << "Splits\tNum Scaffolds\n";
    for (auto it = summary.begin(); it != summary.end(); it++) {
	cerr   << it->first << "\t" << it->second << "\n";
	doc_fh << it->first << "\t" << it->second << "\n";
    }

    cerr   << "\n"
	   << marker_cnt << " markers in the final map; " << conflict_markers.size() << " inconsistent markers excluded.\n";
    doc_fh << "\n"
	   << marker_cnt << " markers in the final map; " << conflict_markers.size() << " inconsistent markers excluded.\n";

    for (auto it = conflict_markers.begin(); it != conflict_markers.end(); it++) {
	if (it->second.old.size() > 0 && prob_scaf_map.count(it->second.old.back().first) > 0)
	    prob_scaf_map[it->second.old.back().first].conflict_markers++;
	else if (prob_scaf_map.count(it->second.scaffold) > 0)
	    prob_scaf_map[it->second.scaffold].conflict_markers++;
    }

    vector<ProbScaf> problem_scaffolds;
    for (auto it = prob_scaf_map.begin(); it != prob_scaf_map.end(); it++)
	problem_scaffolds.push_back(it->second);

    sort(problem_scaffolds.begin(), problem_scaffolds.end(), compare_problem_scaffolds);

    //
    // Log problem scaffolds and markers.
    //
    path = out_path + "problem_scaffolds.tsv";
    ofstream fh(path.c_str());
    if (fh.fail()) {
	cerr << "Error: unable to open problem scaffolds file '" << path << "'.\n";
	exit(1);
    }
    cerr << "\n"
	 << "Writing problem scaffolds to '" << path << "' for further investigation.\n";

    fh << "# Scaffold\tLG\tSplit Fails\tConflicted Markers\tSplit Cnt\n";

    for (uint i = 0; i < problem_scaffolds.size(); i++)
	if (problem_scaffolds[i].conflict_markers == 0 && problem_scaffolds[i].splits == 0)
	    continue;
	else
	    fh << problem_scaffolds[i].id << "\t"
	       << problem_scaffolds[i].lg << "\t"
	       << (problem_scaffolds[i].split_fail == true ? "true" : "false") << "\t"
	       << problem_scaffolds[i].conflict_markers << "\t"
	       << problem_scaffolds[i].splits << "\n";

    fh.close();

    cerr << "\n" 
         << "Chromosome lengths:\n";
    doc_fh << "\n" 
	   << "Chromosome lengths:\n";
    for (auto it = lg_lens.begin(); it != lg_lens.end(); it++) {
        snprintf(doc_str, max_len, "%5s: %'10ldbp\n", it->first.c_str(), it->second);
        cerr   << doc_str;
        doc_fh << doc_str;
    }

    snprintf(doc_str, max_len, "%'10ldbp", tot_len);
    cerr << "Total length: "
	 << doc_str << "\n";
    doc_fh << "Total length: "
	   << doc_str << "\n";

    doc_fh.close();

    return 0;
}

int
build_linkage_map(map<string, Scaffold> &scaffolds, map<string, list<Node>> &linkage_map)
{
    map<string, map<double, Node>> node_map;
    cerr << "Assembling " << scaffolds.size() << " scaffolds into the linkage map.\n";
    //
    // Organize scaffolds onto their appropriate linkage nodes.
    //
    for (auto it = scaffolds.begin(); it != scaffolds.end(); it++) {
	Scaffold &scaffold = it->second;

	for (uint i = 0; i < scaffold.markers.size(); i++) {
	    if (node_map[scaffold.lg].count(scaffold.markers[i].cM) == 0)
		node_map[scaffold.lg][scaffold.markers[i].cM] = Node(scaffold.markers[i].cM);

	    if (node_map[scaffold.lg][scaffold.markers[i].cM].scaffolds.count(scaffold.id) == 0) {
		node_map[scaffold.lg][scaffold.markers[i].cM].scaffolds[scaffold.id] = 
		    Scaffold(scaffold.id, scaffold.start_bp, scaffold.end_bp, scaffold.len, scaffold.ori, scaffold.split_failure);
		node_map[scaffold.lg][scaffold.markers[i].cM].scaffolds[scaffold.id].old = scaffold.old;
	    }
	    node_map[scaffold.lg][scaffold.markers[i].cM].scaffolds[scaffold.id].markers.push_back(scaffold.markers[i]);
	}
    }

    //
    // Create a list of nodes for each linkage group, representing each mapping location
    //
    for (auto it = node_map.begin(); it != node_map.end(); it++) 
	for (auto j = it->second.begin(); j != it->second.end(); j++)
	    linkage_map[it->first].push_back(j->second);

    for (auto it = linkage_map.begin(); it != linkage_map.end(); it++)
	it->second.sort(compare_node_cM);

    //
    // Create nodes in the map to represent contigs that safely span more than one map position.
    //
    for (auto it = linkage_map.begin(); it != linkage_map.end(); it++) {
	string lg       = it->first;
	list<Node> &map = it->second;

	cerr << "Assembling linkage group '" << lg << "'...\n";

	span_scaffolds_across_nodes(map);
    }

    cerr << "\n";

    return 0;
}

int
span_scaffolds_across_nodes(list<Node> &map)
{
    string cmn;
    Node   n;
    list<Node>::iterator cur, next;

    cur  = map.begin();
    next = cur;
    next++;

    while (next != map.end()) {
	//
	// Are there any scaffolds that span these two mapping nodes?
	//
	cmn = find_common_scaffolds(*cur, *next);

	if (cmn.length() > 0) {
	    // cerr << cmn << " is common between " << cur->id << " and " << next->id << "\n";

	    //
	    // Extract out the scaffold common to both mapping nodes and place it in a separate node.
	    //
	    n = extract_node(*cur, *next, cmn);

	    //
	    // If we removed the only contig at the current node, discard.
	    //
	    if (cur->scaffolds.size() == 0)
		map.erase(cur);
	    
	    cur = map.insert(next, n);

	    //
	    // If we removed the only contig at the next node, discard.
	    //
	    if (next->scaffolds.size() == 0)
		map.erase(next);

	    next = cur;
	    next++;

	} else {
	    cur++;
	    next++;
	}
    }

    return 0;
}

string
find_common_scaffolds(Node &cur, Node &next)
{
    string cmn_scaf;
    uint   cmn_cnt = 0;

    for (auto i = cur.scaffolds.begin(); i != cur.scaffolds.end(); i++)
	if (next.scaffolds.count(i->first) > 0) {
	    cmn_scaf = i->first;
	    cmn_cnt++;
	}

    if (cmn_cnt > 1)
	cerr << "  Warning: more than one scaffold common to map nodes " << cur.cM << " and " << next.cM << "\n";

    return cmn_scaf;
}

Node
extract_node(Node &cur, Node &next, string scaffold_id)
{
    //
    // Create a new node that will hold the scaffold that is common to these two mapping
    // nodes along with the subset of markers from each genetic map node that map to this 
    // scaffold.
    //
    Node node;

    node.cM = cur.cM;

    for (auto i = cur.cMs.begin(); i != cur.cMs.end(); i++)
	node.cMs.insert(*i);
    for (auto i = next.cMs.begin(); i != next.cMs.end(); i++)
	node.cMs.insert(*i);
    node.scaffolds[scaffold_id] = cur.scaffolds[scaffold_id];

    //
    // Add the markers that were in the scaffold in the node next.
    //
    for (uint i = 0; i < next.scaffolds[scaffold_id].markers.size(); i++)
	node.scaffolds[scaffold_id].markers.push_back(next.scaffolds[scaffold_id].markers[i]);

    cur.scaffolds.erase(scaffold_id);
    next.scaffolds.erase(scaffold_id);

    return node;
}

int
traverse(list<Node> &lg, ofstream &fh, bool write_markers)
{
    char doc_str[max_len];
    map<ori_type, string> o;
    o[ori_unknown] = "Unknown";
    o[ori_forward] = "Forward";
    o[ori_reverse] = "Reverse";

    for (auto it = lg.begin(); it != lg.end(); it++) {
	Node &node = *it;

	fh << node.id() << "\n";

	for (auto i = node.scaffolds.begin(); i != node.scaffolds.end(); i++) {
	    Scaffold &scaffold = i->second;

	    if (scaffold.ori == ori_reverse)
		fh << "  " << scaffold.id << " " << scaffold.end_bp << "-" << scaffold.start_bp << "bp; orientation: '" << o[scaffold.ori] << "'";
	    else
		fh << "  " << scaffold.id << " " << scaffold.start_bp << "-" << scaffold.end_bp << "bp; orientation: '" << o[scaffold.ori] << "'";

	    if (write_markers == true)
		fh << " [" << scaffold.markers.size() << " markers] ";

	    // if (node.scaffolds[i].) {
	    //     print $fh "[$ctg->{'old'}]";
	    // }
	    fh << "\n";

	    if (write_markers == true) {
		if (scaffold.ori == ori_forward)
		    sort(scaffold.markers.begin(), scaffold.markers.end(), compare_marker_cM_bp);
		else
		    sort(scaffold.markers.begin(), scaffold.markers.end(), compare_marker_cM_bp_reverse);

		for (uint j = 0; j < scaffold.markers.size(); j++) {
		    snprintf(doc_str, max_len, 
			     "    M%s: %' 9dbp, %.03fcM\n", 
			     scaffold.markers[j].id.c_str(), scaffold.markers[j].bp, scaffold.markers[j].cM);
		    fh << doc_str;
		}
	    }
	}
    }

    return 0;
}

int
prune_outoforder_markers(map<string, Scaffold> &scaffolds, map<string, Marker> &conflict_markers)
{
    char     doc_str[max_len];
    string   path;
    ofstream doc_fh;

    cerr << "Removing out of order markers from scaffolds with known orientation:\n";
    uint total_removed = 0;
    uint total         = 0;

    for (auto it = scaffolds.begin(); it != scaffolds.end(); it++) {
	Scaffold       &scaffold = it->second;
	vector<Marker> &markers  = scaffold.markers;

	total += markers.size();

	if (scaffold.ori == ori_unknown) continue;

	//
	// Open a log file to document changes made to this scaffold.
	//
	path = out_path + it->first + ".log";
	if (logged_scaffolds.count(path) > 0) {
	    doc_fh.open(path.c_str(), std::ofstream::app);
	} else {
	    doc_fh.open(path.c_str());
	    logged_scaffolds.insert(path);
	}
	if (doc_fh.fail()) {
	    cerr << "Error: unable to open scaffold documentation '" << path << "'.\n";
	    exit(1);
	}

	//
	// Sort the markers according to cM and bp, depending on the orientation.
	//
	if (scaffold.ori == ori_forward)
	    sort(markers.begin(), markers.end(), compare_marker_cM_bp);
	else
	    sort(markers.begin(), markers.end(), compare_marker_cM_bp_reverse);

	doc_fh << "Orientation of " << it->first << " is " << (scaffold.ori == ori_forward ? "forward" : "reverse") <<  "\n"
	       << "Unmodified marker ordering:\n";

	for (uint i = 0; i < markers.size(); i++) {
	    snprintf(doc_str, max_len,
		     "%03d\t%s\t%s\t%*.*fcM <=> %9dbp\t%s\t[%s]\n",
		     i, markers[i].id.c_str(), markers[i].lg.c_str(), fieldw+4, fieldw, markers[i].cM, markers[i].bp, markers[i].scaffold.c_str(), markers[i].cigar.c_str());
	    doc_fh << doc_str;
	}

	//
	// Examine each coherent path through the sorted markers (markers consistently down or up ordered).
	//
	int marker_index, cur, next, cnt;
	int stop             = (int) markers.size() - 1;
	int max_marker_cnt   = 0;
	int max_marker_index = 0;

	vector<int> path(markers.size(), 0);
	vector<int> blacklist(markers.size(), 0);
	vector<int> max_path;

	for (marker_index = 0; marker_index < stop; marker_index++) {
	    for (cur = 0; cur < (int) path.size(); cur++) { path[cur] = 0; }

	    cur       = marker_index;
	    path[cur] = 1;
	    cnt       = 1;

	    if (scaffold.ori == ori_forward) {
		do {
		    next = cur + 1;
		    //
		    // Skip out of order markers.
		    //
		    while (markers[next].bp < markers[cur].bp && next <= stop) next++;
		    if (next <= stop) {
			path[next] = 1;
			cnt++;
		    }
		    cur = next;
		} while (cur < stop);

	    } else {
		do {
		    next = cur + 1;
		    //
		    // Skip out of order markers.
		    //
		    while (markers[next].bp > markers[cur].bp && next <= stop) next++;
		    if (next <= stop) {
			cnt++;
			path[next] = 1;
		    }
		    cur = next;
		} while (cur < stop);
	    }

	    if (cnt > max_marker_cnt) {
		max_marker_cnt   = cnt;
		max_marker_index = marker_index;
		max_path         = path;
	    }
	}

	// cerr << "    Max path: ";
	// for (uint j = 0; j < max_path.size(); j++)
	//     cerr << max_path[j];
	// cerr << "\n";

	doc_fh << "\n" 
	       << "Removing out of order markers:\n";

	//
	// Test to see if we can add markers to the maximal path.
	//
	for (int i = 0; i < max_marker_index; i++)
	    blacklist[i] = 1;

	int index = (int) path.size() - 1;
	do {
	    path = max_path;
	    while (index > 0 && blacklist[index] == 0 && path[index] == 0) index--;
	    //
	    // Test this position to see if removing this marker adds more markers overall.
	    //
	    blacklist[index] = 1;
	    cnt = markers_in_path(markers, path, blacklist, scaffold.ori);

	    //
	    // If it didn't add more markers, take it back out.
	    //
	    if (cnt <= max_marker_cnt) {
		blacklist[index] = 0;
	    } else {
		max_marker_cnt = cnt;
		max_path       = path;
	    }

	    index--;
	} while (index > max_marker_index);

	index = 0;
	do {
	    //
	    // Test this position to see if adding this marker adds more markers overall.
	    //
	    blacklist[index] = 0;
	    cnt = markers_in_path(markers, path, blacklist, scaffold.ori);

	    //
	    // If it didn't add more markers, take it back out.
	    //
	    if (cnt <= max_marker_cnt) {
		blacklist[index] = 1;
	    } else {
		max_marker_cnt = cnt;
		max_path       = path;
	    }

	    index++;
	} while (index < max_marker_index);
	    
	vector<Marker> tmp;
	for (uint i = 0; i < markers.size(); i++)
	    tmp.push_back(markers[i]);
	markers.clear();

	//
	// Discard markers that are not on the final marker path.
	//
	uint removed_cnt = 0;
	for (uint i = 0; i < max_path.size(); i++)
	    if (max_path[i]) {
		markers.push_back(tmp[i]);
	    } else {
		conflict_markers[tmp[i].id] = tmp[i];
		snprintf(doc_str, max_len, "  %03d: removing marker %s [%8dbp]\n", i, tmp[i].id.c_str(), tmp[i].bp);
		doc_fh << doc_str;
		removed_cnt++;
	    }

	if (removed_cnt > 0 && verbose) cerr   << "  Scaffold '" << it->first << "': removed " << removed_cnt << " markers.\n";
	doc_fh << "Scaffold '" << it->first << "': removed " << removed_cnt << " markers.\n\n"
	       << "Final marker ordering:\n";
	for (uint i = 0; i < markers.size(); i++) {
	    snprintf(doc_str, max_len,
		     "%03d\t%s\t%s\t%*.*fcM <=> %9dbp\t%s\t[%s]\n",
		     i, markers[i].id.c_str(), markers[i].lg.c_str(), fieldw+4, fieldw, markers[i].cM, markers[i].bp, markers[i].scaffold.c_str(), markers[i].cigar.c_str());
	    doc_fh << doc_str;
	}
	doc_fh << "\n";

	doc_fh.close();

	total_removed += removed_cnt;
    }

    cerr << "Removed " << total_removed << " out of order markers from a total of " << total << " markers.\n\n";

    return 0;
}

int
markers_in_path(vector<Marker> &markers, vector<int> &path, vector<int> &blacklist, ori_type orientation)
{
    for (uint i = 0; i < path.size(); i++) path[i] = 0;

    uint next;
    uint stop = markers.size() - 1;
    uint cur  = 0;
    uint cnt  = 1;

    while (cur <= stop && blacklist[cur] == 1)
	cur++;
    path[cur] = 1;

    if (orientation == ori_forward) {
	do {
	    next = cur + 1;
	    //
	    // Skip out of order markers.
	    //
	    while (blacklist[next] == 1 || (markers[next].bp < markers[cur].bp && next <= stop)) next++;
	    if (next <= stop) {
		path[next] = 1;
		cnt++;
	    }
	    cur = next;
	} while (cur < stop);

    } else {
	do {
	    next = cur + 1;
	    //
	    // Skip out of order markers.
	    //
	    while (blacklist[next] == 1 || (markers[next].bp > markers[cur].bp && next <= stop)) next++;
	    if (next <= stop) {
		cnt++;
		path[next] = 1;
	    }
	    cur = next;
	} while (cur < stop);
    }

    return cnt;
}

int
determine_scaffold_orientation(map<string, Scaffold> &scaffolds)
{
    //
    // We will caclulate the linear regression line if plotting cM positions on the X-axis
    // versus basepair positions on the Y-axis. If the slope is positive, the orientation is
    // 'forward'; if the slope isnegative, the orientation is 'reverse'.
    //
    uint total   = 0;
    uint forward = 0;
    uint reverse = 0;

    cerr << "Examining scaffolds to determine orientation...\n";

    for (auto it = scaffolds.begin(); it != scaffolds.end(); it++) {
	total++;

	//
	// We need at least two mapping nodes (cM) to determine orientation.
	//
	set<double> nodes;
	for (uint i = 0; i < it->second.markers.size(); i++)
	    nodes.insert(it->second.markers[i].cM);
	if (nodes.size() < 2) continue;

	double n      = it->second.markers.size();
	double avg_x  = 0.0;
	double avg_y  = 0.0;
	double avg_xy = 0.0;
	double avg_x2 = 0.0;

	for (uint i = 0; i < it->second.markers.size(); i++) {
	    // cerr << "  X: " << it->second.markers[i].cM << ", Y: " << it->second.markers[i].bp << "\n";
 	    avg_x  += it->second.markers[i].cM;
	    avg_y  += (double) it->second.markers[i].bp;
	    avg_x2 += pow(it->second.markers[i].cM, 2);
	    avg_xy += it->second.markers[i].cM * (double) it->second.markers[i].bp;
	}

	avg_x  = avg_x  / n;
	avg_y  = avg_y  / n;
	avg_x2 = avg_x2 / n;
	avg_xy = avg_xy / n;

	// cerr << "Avg x: " << avg_x << "; Avg y: " << avg_y << "; Avg x^2: " << avg_x2 << "; Avg XY: " << avg_xy << "\n";

	double m = ((avg_x * avg_y) - avg_xy) / (pow(avg_x, 2) - avg_x2);
	// double b = avg_y - (m * avg_x);

	//cerr << "Scaffold: " << it->first << ", regression, slope (m): " << m << ", Y-intercept (b): " << b << "\n";

	if (m < 0) {
	    it->second.ori = ori_reverse;
	    reverse++;
	} else if (m > 0) {
	    it->second.ori = ori_forward;
	    forward++;
	} else {
	    it->second.ori = ori_unknown;
	}
    }

    cerr << "Examined " << total << " scaffolds; " 
	 << forward << " in forward orientation; " 
	 << reverse << " in reverse orientation; " 
	 << (total - forward - reverse) << " unknown.\n\n";

    return 0;
}

ori_type
recheck_scaffold_orientation(vector<Marker> &markers)
{
    //
    // After splitting scaffolds, re-check their orientation.
    //   1) Make sure there are two or more mapping nodes
    //   2) If so, sort markers and check orientation.
    //

    //
    // We need at least two mapping nodes (cM) to determine orientation.
    //
    set<double> nodes;
    for (uint i = 0; i < markers.size(); i++)
        nodes.insert(markers[i].cM);
    if (nodes.size() < 2)
        return ori_unknown;

    sort(markers.begin(), markers.end(), compare_marker_bp);
    
    if (markers.front().cM < markers.back().cM)
        return ori_forward;

    else if (markers.front().cM > markers.back().cM)
        return ori_reverse;

    else
        return ori_unknown;
}

int
generate_scaffold_list(map<string, Marker> &markers, 
		       map<string, vector<ScafComp>> &agp, 
		       map<string, Scaffold> &scaffolds)
{
    pair<map<string, Scaffold>::iterator, bool> ret;
    string id;

    for (auto i = markers.begin(); i != markers.end(); i++) {
	id = i->second.scaffold;

	if (scaffolds.count(id) == 0) {
	    ret = scaffolds.insert(pair<string, Scaffold>(id, Scaffold(id)));
            Scaffold &s = ret.first->second;

	    //
	    // Look this scaffold up in the AGP definitions.
	    //
            map<string, vector<ScafComp>>::iterator it = agp.find(id);
	    if (it == agp.end()) {
		cerr << "Warning: unable to locate scaffold '" << id << "' in AGP definitions.\n";
		exit(1);
	    }

            //
            // Check if any split failures occurred.
            //
            vector<ScafComp> &scomps = it->second;
            for (uint j = 0; j < scomps.size(); j++)
                if (scomps[j].split_fail == true) {
                    s.split_failure = true;
                    break;
                }

	    s.start_bp = scomps.front().start;
	    s.end_bp   = scomps.back().end;
	    s.len      = s.end_bp - s.start_bp + 1;
	    s.lg       = i->second.lg;

	    //
	    // If the AGP components have been renamed, this scaffold was split between
	    // linkage groups (resolve_interlg_conflicts()). If so, record the previous 
	    // history into the new Scaffold object.
	    //
	    if (scomps.front().old.size() > 0)
		s.old.push_back(OldScafComp(scomps.front().old.front().scaffold,
                                            scomps.front().old.front().start,
                                            scomps.back().old.front().end));
	}

	scaffolds[i->second.scaffold].markers.push_back(i->second);
	scaffolds[i->second.scaffold].nodes.insert(i->second.cM);
    }

    return 0;
}

int
remove_conflict_markers(map<string, Marker> &markers, map<string, Marker> &conflicts)
{
    map<string, Marker> tmp;

    for (auto it = markers.begin(); it != markers.end(); it++)
	if (conflicts.count(it->first) == 0)
	    tmp[it->first] = it->second;

    markers = tmp;

    return 0;
}

int
prune_map_nodes(string lg, list<Node> &lg_map, map<string, Marker> &conflict_markers)
{
    char     doc_str[max_len];
    string   path;
    ofstream doc_fh;

    uint scaffold_cnt = 0;
    uint marker_cnt   = 0;

    //
    // Create a map of which nodes on this linkage map contain each scaffold.
    //
    map<string, vector<Node>> scaffold_map;

    for (auto node = lg_map.begin(); node != lg_map.end(); node++)
	for (auto j = node->scaffolds.begin(); j != node->scaffolds.end(); j++)
	    scaffold_map[j->first].push_back(*node);

    //
    // Iterate over each scaffold on this linkage map chromosome and decide which nodes to prune.
    //
    for (auto i = scaffold_map.begin(); i != scaffold_map.end(); i++) {
	string id           = i->first;
	vector<Node> &nodes = i->second;

	if (nodes.size() == 1) 
	    continue;

	scaffold_cnt++;

	//
	// Open a log file to document changes made to this scaffold.
	//
	path = out_path + id + ".log";
	if (logged_scaffolds.count(path) > 0) {
	    doc_fh.open(path.c_str(), std::ofstream::app);
	} else {
	    doc_fh.open(path.c_str());
	    logged_scaffolds.insert(path);
	}
	if (doc_fh.fail()) {
	    cerr << "Error: unable to open scaffold documentation '" << path << "'.\n";
	    exit(1);
	}

	vector<pair<Node, uint>> prune_list;

	for (uint i = 0; i < nodes.size(); i++)
	    if (nodes[i].scaffolds[id].markers.size() < min_node_markers)
		prune_list.push_back(make_pair(nodes[i], nodes[i].scaffolds[id].markers.size()));

	//
	// If the scaffold is on more than one node, but all nodes are below the minimum marker limit,
	// just keep the first node.
	//
	if (prune_list.size() == nodes.size())
	    prune_list.erase(prune_list.begin());

	//
	// Prune the scaffolds out of the linkage map.
	//
	list<Node>::iterator it, prev, next;
	for (uint i = 0; i < prune_list.size(); i++) {
	    it = find(lg_map.begin(), lg_map.end(), prune_list[i].first);

	    if (it == lg_map.end()) {
		cerr << "Can no longer find " << prune_list[i].first.id() << "\n";
		continue;
	    }

	    doc_fh << "Removing scaffold '" << id << "' from node '" << it->id() << "' as it is only supported by " << prune_list[i].second << " markers.\n";

	    //
	    // Blacklist markers
	    //
	    vector<Marker> &markers = it->scaffolds[id].markers;

	    for (uint j = 0; j < markers.size(); j++) {
		snprintf(doc_str, max_len, "  Removing marker %s\n", markers[j].id.c_str());
		doc_fh << doc_str;
		conflict_markers[markers[j].id] = markers[j];
		marker_cnt++;
	    }

	    //
	    // Erase the scaffold from this node.
	    //
	    it->scaffolds.erase(id);

	    //
	    // If it is the only scaffold on this node, erase the node, too.
	    //
	    if (it->scaffolds.size() == 0)
		lg_map.erase(it);
	}

	doc_fh << "\n";
	doc_fh.close();
    }

    //
    // As we have pruned a bunch of nodes, we need to search again for scaffolds that
    // are common across mapping nodes and can be extracted into their own nodes so
    // the scaffold is not split in the next stage of the algorithm.
    //
    span_scaffolds_across_nodes(lg_map);

    cerr << scaffold_cnt << " scaffolds pruned from map nodes supported by fewer than " << min_node_markers << " markers; removed " << marker_cnt << " markers supporting those nodes.\n";

    return 0;
}

int
resolve_conflicts(string lg, list<Node> &lg_map, 
		  map<string, vector<ScafComp>> &agp, 
		  map<string, Marker> &markers,
		  map<string, Marker> &conflict_markers,
		  int &scaffold_id, int &tot_scaf, int &scaf_split)
{
    //
    // Create a map of which nodes on this linkage map contain each scaffold.
    //
    map<string, vector<Node>> scaffold_map;

    for (auto node = lg_map.begin(); node != lg_map.end(); node++)
	for (auto j = node->scaffolds.begin(); j != node->scaffolds.end(); j++)
	    scaffold_map[j->first].push_back(*node);

    uint     confounded_scaffold_cnt = 0;
    int      split  = 0;
    int      failed = 0;
    string   path;
    ofstream doc_fh;
    char     doc_str[max_len];

    //
    // Iterate over each scaffold on this linkage map chromosome.
    //
    for (auto i = scaffold_map.begin(); i != scaffold_map.end(); i++) {
	string id           = i->first;
	vector<Node> &nodes = i->second;

	ori_type ori = nodes[0].scaffolds[id].ori;

	if (nodes.size() == 1) 
	    continue;

	confounded_scaffold_cnt++;

	stringstream node_str;
	for (uint i = 0; i < nodes.size(); i++)
	    node_str << nodes[i].id() << ", ";

	//
	// Open a log file to document changes made to this scaffold.
	//
	path = out_path + id + ".log";
	if (logged_scaffolds.count(path) > 0) {
	    doc_fh.open(path.c_str(), std::ofstream::app);
	} else {
	    doc_fh.open(path.c_str());
	    logged_scaffolds.insert(path);
	}
	if (doc_fh.fail()) {
	    cerr << "Error: unable to open scaffold documentation '" << path << "'.\n";
	    exit(1);
	}

	if (verbose)
	    cerr << "  Scaffold " << id <<  " is mapped to multiple linkage nodes: " << node_str.str().substr(0, node_str.str().length() - 2) << "\n";
	doc_fh << "Linkage group " << lg << "; Scaffold '" << id << "'; Orientation: " << (ori == ori_forward ? "Forward" : "Reverse") << "\n\n"
	       << "Scaffold " << id <<  " is mapped to multiple linkage nodes: " << node_str.str().substr(0, node_str.str().length() - 2) << "\n\n";

	int segment = 1;
	map<int, vector<Marker>> segments;
	map<string, int> segment_map;

	for (uint j = 0; j < nodes.size(); j++) {
	    vector<Marker> &marker_list = nodes[j].scaffolds[id].markers;

	    segment_map[nodes[j].id()] = segment;

	    for (uint k = 0; k < marker_list.size(); k++) {
		segments[segment].push_back(marker_list[k]);

                snprintf(doc_str, max_len, 
                         "%03d\t%s\t%4s\t%*.*fcM <=> %'10dbp => segment %2d [%s]\n", 
                         k+1, marker_list[k].id.c_str(), marker_list[k].lg.c_str(), fieldw+4, fieldw, marker_list[k].cM, marker_list[k].bp, segment, marker_list[k].cigar.c_str());
		doc_fh << doc_str;
	    }
	    segment++;
	}
	doc_fh << "\n"
	       << "Scaffold '" << id << "' will be split into " << segments.size()  << " pieces.\n";

	for (auto j = segments.begin(); j != segments.end(); j++) {
	    if (ori == ori_forward)
		sort(j->second.begin(), j->second.end(), compare_marker_cM_bp);
	    else
		sort(j->second.begin(), j->second.end(), compare_marker_cM_bp_reverse);

	    snprintf(doc_str, max_len, 
		     "Segment %2d: %'9dbp - %'9dbp\n", 
		     j->first, j->second[0].bp, j->second.back().bp);
	    doc_fh << doc_str;
	}
	doc_fh << "\n";

        if (id == "JH556705.1") {
            cerr << "Here.\n";
        }
        
	if (split_scaffold(id, segments, ori, agp, markers, conflict_markers, scaffold_id, doc_fh)) {
	    split++;
	    //
	    // Adjust the scaffolds in the linkage map to reflect the split components.
	    //
	    // Each element in the scaffold_map is one node that contains the old scaffold. The
	    // split_scaffold() function split the occurance of each scaffold in a node into a
	    // new scaffold. Here we create a new Scaffold object for each one adjusting the 
	    // start/end basepair coordinates.
	    //
	    list<Node>::iterator it;
	    vector<Node> &nodes = scaffold_map[id];

	    for (uint j = 0; j < nodes.size(); j++) {

                //
                // It is possible that a segment was pruned out of the map if it could not be split
                // cleanly. In this case, the segment will no longer be defined and that segment
                // should be pruned from the map.
                //
                segment = segment_map[nodes[j].id()];
                if (segments.count(segment) == 0) {
                    nodes[j].scaffolds.erase(id);
                
                    it = find(lg_map.begin(), lg_map.end(), nodes[j]);
                    it->scaffolds.erase(id);
                    if (it->scaffolds.size() == 0)
                        lg_map.erase(it);
                    continue;
                }

                vector<Marker> &marker_list = nodes[j].scaffolds[id].markers;

		//
		// The split_scaffold function modified the main markers list. Create a new
		// scaffold object and Recopy the markers into it.
		//
		Scaffold new_scaffold;
		new_scaffold.id       = markers[marker_list[0].id].scaffold;
		new_scaffold.start_bp = 1;
		new_scaffold.end_bp   = agp[new_scaffold.id].back().end - agp[new_scaffold.id].front().start + 1;
		new_scaffold.len      = new_scaffold.end_bp - new_scaffold.start_bp + 1;

		//
		// Record the previous scaffold components. Find the first and last AGP component, then
		// cycle through their previous coordinates creating the scaffold boundaries.
		//
		vector<OldScafComp> &os_front = agp[new_scaffold.id].front().old;
		vector<OldScafComp> &os_back  = agp[new_scaffold.id].back().old;
		if (os_front.size() != os_back.size())
		    cerr << "Error recording prevous scaffold components for '" << new_scaffold.id << "'\n";

		for (uint k = 0; k < os_front.size(); k++)
		    new_scaffold.old.push_back(OldScafComp(os_front[k].scaffold, os_front[k].start, os_back[k].end));

		for (uint k = 0; k < marker_list.size(); k++)
		    new_scaffold.markers.push_back(markers[marker_list[k].id]);

                //
                // Re-check the orientation of the new scaffold.
                //
                new_scaffold.ori = recheck_scaffold_orientation(new_scaffold.markers);

		//
		// Replace the old copy of the scaffold in the scaffold_map and the linkage map.
		//
		nodes[j].scaffolds.erase(id);
		nodes[j].scaffolds[new_scaffold.id] = new_scaffold;
                
		it = find(lg_map.begin(), lg_map.end(), nodes[j]);
		it->scaffolds.erase(id);
		it->scaffolds[new_scaffold.id] = new_scaffold;
	    }

	    //
	    // Delete the existing scaffold representation, since it has been split out onto new contigs.
	    //
	    agp.erase(id);

	} else {
	    failed++;

	    //
	    // Remove scaffold from linkage map; delete node if it is the only scaffold present.
	    //
            int segment = segments.begin()->first;
	    list<Node>::iterator it;

	    for (uint j = 0; j < scaffold_map[id].size(); j++) {
		vector<Node> &nodes = scaffold_map[id];

		if (segment_map[nodes[j].id()] == segment) {
		    //
		    // Record this split failure in the remaining scaffold segment.
		    //
		    it = find(lg_map.begin(), lg_map.end(), nodes[j]);
		    it->scaffolds[id].split_failure = true;
		    continue;
		}

 		it = find(lg_map.begin(), lg_map.end(), nodes[j]);
		it->scaffolds.erase(id);

		if (it->scaffolds.size() == 0)
		    lg_map.erase(it);
	    }
	}
	doc_fh << "\n";
	doc_fh.close();
    }

    tot_scaf   = scaffold_map.size();
    scaf_split = split;
    cerr << scaffold_map.size() << " total scaffolds; " 
	 << confounded_scaffold_cnt << " scaffolds mapped to multiple nodes in a linkage group; " 
	 << scaf_split << " successfully split; " << failed << " failed to split.\n";

    doc_fh.close();

    return 0;
}

int
resolve_interlg_conflicts(map<string, Marker> &markers, 
			  map<string, vector<ScafComp>> &agp, 
			  map<string, Marker> &conflict_markers,
			  int &scaffold_id, int &scaf_tot, int &scaf_split)
{
    map<string, vector<Marker>> scaffold_map;

    //
    // Record which markers are mapped to which scaffolds.
    //
    for (auto i = markers.begin(); i != markers.end(); i++)
	scaffold_map[i->second.scaffold].push_back(i->second);

    uint     confounded_scaffold_cnt = 0;
    string   path;
    ofstream doc_fh;
    char     doc_str[max_len];
    set<string>::iterator s;
    map<int, vector<Marker>>::iterator t;

    //
    // Iterate over each scaffold and identify those that are mapped to multiple linkage groups.
    //
    for (auto j = scaffold_map.begin(); j != scaffold_map.end(); j++) {
	string          scaffold    = j->first;
	vector<Marker> &marker_list = j->second;

	set<string> lgs;
	for (uint k = 0; k < marker_list.size(); k++)
	    lgs.insert(marker_list[k].lg);

	//
	// Scaffold is mapped to only one linkage group, no modification necessary.
	//
	if (lgs.size() == 1)
	    continue;

	confounded_scaffold_cnt++;

	stringstream lg_str;
	for (s = lgs.begin(); s != lgs.end(); s++)
	    lg_str << *s << ", ";

	if (verbose)
	    cerr << "  " << scaffold <<  " is mapped to multiple linkage groups: " << lg_str.str().substr(0, lg_str.str().length() - 2) << "\n";

	//
	// Open a log file to document changes made to this scaffold.
	//
	path = out_path + scaffold + ".log";
	doc_fh.open(path.c_str());
        logged_scaffolds.insert(path);

	if (doc_fh.fail()) {
	    cerr << "Error: unable to open scaffold documentation '" << path << "'.\n";
	    exit(1);
	}

	doc_fh << "Scaffold " << scaffold <<  " is mapped to multiple linkage groups: " << lg_str.str().substr(0, lg_str.str().length() - 2) << "\n\n";

	//
	// Divide the scaffold into segments, each segment appearing on a different linkage group.
	//
	sort(marker_list.begin(), marker_list.end(), compare_marker_bp);

	string cur     = marker_list[0].lg;
	uint   segment = 1;
	map<int, vector<Marker>> segments;

	for (uint k = 0; k < marker_list.size(); k++) {
	    if (marker_list[k].lg != cur) {
		segment++;
		cur = marker_list[k].lg;
	    }
	    segments[segment].push_back(marker_list[k]);

	    snprintf(doc_str, max_len, 
		     "%03d\t%s\t%4s\t% 8.3fcM <=> %'10dbp => segment %2d [%s]\n", 
                     k+1, marker_list[k].id.c_str(), marker_list[k].lg.c_str(), marker_list[k].cM, marker_list[k].bp, segment, marker_list[k].cigar.c_str());
	    doc_fh << doc_str;
	}
	doc_fh << "\n";

	for (t = segments.begin(); t != segments.end(); t++) {
	    sort(t->second.begin(), t->second.end(), compare_marker_bp);

	    snprintf(doc_str, max_len, 
		     "Segment %2d: %'10dbp - %'10dbp\n", 
		     t->first, t->second[0].bp, t->second.back().bp);
	    doc_fh << doc_str;
	}
	doc_fh << "\n";

	if (split_scaffold(scaffold, segments, ori_unknown, agp, markers, conflict_markers, scaffold_id, doc_fh)) {
	    //
	    // Delete the existing scaffold representation, since it has been split out onto new contigs.
	    //
	    agp.erase(scaffold);
	    scaf_split++;
	}

	doc_fh << "\n";
	doc_fh.close();
    }

    scaf_tot = scaffold_map.size();

    cerr << scaffold_map.size() << " total scaffolds; "
	 << confounded_scaffold_cnt << " scaffolds mapped to multiple linkage groups.\n";

    return 0;
}

int
print_scaffold_map(vector<ScafComp> &s, ofstream &doc_fh, bool compress)
{ 
    char doc_str[max_len];

    int cnt = s.size();
    doc_fh << "Scaffold " << s[0].scaffold << ", " << cnt << " components.\n";

    //
    // Find components that contain marker annotations and/or split failures.
    //
    vector<int> pattern, indexes;
    if (compress == true) {
	for (int i = 0; i < cnt; i++) {
	    pattern.push_back(0);
	    if (s[i].annotations.size() > 0)
		indexes.push_back(i);
            else if (s[i].split_fail)
                indexes.push_back(i);
	}

	int start, end, index;
	for (uint i = 0; i < indexes.size(); i++) {
	    index = indexes[i];
	    start = index - 3 < 0 ? 0 : index - 3;
	    end   = index + 3 > cnt - 1 ? cnt - 1 : index + 3;

	    for (int j = start; j <= end; j++)
		pattern[j] = 1;
	}
    } else {
	for (int i = 0; i < cnt; i++)
	    pattern.push_back(1);
    }

    int dots = 0;
    for (int i = 0; i < cnt; i++) {
	if (i == 0 && pattern[0] == 0) {
	    snprintf(doc_str, max_len, "%03d   ", i);
	    doc_fh << doc_str;
	}

	if (pattern[i] == 0) {
	    doc_fh << ".";
	    dots++;
	    if (dots >= 60) {
	        doc_fh << "\n      ";
		dots = 0;
	    }
	    if (i < cnt - 1 && pattern[i + 1] == 1)
		doc_fh << "\n";

	} else {
	    dots = 0;
	    if (s[i].type == gap) {
		snprintf(doc_str, max_len, 
			"%03d      Gap start: % 7d, end: % 7d, size: % 7d\n", 
			i, s[i].start, s[i].end, s[i].len);
	    } else {
                if (s[i].split_fail)
                    snprintf(doc_str, max_len, 
                             "%03d   Contig start: % 7d, end: % 7d, size: % 7d [split failure]\n", 
                             i, s[i].start, s[i].end, s[i].len);
                else
                    snprintf(doc_str, max_len, 
                             "%03d   Contig start: % 7d, end: % 7d, size: % 7d\n", 
                             i, s[i].start, s[i].end, s[i].len);
	    }
	    doc_fh << doc_str;

	    if (s[i].annotations.size() > 0) {
		for (uint k = 0; k < s[i].annotations.size(); k++)
		    doc_fh << "         Segment " << s[i].annotations[k].segment <<  " " << (s[i].annotations[k].type == seg_beg ? "begin" : "end") << "\n";
	    }

	    if (i < cnt - 1 && pattern[i + 1] == 0) {
		snprintf(doc_str, max_len, "%03d   ", i + 1);
		doc_fh << doc_str;
	    }
	}
    }

    doc_fh << "\n\n";

    return 0;
}

int    
write_scaffold_map(string out_path, map<string, list<Node>> &linkage_map)
{
    //
    // Open a log to record scaffold splits.
    //
    string path = out_path + "split_scaffold_map.tsv";
    ofstream fh(path.c_str());
    if (fh.fail()) {
	cerr << "Error: unable to open scaffold log file '" << path << "'.\n";
	exit(1);
    }

    cerr << "Writing split scaffold map to '" << path << "'.\n";

    fh << "# Old scaffold	old start BP	old end BP	new scaffold	new start BP	new end BP	linkage group\n";

    for (auto i = linkage_map.begin(); i != linkage_map.end(); i++) {
	string lg         = i->first;
	list<Node> &nodes = i->second;

	map<string, vector<Node>> scaffold_map;

	for (auto node = nodes.begin(); node != nodes.end(); node++)
	    for (auto j = node->scaffolds.begin(); j != node->scaffolds.end(); j++)
		scaffold_map[j->first].push_back(*node);

	//
	// Iterate over each scaffold on this linkage map chromosome.
	//
	for (auto k = scaffold_map.begin(); k != scaffold_map.end(); k++) {
	    string id           = k->first;
	    vector<Node> &nodes = k->second;

	    for (auto node = nodes.begin(); node != nodes.end(); node++) {

		Scaffold &scaffold = node->scaffolds[id];

		//
		// If there are two 'old' entries, then this scaffold was split twice, once
		// to split between linkage groups, and once within a linkage group.
		//
		if (scaffold.old.size() > 1)
		    fh << scaffold.old[1].scaffold << "\t"
		       << scaffold.old[1].start    << "\t"
		       << scaffold.old[1].end      << "\t"
		       << scaffold.old[0].scaffold << "\t"
		       << scaffold.old[0].start    << "\t"
		       << scaffold.old[0].end      << "\t"
		       << "\n";
		if (scaffold.old.size() > 0)
		    fh << scaffold.old[0].scaffold << "\t"
		       << scaffold.old[0].start    << "\t"
		       << scaffold.old[0].end      << "\t"
		       << scaffold.id              << "\t"
		       << scaffold.start_bp        << "\t"
		       << scaffold.end_bp          << "\t"
		       << lg                       << "\n";
	    }
	}
    }

    fh.close();

    return 0;
}

int
write_lg_agp(string out_path, map<string, list<Node>> &linkage_map, map<string, vector<ScafComp>> &agp)
{
    string path = out_path + scaffold_prefix + "_genome.agp";
    ofstream fh(path.c_str());
    if (fh.fail()) {
	cerr << "Error: unable to AGP output file, '" << path << "'.\n";
	exit(1);
    }

    //
    // We will record the scaffolds that are in the linkage map so we can output the remaining unplaced scaffolds.
    //
    set<string> scaffolds_in_map;
    
    cerr << "Writing AGP genome description to '" << path << "'.\n";

    fh << "##agp-version	2.0\n"
       << "# Format: object object_beg object_end part_number  W  component_id component_beg component_end  orientation\n"
       << "#   Gaps: object object_beg object_end part_number  N  gap_length    gap_type       linkage\n";

    for (auto i = linkage_map.begin(); i != linkage_map.end(); i++) {
	string lg         = i->first;
	list<Node> &nodes = i->second;

	uint part_num = 1;
	uint cur_bp   = 1;
	uint end_bp   = 1;

	for (auto node = nodes.begin(); node != nodes.end(); node++) {

	    for (auto j = node->scaffolds.begin(); j != node->scaffolds.end(); j++) {
		Scaffold &scaffold = j->second;

		end_bp = cur_bp + scaffold.len - 1;

		fh << lg       << "\t"
		   << cur_bp   << "\t"
		   << end_bp   << "\t"
		   << part_num << "\t"
		   << "W"      << "\t";
		if (scaffold.old.size() > 0) {
		    scaffolds_in_map.insert(scaffold.old.front().scaffold);
		    
		    fh << scaffold.old.front().scaffold << "\t"
		       << scaffold.old.front().start    << "\t"
		       << scaffold.old.front().end      << "\t"
		       << (scaffold.ori == ori_forward ? "+" : "-") << "\n";
		} else {
		    scaffolds_in_map.insert(scaffold.id);
		    
		    fh << scaffold.id       << "\t"
		       << scaffold.start_bp << "\t"
		       << scaffold.end_bp   << "\t"
		       << (scaffold.ori == ori_forward ? "+" : "-") << "\n";
		}
		
		cur_bp = end_bp + 1;
		part_num++;

		if (next(j) != node->scaffolds.end() || next(node) != nodes.end()) {
		    //
		    // Add a join_gap_size (100bp) length of Ns, representing a gap of unknown size.
		    //
		    end_bp   = cur_bp + join_gap_size - 1;

		    fh << lg            << "\t"
		       << cur_bp        << "\t"
		       << end_bp        << "\t"
		       << part_num      << "\t"
		       << "U"           << "\t"
		       << join_gap_size << "\t"
		       << "scaffold"    << "\t"
		       << "yes"         << "\t"
		       << "map"         << "\n";

		    cur_bp = end_bp + 1;
		    part_num++;
		}
	    }
	}
    }

    fh.close();

    //
    // Write the remaining, unplaced scaffolds.
    //
    path = out_path + scaffold_prefix + "_unplaced_scaffolds.agp";
    fh.open(path.c_str());
    if (fh.fail()) {
	cerr << "Error: unable to AGP output file, '" << path << "'.\n";
	exit(1);
    }

    fh << "##agp-version	2.0\n"
       << "# Format: object object_beg object_end part_number  W  component_id component_beg component_end  orientation\n"
       << "#   Gaps: object object_beg object_end part_number  N  gap_length    gap_type       linkage\n";

    for (auto it = agp.begin(); it != agp.end(); it++) {

	if (scaffolds_in_map.count(it->first))
	    continue;

	vector<ScafComp> &comp = it->second;

	for (uint i = 0; i < comp.size(); i++) {
	    fh << comp[i].scaffold << "\t"
	       << comp[i].start    << "\t"
	       << comp[i].end      << "\t"
	       << comp[i].part_num << "\t";
	    if (comp[i].type == gap)
		fh << "N"              << "\t"
		   << comp[i].len      << "\t"
		   << comp[i].gap_type << "\t"
		   << (comp[i].linkage ? "yes" : "no") << "\n";
	    else
		fh << "W"               << "\t"
		   << comp[i].contig_id << "\t"
		   << comp[i].comp_beg  << "\t"
		   << comp[i].comp_end  << "\t"
		   << (comp[i].ori == ori_forward ? "+" : "-") << "\n";
	}
    }

    fh.close();

    return 0;
}

int
write_scaffold_agp(string out_path, map<string, vector<ScafComp>> &agp)
{
    string path = out_path + scaffold_prefix + "_scaffolds.agp";
    ofstream fh(path.c_str());
    if (fh.fail()) {
	cerr << "Error: unable to AGP output file, '" << path << "'.\n";
	exit(1);
    }

    cerr << "Writing AGP scaffold descriptions to '" << path << "'.\n";

    fh << "##agp-version	2.0\n"
       << "# Format: object object_beg object_end part_number  W  component_id component_beg component_end  orientation\n"
       << "#   Gaps: object object_beg object_end part_number  N  gap_length    gap_type       linkage\n";

    for (auto it = agp.begin(); it != agp.end(); it++) {
	vector<ScafComp> &comp = it->second;

	for (uint i = 0; i < comp.size(); i++) {
	    fh << comp[i].scaffold << "\t"
	       << comp[i].start    << "\t"
	       << comp[i].end      << "\t"
	       << comp[i].part_num << "\t";
	    if (comp[i].type == gap)
		fh << "N"              << "\t"
		   << comp[i].len      << "\t"
		   << comp[i].gap_type << "\t"
		   << (comp[i].linkage ? "yes" : "no") << "\n";
	    else
		fh << "W"               << "\t"
		   << comp[i].contig_id << "\t"
		   << comp[i].comp_beg  << "\t"
		   << comp[i].comp_end  << "\t"
		   << (comp[i].ori == ori_forward ? "+" : "-") << "\n";
	}
    }

    fh.close();

    return 0;
}

int
write_map(string out_path, string lg, list<Node> &lg_map)
{
    string path = out_path + lg + "_scaffold.tsv";
    ofstream fh(path.c_str());
    if (fh.fail()) {
	cerr << "Error: unable to open file to write linkage group map, '" << path << "'.\n";
	exit(1);
    }

    cerr << "Writing linkage group '" << lg << "' to '" << path << "'\n";

    uint master_bp = 1;

    for (auto node = lg_map.begin(); node != lg_map.end(); node++) {
	for (auto it = node->scaffolds.begin(); it != node->scaffolds.end(); it++) {
	    Scaffold &scaf = it->second;

	    fh << master_bp << "\t"
	       << scaf.id   << "\t"
	       << (scaf.ori == ori_reverse ? "-" : "+") << "\t";
	    if (scaf.ori == ori_reverse)
		fh << scaf.end_bp << "\t" << scaf.start_bp;
	    else
		fh << scaf.start_bp << "\t" << scaf.end_bp;
	    if (scaf.old.size() > 0)
		fh << "\t[" << scaf.old.back().scaffold << "]\n";
	    else
		fh << "\n";

	    master_bp += scaf.end_bp - scaf.start_bp + 1;
	}
    }

    fh.close();

    return 0;
}

int
translate_fasta(string out_path, string fasta_path,
                map<string, list<Node>> &linkage_map,
                map<string, vector<ScafComp>> &agp)
{
    char clen[max_len], ulen[max_len], tlen[max_len], olen[max_len], glen[max_len];
    long unmodified_fasta_len = 0;
    long unplaced_scaff_len   = 0;
    long chromosome_len       = 0;
    long added_gap_len        = 0;
    long tot_added_gap_len    = 0;

    //
    // Create a list of scaffolds that have been integrated into the chromonome; those
    // that haven't will be immediately output.
    //
    set<string> scaffold_list;
    for (auto i = linkage_map.begin(); i != linkage_map.end(); i++) {
        for (auto node = i->second.begin(); node != i->second.end(); node++) {
            for (auto j = node->scaffolds.begin(); j != node->scaffolds.end(); j++) {
                Scaffold scaf = j->second;

                if (scaf.old.size() > 0)
                    scaffold_list.insert(scaf.old.front().scaffold);
                else
                    scaffold_list.insert(scaf.id);
            }
        }
    }

    string path = out_path + scaffold_prefix + "_integrated.fa";
    ofstream out_fh(path);
    if (out_fh.fail()) {
        cerr << "Error opening FASTA output file: '" << path << "'\n";
        return 0;
    }

    cerr << "Translating FASTA file and writing new FASTA to '" << path << "'.\n";
    
    //
    // Open a temporary file. We will store unintegrated sequences here until
    // all the integrated sequences have been written.
    //
    FILE *tmp_fh;
    if ((tmp_fh = tmpfile()) == NULL) {
        cerr << "Unable to open temporary file to cache FASTA sequences.\n";
        return 0;
    }

    Input *in_fh = NULL;

    uint pos = fasta_path.find_last_of(".");
    
    if (fasta_path.substr(pos + 1) == "gz")
        in_fh = new GzFasta(fasta_path);
    else
        in_fh = new Fasta(fasta_path);

    map<string, Seq *> seqs;
    map<string, Seq *>::iterator seqs_it;
    Seq *s;

    while ((s = in_fh->next_seq()) != NULL) {
        //
        // Sum the length of the FASTA sequences.
        unmodified_fasta_len += strlen(s->seq);

        if (scaffold_list.count(string(s->id)) == 0) {
            //
            // Scaffold is not in the integrated assembly.
            //
            write_fasta(tmp_fh, s);
            unplaced_scaff_len += strlen(s->seq);
            delete s;
        } else {
            seqs[s->id] = s;
        }
    }

    delete in_fh;
    
    //
    // Write out the integrated sequences.
    //
    char *p;
    string seqbuf;
    stringstream chr_seq;
    for (auto i = linkage_map.begin(); i != linkage_map.end(); i++) {
	string lg         = i->first;
	list<Node> &nodes = i->second;
        chr_seq.str("");
        added_gap_len = 0;

	for (auto node = nodes.begin(); node != nodes.end(); node++) {
            
	    for (auto j = node->scaffolds.begin(); j != node->scaffolds.end(); j++) {
		Scaffold &scaffold = j->second;

		if (scaffold.old.size() > 0) {
		    OldScafComp &old_scaffold = scaffold.old.front();

                    seqs_it = seqs.find(old_scaffold.scaffold);
                    if (seqs_it != seqs.end()) {
                        p  = seqs_it->second->seq;
                        p += old_scaffold.start - 1;
                        seqbuf.assign(p, old_scaffold.end - old_scaffold.start + 1);
                    } else {
                        cerr << "  Error: unable to find scaffold '" << old_scaffold.scaffold << "', excluding from FASTA file.\n";
                        continue;
                    }

		} else {
                    seqs_it = seqs.find(scaffold.id);
                    if (seqs_it != seqs.end()) {
                        p  = seqs_it->second->seq;
                        p += scaffold.start_bp - 1;
                        seqbuf.assign(p, scaffold.end_bp - scaffold.start_bp + 1);
                    } else {
                        cerr << "  Error: unable to find scaffold '" << scaffold.id << "', excluding from FASTA file.\n";
                        continue;
                    }
		}

                if (scaffold.ori == ori_reverse)
                    rev_complement(seqbuf);

                chr_seq << seqbuf;
                
		if (next(j) != node->scaffolds.end() || next(node) != nodes.end()) {
		    //
		    // Add a join_gap_size (100bp by default) length of Ns, representing a gap of unknown size.
		    //
                    seqbuf.assign(join_gap_size, 'N');
                    added_gap_len += join_gap_size;
                    chr_seq << seqbuf;
		}
            }
	}

        seqbuf = chr_seq.str();
        write_fasta(out_fh, lg, seqbuf);

        chromosome_len    += seqbuf.length();
        tot_added_gap_len += added_gap_len;
        if (verbose) {
            snprintf(clen, max_len, "%' 9d", (int) seqbuf.length());
            snprintf(glen, max_len, "%'ld", added_gap_len);
            cerr << "  Writing chromosome '" << lg << "': " << clen << "bp [" << glen << "bp added as gaps between scaffold joins]\n";
        }
    }

    //
    // Write out the unintegrated sequences from the temporary file.
    //
    rewind(tmp_fh);
    char buf[max_len];
    
    while (!feof(tmp_fh)) {
        if (fgets(buf, max_len, tmp_fh) == NULL)
            break;
        out_fh << buf;
    }

    fclose(tmp_fh);
    out_fh.close();

    for (auto i = seqs.begin(); i != seqs.end(); i++)
        delete i->second;
    seqs.clear();

    snprintf(olen, max_len, "%'ld", unmodified_fasta_len);
    snprintf(clen, max_len, "%'ld", chromosome_len);
    snprintf(glen, max_len, "%'ld", tot_added_gap_len);
    snprintf(ulen, max_len, "%'ld", unplaced_scaff_len);
    snprintf(tlen, max_len, "%'ld", (chromosome_len + unplaced_scaff_len));

    cerr << "  Original, unmodified FASTA length: " << olen << "bp.\n"
         << "  Wrote "
         << clen << "bp of integrated sequence [" << glen << "bp scaffold join gaps]; "
         << ulen << "bp of unplaced sequences; "
         << "total length: " << tlen << "bp.\n";

    path = out_path + "chromonomer_summary.log";
    ofstream doc_fh(path, ofstream::app);
    if (doc_fh.fail()) {
	cerr << "Error: unable to open summary '" << path << "'.\n";
	exit(1);
    }
    snprintf(clen, max_len, "%' 12ld", chromosome_len);
    snprintf(ulen, max_len, "%' 12ld", unplaced_scaff_len);
    snprintf(tlen, max_len, "%' 12ld", (chromosome_len + unplaced_scaff_len));

    doc_fh << "\nUnmodified genome sequence FASTA: " << olen << "bp.\n"
           << "Integrated genome sequence FASTA:\n"
           << "  Wrote "
           << clen << "bp of integrated sequence [" << glen << "bp scaffold join gaps];\n"
           << "        " << ulen << "bp of unplaced sequences;\n"
           << "        " << tlen << "bp total length.\n";
    doc_fh.close();

    return 1;
}

int
rev_complement(string &seq)
{
    const char *p, *q;

    int len   = seq.length();
    int j     = 0;
    char *com = new char[len + 1];

    q = seq.c_str();

    for (p = q + len - 1; p >= q; p--) {
        switch (*p) {
        case 'A':
        case 'a':
            com[j] = 'T';
            break;
        case 'C':
        case 'c':
            com[j] = 'G';
            break;
        case 'G':
        case 'g':
            com[j] = 'C';
            break;
        case 'T':
        case 't':
            com[j] = 'A';
            break;
        case 'N':
        case 'n':
            com[j] = 'N';
            break;
        case 'W':
        case 'w':
            com[j] = 'W';
            break;
        case 'S':
        case 's':
            com[j] = 'S';
            break;
        case 'M':
        case 'm':
            com[j] = 'K';
            break;
        case 'K':
        case 'k':
            com[j] = 'M';
            break;
        case 'R':
        case 'r':
            com[j] = 'Y';
            break;
        case 'Y':
        case 'y':
            com[j] = 'R';
            break;
        default:
            com[j] = 'N';
            break;
        }
        j++;
    }
    com[len] = '\0';

    seq.assign(com);

    delete [] com;

    return 0;
}

int
write_fasta(FILE *fh, Seq *s)
{
    char *p     = s->seq;
    int   plen  = strlen(p);
    int   lines = ceil((double) plen / (double) fasta_line_len);

    string id = ">" + string(s->id) + "\n";
    fputs(id.c_str(), fh);

    while (lines > 1) {
        fwrite(p, sizeof(char), fasta_line_len, fh);
        fputs("\n", fh);
        p += fasta_line_len;
        lines--;
    }

    plen = strlen(p);
    fwrite(p, sizeof(char), plen, fh);
    fputs("\n", fh);
        
    return 0;
}

int
write_fasta(ofstream &fh, string &id, string &s)
{
    const char *p = s.c_str();
    int   plen    = strlen(p);
    int   lines   = ceil((double) plen / (double) fasta_line_len);

    fh << ">" << id << "\n";

    char buf[fasta_line_len + 1];
    buf[fasta_line_len] = '\0';

    while (lines > 1) {
        strncpy(buf, p, fasta_line_len);
        fh << buf << "\n";
        p += fasta_line_len;
        lines--;
    }

    plen = strlen(p);
    strncpy(buf, p, plen);
    buf[plen] = '\0';
    fh << buf << "\n";
        
    return 0;
}

int
write_lg_php(string doc_path, string lg, list<Node> &lg_map, string suffix)
{
    string path = doc_path + lg + "_" + suffix + ".php";
    ofstream fh(path);
    if (fh.fail()) {
	cerr << "Error: unable to open LG PHP file '" << path << "'.\n";
	exit(1);
    }

    fh << "<?php\n"
       << "include(\"../header.php\");\n"
       << "write_compact_header(\"" << lg << "\");\n\n";

    int i = 1;
    for (auto node = lg_map.begin(); node != lg_map.end(); node++) {
	dump_php_node(*node, doc_path, i, fh);
	i++;
    }

    fh << "\n\ninclude(\"../footer.php\");\n" 
       << "?>\n";

    fh.close();

    return 0;
}

int
dump_php_node(Node node, string doc_path, int &i, ofstream &fh) 
{
    char doc_str[max_len];
    map<ori_type, string> o;
    o[ori_unknown] = "Unknown";
    o[ori_forward] = "Forward";
    o[ori_reverse] = "Reverse";
    struct stat buffer;

    fh << "echo <<< EOQ\n"
       << "<h2 class=\"node\">+ " << node.id() << "</h2>\n";

    for (auto it = node.scaffolds.begin(); it != node.scaffolds.end(); it++) {
	Scaffold scaf = it->second;

	//
	// Check if a log file exists for this scaffold
	//
	string path = doc_path + scaf.id + ".log";
	string str;
	if (stat(path.c_str(), &buffer) == 0) {
	    str = "<a target=\"blank\" href=\"/chromonomer/view_log.php?v=" + data_version + "&log=" + scaf.id + "\"><span class=\"scaffold\">" + scaf.id + "</span></a>";
	} else { 
	    str = "<span class=\"scaffold\">" + scaf.id + "</span>";
	}

	stringstream id;
	id << node.id() << "_" << i;

	fh << "<div class=\"node\">\n"
	   << "<h3 class=\"scaffold\">"
	   << "<a onclick=\"toggle_markers('" << id.str() << "', '$img_path');\">\n"
	   <<  "<img id=\"" << id.str() << "_img\" src=\"$img_path/caret-u.png\" /></a>\n";

	if (scaf.ori == ori_reverse) {
	    fh << str
	       << "<a onclick=\"toggle_markers('" << id.str() << "', '$img_path');\">"
	       << " <span class=\"range\">" << scaf.end_bp << "-" << scaf.start_bp << "bp</span>; ori: <span class=\"rev\">"
	       << o[scaf.ori] << "</span>";
	} else if (scaf.ori == ori_forward) {
	    fh << str
	       << "<a onclick=\"toggle_markers('" << id.str() << "', '$img_path');\">"
	       << " <span class=\"range\">" << scaf.start_bp << "-" << scaf.end_bp << "bp</span>; ori: <span class=\"for\">"
	       << o[scaf.ori] << "</span>";
	} else {
	    fh << str
	       << "<a onclick=\"toggle_markers('" << id.str() << "', '$img_path');\">"
	       << " <span class=\"range\">" << scaf.start_bp << "-" << scaf.end_bp << "bp</span>; ori: <span class=\"unk\">"
	       << o[scaf.ori] << "</span>";
	}

	fh << " [" << scaf.markers.size() << " markers]</a> ";

	if (scaf.old.size() > 0) {
	    path = doc_path + scaf.old.back().scaffold + ".log";
	    if (stat(path.c_str(), &buffer) == 0) {
		str = "<a target=\"blank\" href=\"/chromonomer/view_log.php?v=" + data_version + "&log=" + 
		    scaf.old.back().scaffold + "\"><span class=\"scaffold\">" + scaf.old.back().scaffold + "</span></a>";
	    } else { 
		str = "<span class=\"scaffold\">" + scaf.old.back().scaffold + ".log</span>";
	    }
	    fh << "[" << str << "]";
	}
	    
	fh << "</h3>\n"
	   << "<div id=\"" << id.str() << "\" style=\"display: none;\">\n"
	   << "<ul class=\"marker_list\">\n";

	for (uint j = 0; j < scaf.markers.size(); j++) {
	    snprintf(doc_str, max_len,
		     "    <li><a target=\"blank\" href=\"${stacks_url}%s\"><span class=\"marker\">M%s</span></a>: <span class=\"bp\">%07dbp</span>, <span class=\"cm\">%.*fcM</span></li>\n", 
		     scaf.markers[j].id.c_str(), scaf.markers[j].id.c_str(), scaf.markers[j].bp, fieldw, scaf.markers[j].cM);
	    fh << doc_str;
 	}

	fh << "</ul>\n"
	   << "</div>\n"
	   << "</div>\n";
	i++;
    }

    fh << "EOQ;\n";

    return 0;
}

int
write_lg_json(string doc_path, string lg, list<Node> &lg_map, string suffix)
{
    string path = doc_path + lg + "_" + suffix + ".json";
    ofstream fh(path);
    if (fh.fail()) {
	cerr << "Error: unable to open LG JSON file '" << path << "'.\n";
	exit(1);
    }

    map<ori_type, string> o;
    o[ori_unknown] = "Unknown";
    o[ori_forward] = "Forward";
    o[ori_reverse] = "Reverse";

    fh << "{\"lg\":\"" << lg << "\",\"path\":" << "\"/chromonomer/view_log.php?v=" <<  data_version << "&log=\"" << ",\"cMs\":[";

    set<double> cMs;
    for (auto i = lg_map.begin(); i != lg_map.end(); i++)
	for (auto j = i->cMs.begin(); j != i->cMs.end(); j++)
	    cMs.insert(*j);

    for (auto j = cMs.begin(); j != cMs.end(); j++) {
	if (j != cMs.begin())
	    fh << ",";
	fh << "\"" << std::setprecision(fieldw) << *j << "\"";
    }
    
    fh << "], \"nodes\":[";

    for (auto i = lg_map.begin(); i != lg_map.end(); i++) {
	Node node = *i;
	
	if (i != lg_map.begin())
	    fh << ",";
	fh << "{\"cMs\":[";

	for (auto j = node.cMs.begin(); j != node.cMs.end(); j++) {
	    if (j != node.cMs.begin())
		fh << ",";
	    fh << "\"" << *j << "\"";
	}

	fh << "],\"scaffolds\":[";

	for (auto k = node.scaffolds.begin(); k != node.scaffolds.end(); k++) {
	    Scaffold scaf = k->second;

	    string log_path = out_path + scaf.id + ".log";
	    bool   logged   = logged_scaffolds.count(log_path) > 0 ? true : false;

	    if (k != node.scaffolds.begin())
		fh << ",";
	    
	    fh << "{\"id\":\"" << scaf.id << "\",\"s_bp\":\"" << scaf.start_bp << "\",\"e_bp\":\"" << scaf.end_bp << "\",\"ori\":\"" << o[scaf.ori] << "\",\"log\":\"" << logged << "\",\"old\":[";

	    for (uint n = 0; n < scaf.old.size(); n++) {
		if (n > 0)
		    fh << ",";
		fh << "\"" << scaf.old[n].scaffold << "\"";
	    }

	    fh << "]}";
	}
	fh << "]}";
    }
    
    fh << "],\"markers\":[";

    int node_index = 0;
    for (auto i = lg_map.begin(); i != lg_map.end(); i++) {
	Node node = *i;

	for (auto k = node.scaffolds.begin(); k != node.scaffolds.end(); k++) {
	    Scaffold scaf = k->second;

	    for (uint n = 0; n < scaf.markers.size(); n++) {
		if (i != lg_map.begin() || k != node.scaffolds.begin() ||  n > 0)
		    fh << ",";
		fh << "{"
		   << "\"id\":\"" << scaf.markers[n].id
		   << "\",\"bp\":\"" << scaf.markers[n].bp
		   << "\",\"cM\":\"" << scaf.markers[n].cM
		   << "\",\"s\":\"" << scaf.id
		   << "\",\"node\":\"" << node_index
		   << "\"}";
	    }
	}
	node_index++;
    }
    fh << "]}";
    
    fh.close();

    return 0;
}

int
write_final_linkage_map(string doc_path, map<string, list<Node>> &linkage_map)
{
    string path = doc_path + scaffold_prefix + "_linkage_map.tsv";
    ofstream fh(path);
    if (fh.fail()) {
	cerr << "Error: unable to open final linkage map file '" << path << "'.\n";
	exit(1);
    }

    fh << "# LG\t"
       << "Marker\t"
       << "cM\t"
       << "Scaffold\t"
       << "Scaffold Orientation\t"
       << "Scaffold BP\t"
       << "LG BP\t"
       << "NodeIndex\n";

    for (auto it = linkage_map.begin(); it != linkage_map.end(); it++) {
	string lg = it->first;

        uint cur_bp     = 1;
        uint node_index = 0;

        for (auto i = it->second.begin(); i != it->second.end(); i++) {
            Node node = *i;

            for (auto k = node.scaffolds.begin(); k != node.scaffolds.end(); k++) {
                Scaffold scaf = k->second;

                for (uint n = 0; n < scaf.markers.size(); n++) {
                    fh << lg                 << "\t"
                       << scaf.markers[n].id << "\t"
                       << scaf.markers[n].cM << "\t"
                       << scaf.id            << "\t"
                       << (scaf.ori == ori_forward ? "+" : "-") << "\t"
                       << scaf.markers[n].bp << "\t"
                       << (scaf.ori == ori_forward ? cur_bp + scaf.markers[n].bp - 1 : cur_bp + (scaf.len - scaf.markers[n].bp)) << "\t"
                       << node_index         << "\n";
                }

                cur_bp += scaf.len + join_gap_size;
            }
            node_index++;
        }
    }

    fh.close();

    return 0;
}

int
prune_misaligned_markers(map<string, Marker> &markers)
{
    map<string, vector<string>> aln_map;
    map<string, Marker> tmp;

    char buf[id_len + 1];
    uint multiple_alns       = 0;
    uint mult_markers_at_pos = 0;

    buf[id_len] = '\0';
    
    for (auto it = markers.begin(); it != markers.end(); it++) {
        Marker &m = it->second;

        //
        // Skip markers with no alignment.
        //
        if (m.scaffold.length() == 0) {
            tmp[m.id] = m;
            continue;
        }

        snprintf(buf, id_len, "%s|%d|%s", m.scaffold.c_str(), m.bp, (m.strand == strand_minus ? "minus" : "plus"));
        aln_map[buf].push_back(m.id);
    }

    for (auto j = aln_map.begin(); j != aln_map.end(); j++) {
        vector<string> &marker_ids = j->second;

        if (marker_ids.size() > 1) {
            mult_markers_at_pos += marker_ids.size();

            // for (uint k = 0; k < marker_ids.size(); k++)
            //     cerr << "Marker " << marker_ids[k] << " aligned to " << j->first << "\n";
            continue;
        }

        tmp[marker_ids[0]] = markers[marker_ids[0]];

        if (markers[marker_ids[0]].aln_cnt > 1)
            multiple_alns++;
    }
    
    markers.clear();
    markers = tmp;

    cerr << "  Removed " << mult_markers_at_pos << " markers where at least two markers aligned to the same genomic position. " << markers.size() << " markers remain.\n";

    if (multiple_alns > 0)
        cerr << "  WARNING: more than one alignment was submitted for " << multiple_alns << " markers; only the last provided alignment was retained for each.\n";

    return 0;
}

int
prune_unaligned_markers(map<string, Marker> &markers)
{
    map<string, Marker> tmp;

    uint pruned = 0;

    for (auto it = markers.begin(); it != markers.end(); it++)
	if (it->second.bp > 0)
	    tmp[it->first] = it->second;
	else
	    pruned++;

    markers.clear();
    markers = tmp;

    cerr << "  Pruned " << pruned << " markers that have no alignments; " << markers.size() << " markers remain.\n";

    return 0;
}

int
parse_lg_definitions(string path, map<string, Marker> &markers)
{
    char *line       = new char[max_len];
    int   size       = max_len;
    int   fh_status  = 1;
    int   line_num   = 1;

    set<string>    lgs;
    vector<string> parts;

    ifstream fh;
    gzFile   gzfh;
    bool     compressed = false;

    uint pos = path.find_last_of(".");
    
    if (path.substr(pos + 1) == "gz")
        compressed = true;

    if (compressed) {
        gzfh = gzopen(path.c_str(), "rb");
        if (!gzfh) {
            cerr << "Unable to open gzipped file '" << path << "': " << strerror(errno) << ".\n";
            return 0;
        }
    } else {
        fh.open(path.c_str(), ifstream::in);
        if (fh.fail()) {
            cerr << " Unable to open '" << path << "'\n";
            return 0;
        }
    }
    cerr << "  Parsing '" << path << "'.\n";

    while (fh_status) {
        if (compressed)
            fh_status = read_gzip_line(gzfh, &line, &size);
        else
            fh_status = read_line(fh, &line, &size);

        if (!fh_status && strlen(line) == 0)
	    continue;

	if (is_comment(line)) continue;

	parse_tsv(line, parts);

        if (parts.size() != num_lg_fields) {
            cerr << "Error parsing " << path << " at line: " << line_num << ". (" << parts.size() << " fields).\n";
            return 0;
        }

        //
        // Determine how many decimal places are used to specify the marker position in centiMorgans.
        //
        const char *p = parts[2].c_str();
        uint w = 0;
        for (; *p != '.' && *p != '\0'; p++);
        if (*p == '.')
            w = strlen(p + 1);
        if (w > fieldw)
            fieldw = w;
        
        string lg = parts[0];
        string id = parts[1].c_str();
        double cm = is_double(parts[2].c_str());


	if (markers.count(id) > 0)
	    cerr << "Marker " << id << " id defined more than once in the linkage group file; skipping additional definitions.\n";
	else
	    markers[id] = Marker(id, lg, cm);

	lgs.insert(lg);

        line_num++;
    }

    if (compressed)
        gzclose(gzfh);
    else
        fh.close();

    delete [] line;

    cerr << "    Loaded " << markers.size() << " markers on " << lgs.size() << " linkage groups.\n";

    return 1;
}

int
parse_sam_alignments(string path, map<string, Marker> &markers)
{
    Input *fh = NULL;
    Seq c;

    size_t pos = path.find_last_of(".");

    if (path.substr(pos + 1) == "bam")
        fh = new Bam(path.c_str());
    else
        fh = new Sam(path.c_str());

    cerr << "  Parsing '" << path << "'.\n";

    bool scaf_wl = scaffold_wl.size() > 0 ? true : false;

    int primary_kept    = 0;
    int primary_qual    = 0;
    int primary_clipped = 0;
    int secondary_kept  = 0;
    int secondary_disc  = 0;
    int supplementary   = 0;
    int unmapped        = 0;
    int in_genetic_map  = 0;
    int total_alns      = 0;

    while ((fh->next_seq(c)) != 0) {
        bool keep = false;
        total_alns++;

        switch (c.aln_type) {
        case AlnT::null:
            unmapped++;
            break;
        case AlnT::primary:
            if (c.map_qual < min_mapping_qual)
                primary_qual++;
            else if (c.pct_clipped > max_clipped)
                primary_clipped++;
            else {
                keep = true;
                primary_kept++;
            }
            break;
        case AlnT::secondary:
            if (keep_sec_alns && c.pct_clipped <= max_clipped) {
                secondary_kept++;
                keep = true;
            } else
                secondary_disc++;
            break;
        case AlnT::supplementary:
            supplementary++;
        default:
            break;
        }

        if (!keep) continue;

	if (scaf_wl && scaffold_wl.count(c.loc.chr) == 0) continue;
        
	//
	// Only consider markers defined in the genetic map.
	//
	if (markers.count(c.id) > 0) {
            in_genetic_map++;

            Marker &m = markers[c.id];

	    m.scaffold = c.loc.chr;
	    m.bp       = c.loc.bp;
	    m.cigar    = c.loc.cigar;
	    m.strand   = c.loc.strand;
            m.aln_cnt++;
	}        
    }

    int n_primary = primary_kept+primary_qual+primary_clipped;
    cerr << "  Done reading alignment records:\n"
         << "    Kept " << primary_kept << " primary alignments\n"
         << "    Skipped " << setprecision(2) <<  primary_qual << " (" << (double) primary_qual / n_primary * 100
         << "%) primary alignments with insufficient mapping qualities\n"
         << "    Skipped " << primary_clipped << " (" << (double) primary_clipped / n_primary * 100
         << "%) excessively soft-clipped primary alignments\n"
         << "    Skipped " << secondary_disc << " secondary alignments\n"
         << "    Skipped " << supplementary << " supplementary alignments\n"
         << "    Skipped " << unmapped << " unmapped reads\n";

    if (keep_sec_alns)
        cerr << "  Kept " << secondary_kept << " secondary alignments\n";

    delete fh;
    
    cerr << "  Examined " << total_alns << " alignments; loaded alignment information for " << in_genetic_map << " markers also present in the genetic map.\n";

    return 1;
}

int
parse_agp_file(string path, map<string, vector<ScafComp>> &scaffolds)
{
    char *line      = new char[max_len];
    int   size      = max_len;
    int   fh_status = 1;
    int   line_num  = 1;

    vector<string> parts;

    ifstream fh;
    gzFile   gzfh;
    bool compressed = false;

    uint pos = path.find_last_of(".");
    
    if (path.substr(pos + 1) == "gz")
        compressed = true;

    if (compressed) {
        gzfh = gzopen(path.c_str(), "rb");
        if (!gzfh) {
            cerr << "Unable to open gzipped file '" << path << "': " << strerror(errno) << ".\n";
            return 0;
        }
    } else {
        fh.open(path.c_str(), ifstream::in);
        if (fh.fail()) {
            cerr << " Unable to open '" << path << "'\n";
            return 0;
        }
    }
    cerr << "  Parsing '" << path << "'.\n";

    bool scaf_wl = scaffold_wl.size() > 0 ? true : false;

    while (fh_status) {
        if (compressed)
            fh_status = read_gzip_line(gzfh, &line, &size);
        else
            fh_status = read_line(fh, &line, &size);

        if (!fh_status && strlen(line) == 0)
	    continue;

	if (is_comment(line)) continue;

	parse_tsv(line, parts);

        if (allpaths_agp) {
	    if (parts.size() != num_allpaths_agp_fields) {
		cerr << "Error parsing Allpaths AGP file " << path << " at line: " << line_num << ". (" << parts.size() << " fields).\n";
		return 0;
	    }
	} else {
	    if (parts.size() != num_agp_fields) {
		cerr << "Error parsing " << path << " at line: " << line_num << ". (" << parts.size() << " fields).\n";
		return 0;
	    }
        }

        line_num++;

        string    scaffold = parts[0];
        uint      start    = is_integer(parts[1].c_str());
        uint      end      = is_integer(parts[2].c_str());
        uint      part_num = is_integer(parts[3].c_str());
	comp_type type     = parts[4] == "N" ? gap : contig;
	uint      len;
	uint      comp_beg, comp_end;
	ori_type  ori;
	bool      linkage;

	if (scaf_wl && scaffold_wl.count(scaffold) == 0) continue;

	if (type == contig) {
	    if (allpaths_agp) {
		comp_beg = is_integer(parts[5].c_str());
		comp_end = is_integer(parts[6].c_str());
		len = comp_end - comp_beg + 1;
		ori = parts[7] == "+" ? ori_forward : ori_reverse;
		scaffolds[scaffold].push_back(ScafComp(scaffold, type, parts[4], start, end, part_num, len, comp_beg, comp_end, ori));
	    } else {
		comp_beg = is_integer(parts[6].c_str());
		comp_end = is_integer(parts[7].c_str());
		len = comp_end - comp_beg + 1;
		ori = parts[8] == "+" ? ori_forward : ori_reverse;
		scaffolds[scaffold].push_back(ScafComp(scaffold, type, parts[5], start, end, part_num, len, comp_beg, comp_end, ori));
	    }
	} else {
	    len = is_integer(parts[5].c_str());
	    linkage = parts[7] == "yes" ? true : false;
	    scaffolds[scaffold].push_back(ScafComp(scaffold, type, start, end, part_num, len, parts[6], linkage));
	}
    }

    if (compressed)
        gzclose(gzfh);
    else
        fh.close();

    delete [] line;

    cerr << "    Loaded assembly components for " << scaffolds.size() << " scaffolds.\n";

    return 1;
}

bool 
compare_marker_bp(Marker a, Marker b)
{
    return (a.bp < b.bp);
}

bool 
compare_marker_bp_reverse(Marker a, Marker b)
{
    return (b.bp < a.bp);
}

bool 
compare_node_cM(Node a, Node b)
{
    return (a.cM < b.cM);
}

bool
compare_marker_cM_bp(Marker a, Marker b)
{
    if (a.cM == b.cM)
	return a.bp < b.bp;

    return a.cM < b.cM;
}

bool
compare_marker_cM_bp_reverse(Marker a, Marker b)
{
    if (a.cM == b.cM)
	return b.bp < a.bp;
    
    return a.cM < b.cM;
}

bool
compare_problem_scaffolds(ProbScaf a, ProbScaf b)
{
    if (a.splits == b.splits)
	return b.conflict_markers < a.conflict_markers;

    return b.splits < a.splits;
}

int 
load_marker_list(string path, set<string> &list) {
    char     line[id_len];
    ifstream fh(path.c_str(), ifstream::in);

    if (fh.fail()) {
        cerr << "Error opening white/black list file '" << path << "'.\n";
	exit(1);
    }

    char *p;

    while (fh.good()) {
	fh.getline(line, id_len);

	if (strlen(line) == 0) continue;

	//
	// Skip commented lines.
	//
	for (p = line; isspace(*p) && *p != '\0'; p++);
	if (*p == '#') continue;

	list.insert(line);
    }

    fh.close();

    if (list.size() == 0) {
 	cerr << "Unable to load any markers/scaffolds from '" << path << "'\n";
	exit(1);
    }

    return 0;
}

int 
parse_command_line(int argc, char* argv[]) 
{
    int c;

    while (1) {
	static struct option long_options[] = {
	    {"help",              no_argument,     NULL, 'h'},
            {"version",           no_argument,     NULL, 'v'},
            {"verbose",           no_argument,     NULL, 'V'},
            {"allpaths_agp",      no_argument,     NULL, 'P'},
            {"disable_splitting", no_argument,     NULL, 'I'},
            {"fasta",           required_argument, NULL, 'F'},
	    {"min_markers",     required_argument, NULL, 'M'},
            {"join_gap_size",   required_argument, NULL, 'G'},
	    {"scaffold_prefix", required_argument, NULL, 'C'},
	    {"data_version",    required_argument, NULL, 'A'},
	    {"lg_path",         required_argument, NULL, 'p'},
	    {"sam_path",        required_argument, NULL, 's'},
	    {"agp_path",        required_argument, NULL, 'a'},
	    {"out_path",        required_argument, NULL, 'o'},
	    {"scaffold_wl",     required_argument, NULL, 'S'},
            {"description",     required_argument, NULL, 'D'},
	    {0, 0, 0, 0}
	};

	// getopt_long stores the option index here.
	int option_index = 0;

	c = getopt_long(argc, argv, "hPIVvo:p:s:a:A:C:D:G:M:S:", long_options, &option_index);

	// Detect the end of the options.
	if (c == -1)
	    break;

	switch (c) {
	case 'h':
	    help();
	    break;
	case 'M':
	    min_node_markers = is_integer(optarg);
	    break;
        case 'G':
	    join_gap_size = is_integer(optarg);
	    break;
	case 'C':
	    scaffold_prefix = optarg;
	    break;
	case 'A':
	    data_version = optarg;
	    break;
	case 'p':
	    lg_path = optarg;
	    break;
	case 's':
	    sam_path = optarg;
	    break;
	case 'a':
	    agp_path = optarg;
	    break;
        case 'D':
	    description = optarg;
	    break;
        case 'F':
            parse_fasta = true;
	    fasta_path  = optarg;
	    break;
	case 'o':
	    out_path = optarg;
	    break;
	case 'S':
	    scaffold_wl_path = optarg;
	    break;
        case 'P':
            allpaths_agp = true;
            break;
        case 'I':
            split_scaffolds = false;
            break;
        case 'V':
            verbose = true;
            break;
        case 'v':
            version();
            break;
	case '?':
	    // getopt_long already printed an error message.
	    help();
	    break;
	default:
	    help();
	    abort();
	}
    }

    if (lg_path.length() == 0) {
	cerr << "You must specify a path to the linkage group definition file.\n";
	help();
    }

    if (sam_path.length() == 0) {
	cerr << "You must specify a path to a SAM or BAM file containing marker alignments.\n";
	help();
    }

    if (agp_path.length() == 0) {
	cerr << "You must specify a path to the AGP assembly definition file.\n";
	help();
    }

    if (out_path.length() == 0) {
	cerr << "You must specify a path to the directory to write scaffold documentation output.\n";
	help();
    }

    if (out_path.at(out_path.length() - 1) != '/') 
	out_path += "/";

    if (parse_fasta) {
        struct stat buffer;
	if (stat(fasta_path.c_str(), &buffer) != 0) {
            cerr << "Unable to open FASTA file: '" << fasta_path << "'.\n";
            help();
	}        
    }
    
    return 0;
}

void version() {
    std::cerr << "chromonomer " << VERSION << "\n\n";

    exit(0);
}

void help() {
    std::cerr << "chromonomer " << VERSION << "\n"
              << "chromonomer -p lg_path -s sam_path -a agp_path -o out_path [-v] [-h]" << "\n"
	      << "  p: path to file containing the linkage group definitions.\n"
	      << "  s: path to SAM or BAM formatted alignments linking map markers to genome contigs.\n"
	      << "  a: path to the AGP formatted file defining assembly gaps.\n"
	      << "  o: path to write chromonomer output.\n"
	      << "  h: display this help message.\n"
	      << "  d: turn on debug output.\n"
              << "  --fasta: <path>: supply a scaffold-based FASTA that will be translated according to the integrated assembly.\n"
	      << "  --verbose: turn on detailed console output for each scaffold.\n"
	      << "  --scaffold_prefix <string>: text string to use as a common naming prefix when creating new scaffolds.\n"
              << "  --description <string>: supply a description of this chromonomer execution that will be written into the output log.\n"
	      << "  --min_markers <num>: minimum number of markers required to anchor a scaffold on a particular\n"
	      << "                       map node. Will not remove scaffolds only present on one node (default 2).\n"
              << "  --join_gap_size <num>: number of Ns to insert in between scaffold joins (default: 100).\n"
	      << "  --allpaths_agp: The AllPaths-LG assembler produces a non-standard AGP file. This flag will change\n"
	      << "                  the AGP parser to accommodate it.\n"
              << "  --disable_splitting: do not split scaffolds. When markers indicate a conflict in where to place a scffold, do\n"
              << "                       not split the scaffold, instead place it in the integration where the majority of markers are.\n"
	      << "  Filtering Options:\n"
	      << "    --scaffold_wl <path>: only process scaffolds in this file.\n"
	      << "  HTML/PHP Options:\n"
	      << "    --data_version <string>: used to create versioned PHP files so multiple runs can be\n"
	      << "                             executed and compared in the web interface.\n"
	      << "    --html_prefix <string>: any text that should be prepended onto links in the PHP output\n"
	      << "                            to accommodate where they are located in the web server document root.\n";

    exit(0);
}
