// -*-mode:c++; c-style:k&r; c-basic-offset:4;-*-
//
// Copyright 2016-2018, Julian Catchen <jcatchen@illinois.edu>
//
// This file is part of Chromonomer.
//
// Chromonomer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Chromonomer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Chromonomer.  If not, see <http://www.gnu.org/licenses/>.
//

#include "splits.h"

int
split_scaffold(string scaffold, 
	       map<int, vector<Marker>> &segments,
	       ori_type ori,
	       map<string, vector<ScafComp>> &agp, 
	       map<string, Marker> &markers,
               map<string, Marker> &conflict_markers,
	       int &scaffold_id, ofstream &doc_fh)
{
    if (agp.count(scaffold) == 0 || agp[scaffold].size() == 0) {
	cerr << "Unable to process " << scaffold << ", can't find AGP reference.\n";
	return 0;
    }

    char doc_str[max_len];
    vector<ScafComp> &s = agp[scaffold];
    uint size = s.size() - 1;

    if (split_scaffolds == false) {
        reduce_to_majority_segment(scaffold, segments, s, conflict_markers, doc_fh);
        return 0;
    }

    doc_fh << "Splitting scaffold " << scaffold << ", " << size + 1 << " components\n";

    if (scaffold == "scaffold_215")
        cerr << "Here\n";
    
    hang_annotations(segments, ori, s);

    print_scaffold_map(s, doc_fh, true);

    if (verbose)
        cerr << "    Checking the integrity of marker placements on scaffold " << scaffold << "'...\n";
    doc_fh << "Checking the integrity of marker placements on scaffold " << scaffold << "'...\n";

    vector<ScafAnn> marker_order;
    uint pruned_cnt = 0;
    bool pruned     = false;
    uint next, cur, mo_size;

    do {
        pruned = false;
        //
        // Serialize the start/end annotations to make sure they occur in the proper order.
        //
        serialize_annotations(s, marker_order);

        //
        // Check the integrity of the marker placements, make sure they do not overlap.
        //
        // All markers should be in order: start->end/start->end/start->end. If they are not
        // then two segments overlap; that is, there is not a gap to cleanly separate the two segments.
        //
        // In this case, we want to prune out the smaller (minority) of the two segments
        // and continue with the splitting operation.
        //
        cur     = 0;
        next    = 1;
        mo_size = marker_order.size();

        while (next < mo_size && cur < mo_size) {
            if (marker_order[next].type == marker_order[cur].type) {
                if (verbose)
                    cerr << "      Segments " << marker_order[cur].segment << " and " << marker_order[next].segment << " of scaffold '" << scaffold << "' overlap, ";
                doc_fh << "  Segments " << marker_order[cur].segment << " and " << marker_order[next].segment << " of scaffold '" << scaffold << "' overlap, ";

                prune_segment(segments, marker_order[cur], marker_order[next], s, conflict_markers, doc_fh);

                // print_scaffold_map(s, doc_fh, true);

                pruned_cnt++;
                pruned = true;
                break;
            }
            cur++;
            next++;
        }
    } while (pruned == true && segments.size() > 1);

    if (pruned_cnt == 0) {
        if (verbose)
            cerr << "    Done.\n";
        doc_fh << "Done.\n";
    } else {
        if (verbose)
            cerr << "    Removed " << pruned_cnt << " segments.\n";
        doc_fh << "Removed " << pruned_cnt << " segments.\n";
    }

    if (segments.size() == 1) {
        //
        // If only one segment remains, do not try to split the scaffold any further.
        //
        if (verbose)
            cerr << "    Only one segment remains, scaffold will not be split.\n";
        doc_fh << "Only one segment remains, scaffold will not be split.\n";

        return 0;
    } else {
        doc_fh << "\n";
    }

    //
    // If the end marker of one segment occurs before the starting marker of the next segment,
    // then interpolate where to break the scaffold by finding the largest gap and moving the
    // end/start markers to both point at this gap.
    //
    bool found = false;
    cur  = 0;
    next = 1;
    while (next < size && cur < size) {

	while (cur < size) {
	    while (s[cur].annotations.size() == 0 && cur < size) { cur++; }
	    if (cur >= size) break;
	    if (s[cur].annotations.size() == 1 &&
		s[cur].annotations[0].type == seg_end) break;
	    cur++;
	}

    	next = cur + 1;

	while (next < size) {
	    while (s[next].annotations.size() == 0 && next < size) { next++; }
	    if (next >= size) break;
	    if (s[next].annotations.size() == 1 &&
		s[next].annotations[0].type == seg_beg) break;
	    next++;
	}

	if (cur >= size || next >= size)
	    break;

	//
	// Interpolate between these breakpoints -- find the largest gap and set the breakpoint there.
	//
	found = true;

	snprintf(doc_str, max_len,
		 "Interpolating between markers: Segment %d/%s (%03d) -> Segment %d/%s (%03d)\n",
		 s[cur].annotations[0].segment,  (s[cur].annotations[0].type  == seg_beg ? "begin" : "end"), cur,
		 s[next].annotations[0].segment, (s[next].annotations[0].type == seg_beg ? "begin" : "end"), next);
	doc_fh << doc_str;

	uint max_gap = s[cur].len;
	uint max_pos = cur;
	for (uint i = cur; i <= next; i++) {
	    if (s[i].type == gap && s[i].len > max_gap) {
		max_gap = s[i].len;
		max_pos = i;
	    }
	}

	doc_fh << "  Maximum gap at position " << max_pos << " (" << max_gap << "bp)\n";

	//
	// Move the markers to the maximal gap position.
	//
 	if (cur != max_pos) {
	    for (uint i = 0; i < s[cur].annotations.size(); i++)
		s[max_pos].annotations.push_back(s[cur].annotations[i]);
	    s[cur].annotations.clear();
	}
	if (next != max_pos) {
	    for (uint i = 0; i < s[next].annotations.size(); i++)
		s[max_pos].annotations.push_back(s[next].annotations[i]);
	    s[next].annotations.clear();
	}

	cur = next + 1;
    }

    if (found) {
	doc_fh << "\n";
	print_scaffold_map(s, doc_fh, true);
    }

    //
    // Traverse the scaffold and split it into new contigs.
    //
    stack<uint>    segment_stack;
    vector<string> new_scaffolds;
    uint start = 0;
    uint end   = 0;
    uint last  = 0;

    char new_scaffold[id_len];

    //
    // Find the last marker on the scaffold, record it so we can adjust the end point of the last scaffold later on.
    //
    while (end <= size) {
	if (s[end].annotations.size() > 0) {
	    last = end;
	}
	end++;
    }
    end = 0;

    //
    // Find the first starting marker, add the corresponding segment to the stack.
    //
    while (end < size && s[end].annotations.size() == 0) {
	end++;
    }
    segment_stack.push(s[end].annotations[0].segment);
    // cerr << "Pushing segment " <<  s[end].annotations[0].first << " onto the stack.\n";
    end++;

    //
    // Loop until the stack is empty and we have reached the end of the scaffold.
    //
    while (segment_stack.size() > 0) {
	//
	// Find the next boundary.
	//
	while (end < size && s[end].annotations.size() == 0) {
	    end++;
	}
	if (end == last) end = size;

	// cerr << "Setting start to " << start << "; end to " << end << "\n";

	//
	// Create a new scaffold.
	//
	snprintf(new_scaffold, id_len, "%s%05d.0", scaffold_prefix.c_str(), scaffold_id);
	new_scaffolds.push_back(new_scaffold);
	scaffold_id++;

	//
	// Adjust the endpoint to not include the ending gap. Calculate the basepair adjustment.
	//
	int adj_end = end == size ? end : end - 1;
	int adj_bp  = s[start].start;

	//
	// Adjust markers from the old segment of the scaffold.
	//
	vector<Marker> &marker_list = segments[segment_stack.top()];
	for (uint i = 0; i < marker_list.size(); i++) {
	    //
	    // We go through two rounds of scaffold splitting, first for inter-LG splits, then 
	    // intra-LG splits, so, we want to keep the original scaffold/bp recorded for the 
	    // scaffold map.
	    //
	    markers[marker_list[i].id].old.push_back(make_pair(marker_list[i].scaffold, marker_list[i].bp));

	    markers[marker_list[i].id].scaffold = new_scaffold;
	    markers[marker_list[i].id].bp       = marker_list[i].bp - adj_bp + 1;

	    // doc_fh << "Adjusting marker in segment " << segment_stack.top() << "\n"
	    // 	 << "  Starting region: " << s[start].start << "; Ending region: " << s[adj_end].end << "\n"
	    // 	 << "    Adjusting position of marker " << marker_list[i].id << ": old scaffold: "
	    // 	 << markers[marker_list[i].id].old.back().first << ", old bp: " << markers[marker_list[i].id].old.back().second
	    // 	 << "; new scaffold: " << markers[marker_list[i].id].scaffold << ", new bp: " << markers[marker_list[i].id].bp << "\n";
	}

	//
	// Extract out the scaffold components to create the new scaffold. Remove 
	// marker annotations from the newly created scaffold.
	//
	for (int i = start; i <= adj_end; i++) {
	    //
	    // Rename and renumber.
	    //
	    s[i].relocate(new_scaffold, s[i].start - adj_bp + 1, s[i].end   - adj_bp + 1);
	    s[i].annotations.clear();
	    agp[new_scaffold].push_back(s[i]);
	}

	// cerr << "  Adjusting the end point to " << adj_end << "\n";
	doc_fh << "Creating scaffold '" << new_scaffold << "' with " << agp[new_scaffold].size() << " elements [" << start << " -> " << adj_end << "].\n";

	//
	// Adjust the stack.
	//   If the final marker in this segment was an 'end' marker for a linkage group, 
	//   pop the linkage group off the stack. If it was a 'start' marker, push the new linkage group on.
	//
	if (end == size) {
	    segment_stack.pop();
	    // cerr << "Poping off the stack, " << segment_stack.size() << " elements remain.\n";
	} else {  
	    for (cur = 0; cur < s[end].annotations.size(); cur++)
		if (s[end].annotations[cur].type == seg_end) {
		    segment_stack.pop();
		    // cerr << "Poping segment " << s[end].annotations[cur].first << off the stack, " << segment_stack.size() << " elements remain.\n";
		} 
	    for (cur = 0; cur < s[end].annotations.size(); cur++)
		if (s[end].annotations[cur].type == seg_beg) {
		    segment_stack.push(s[end].annotations[cur].segment);
		    // cerr << "Pushing segment " << s[end].annotations[cur].first << " onto the stack, " << segment_stack.size() << " elements remain.\n";
		}
	}

	//
	// Reset the start/end pointers.
	//
	start = end + 1;
	end   = start;
	// cerr << "Reseting start to " << start << ", end to " << end << ".\n";
    }

    stringstream sstr;

    //
    // Record a map describing how the old scaffold maps to the new scaffolds.
    //
    for (uint i = 0; i < new_scaffolds.size(); i++)
	sstr << new_scaffolds[i] << ", ";

    if (verbose)
	cerr << "    Scaffold '" << scaffold << "' successfully split into scaffolds " << sstr.str().substr(0, sstr.str().length() - 2) << "\n";
    doc_fh << "\nScaffold '" << scaffold << "' successfully split into scaffolds " << sstr.str().substr(0, sstr.str().length() - 2) << "\n";

    return 1;
}

int
hang_annotations(map<int, vector<Marker>> &segments,
                 ori_type ori,
                 vector<ScafComp> &s)
{
    uint size = s.size() - 1;
    map<int, vector<Marker>>::iterator it;

    //
    // For each segment on this scaffold, attach the begining markers to the nearest gap sequence.
    //
    for (it = segments.begin(); it != segments.end(); it++) {
	vector<Marker> &mrks = it->second;
	uint segment = it->first;
	uint cnt     = mrks.size();
	uint cur     = ori == ori_reverse ? size : 0;

	if (ori == ori_reverse) {
	    //
	    // Place the marker that defines the starting position of this LG.
	    //
	    while (cur > 0 && mrks[cnt - 1].bp < s[cur].end)
		cur--;

	    if (cur < size)
	        cur++;
	    //
	    // Move it to the nearest gap.
	    //
	    while (cur > 0 && s[cur].type != gap)
		cur--;

	} else {
	    //
	    // Place the marker that defines the starting position of this LG.
	    //
	    while (cur <= size && s[cur].end < mrks[0].bp)
		cur++;
	    //
	    // Move it to the nearest gap.
	    //
	    while (cur > 0 && s[cur].type != gap)
		cur--;
	}
	s[cur].annotations.push_back(ScafAnn(cur, segment, seg_beg));
    }

    //
    // For each segment on this scaffold, attach the ending markers to the nearest gap sequence.
    //
    for (it = segments.begin(); it != segments.end(); it++) {
	vector<Marker> &mrks = it->second;
	uint segment = it->first;
	uint cnt     = mrks.size();
	uint cur     = ori == ori_reverse ? 0 : size;

	if (ori == ori_reverse) {
	    //
	    // Place the marker that defines the ending position of this segment.
	    //
	    while (cur <= size && s[cur].end < mrks[0].bp)
		cur++;
	    //
	    // Move it to the nearest gap.
	    //
	    while (cur < size && s[cur].type != gap)
		cur++;

	} else {
	    //
	    // Place the marker that defines the ending position of this segment.
	    //
	    while (cur > 0 && mrks[cnt - 1].bp < s[cur].end)
		cur--;

	    if (cur < size)
	    	cur++;
	    //
	    // Move it to the nearest gap.
	    //
	    while (cur < size && s[cur].type != gap)
		cur++;
	}
	s[cur].annotations.push_back(ScafAnn(cur, segment, seg_end));
    }

    return 0;
}

int
serialize_annotations(vector<ScafComp> &s,
                      vector<ScafAnn> &marker_order)
{
    uint ann_cnt;
    uint size = s.size() - 1;

    marker_order.clear();

    for (uint cur = 0; cur <= size; cur++) {
	ann_cnt = s[cur].annotations.size();

        switch(ann_cnt) {
        case 0:
            break;
        case 1:
            marker_order.push_back(s[cur].annotations[0]);
            break;
        case 2:
            if (s[cur].annotations[0].type == seg_end)
                marker_order.push_back(s[cur].annotations[0]);
            if (s[cur].annotations[1].type == seg_end)
                marker_order.push_back(s[cur].annotations[1]);

            if (s[cur].annotations[0].type == seg_beg)
                marker_order.push_back(s[cur].annotations[0]);
            if (s[cur].annotations[1].type == seg_beg)
                marker_order.push_back(s[cur].annotations[1]);
            break;
        default:
            for (uint i = 0; i < ann_cnt; i++)
                if (s[cur].annotations[i].type == seg_beg)
                    marker_order.push_back(s[cur].annotations[i]);
            for (uint i = 0; i < ann_cnt; i++)
                if (s[cur].annotations[i].type == seg_end)
                    marker_order.push_back(s[cur].annotations[i]);
            break;
        }
    }

    return 0;
}

int
prune_segment(map<int, vector<Marker>> &segments, ScafAnn seg_1, ScafAnn seg_2,
              vector<ScafComp> &agp_scaf,
              map<string, Marker> &conflict_markers,
              ofstream &doc_fh)
{
    char doc_str[max_len];
    int  agp_size = agp_scaf.size();

    //
    // Find the majority node in the linkage map -- the one with the most markers.
    //
    ScafAnn majority_segment = segments[seg_1.segment].size() > segments[seg_2.segment].size() ? seg_1 : seg_2;
    ScafAnn minority_segment = segments[seg_1.segment].size() > segments[seg_2.segment].size() ? seg_2 : seg_1;

    if (verbose)
        cerr << "pruning segment " << minority_segment.segment << " (" << segments[minority_segment.segment].size() << " markers), "
             << "retaining segment " << majority_segment.segment << " (" << segments[majority_segment.segment].size() << " markers).\n";
    doc_fh << "pruning segment " << minority_segment.segment << " (" << segments[minority_segment.segment].size() << " markers), "
           << "retaining segment " << majority_segment.segment << " (" << segments[majority_segment.segment].size() << " markers).\n";

    //
    // Re-attach annotations so the majority segment spans both the minority and
    // majority segment. Record split failures in the minority segment.
    //
    ScafAnn min_start, min_end, maj_start, maj_end;
    int     cur;

    if (minority_segment.type == seg_beg) {
        min_start = minority_segment;
        //
        // Find the ending annotation for the minority segment.
        //
        cur = minority_segment.agp_pos;
        do {
            vector<ScafAnn> &ann = agp_scaf[cur].annotations;

            for (uint j = 0; j < ann.size(); j++)
                if (ann[j].segment == minority_segment.segment &&
                    ann[j].type    == seg_end) {
                    min_end = ann[j];
                    cur = agp_size;
                    break;
                }
            cur++;
        } while (cur < agp_size);

    } else if (minority_segment.type == seg_end) {
        min_end = minority_segment;
        //
        // Find the starting annotation for the minority segment.
        //
        cur = minority_segment.agp_pos;
        do {
            vector<ScafAnn> &ann = agp_scaf[cur].annotations;
        
            for (uint j = 0; j < ann.size(); j++)
                if (ann[j].segment == minority_segment.segment &&
                    ann[j].type    == seg_beg) {
                    min_start = ann[j];
                    cur = -1;
                    break;
                }
            cur--;
        } while (cur >= 0);
    }

    if (majority_segment.type == seg_beg) {
        maj_start = majority_segment;
        //
        // Find the ending annotation for the majority segment.
        //
        cur = majority_segment.agp_pos;
        do {
            vector<ScafAnn> &ann = agp_scaf[cur].annotations;
        
            for (uint j = 0; j < ann.size(); j++)
                if (ann[j].segment == majority_segment.segment &&
                    ann[j].type    == seg_end) {
                    maj_end = ann[j];
                    cur = agp_size;
                    break;
                }
            cur++;
        } while (cur < agp_size);

    } else if (majority_segment.type == seg_end) {
        maj_end = majority_segment;
        //
        // Find the starting annotation for the majority segment.
        //
        cur = majority_segment.agp_pos;
        do {
            vector<ScafAnn> &ann = agp_scaf[cur].annotations;
        
            for (uint j = 0; j < ann.size(); j++)
                if (ann[j].segment == majority_segment.segment &&
                    ann[j].type    == seg_beg) {
                    maj_start = ann[j];
                    cur = -1;
                    break;
                }
            cur--;
        } while (cur >= 0);
    }

    //
    // Remove the annotations for both segments from the AGP structure.
    //
    remove_annotations(agp_scaf[min_start.agp_pos].annotations, min_start.segment);
    remove_annotations(agp_scaf[min_end.agp_pos].annotations,   min_end.segment);
    remove_annotations(agp_scaf[maj_start.agp_pos].annotations, maj_start.segment);
    remove_annotations(agp_scaf[maj_end.agp_pos].annotations,   maj_end.segment);

    //
    // Replace the annotations for the majority segment, spanning both the minority and majority segment.
    //
    if (min_start.agp_pos < maj_start.agp_pos)
        maj_start.agp_pos = min_start.agp_pos;
    agp_scaf[maj_start.agp_pos].annotations.push_back(maj_start);

    if (min_end.agp_pos > maj_end.agp_pos)
        maj_end.agp_pos = min_end.agp_pos;
    agp_scaf[maj_end.agp_pos].annotations.push_back(maj_end);

    //
    // Mark the AGP components containing the minority segment as failed.
    //
    for (uint i = min_start.agp_pos; i <= min_end.agp_pos; i++)
        agp_scaf[i].split_fail = true;

    //
    // Blacklist markers from minority segment.
    //
    vector<Marker> &segment = segments[minority_segment.segment];
    for (uint i = 0; i < segment.size(); i++) {
        snprintf(doc_str, max_len, 
                 "    Removing marker %s; marker segment: %d; majority segment: %d\n",
                 segment[i].id.c_str(), minority_segment.segment, majority_segment.segment);
        doc_fh << doc_str;
        conflict_markers[segment[i].id] = segment[i];
    }

    //
    // Remove the minority segment from the segment map.
    //
    segments.erase(minority_segment.segment);

    return 0;
}

int
reduce_to_majority_segment(string scaffold,
                           map<int, vector<Marker>> &segments,
                           vector<ScafComp> &agp_scaf,
                           map<string, Marker> &conflict_markers,
                           ofstream &doc_fh)
{
    char doc_str[max_len];

    int majority = 0;
    int segment;
    //
    // Find the majority node in the linkage map -- the one with the most markers.
    //
    for (map<int, vector<Marker> >::iterator it = segments.begin(); it != segments.end(); it++)
        if (it->second.size() > (uint) majority) {
            segment  = it->first;
            majority = it->second.size();
        }

    if (verbose)
        cerr << "    Scaffold '" << scaffold << "' will remain on linkage group '" << segments[segment][0].lg
             << "', supported by " << segments[segment].size() << " markers.\n";
    doc_fh << "Scaffold '" << scaffold << "' will remain on linkage group '" << segments[segment][0].lg
           << "', supported by " << segments[segment].size() << " markers.\n";
    
    //
    // Blacklist markers
    //
    for (map<int, vector<Marker> >::iterator it = segments.begin(); it != segments.end(); it++)
        if (it->first != segment) {
            for (uint i = 0; i < it->second.size(); i++) {
                snprintf(doc_str, max_len,
                         "  Removing marker %s; marker segment: %d; majority segment: %d\n",
                         it->second[i].id.c_str(), it->first, segment);
                doc_fh << doc_str;
                conflict_markers[it->second[i].id] = it->second[i];
            }
        }

    //
    // Remove annotations.
    //
    for (uint j = 0; j < agp_scaf.size(); j++) {
        agp_scaf[j].split_fail = true;
        agp_scaf[j].annotations.clear();
    }

    //
    // Remove the non-majority segments.
    //
    vector<int> other_segments;
    for (map<int, vector<Marker> >::iterator it = segments.begin(); it != segments.end(); it++)
        if (it->first != segment)
            other_segments.push_back(it->first);
    for (uint i = 0; i < other_segments.size(); i++)
        segments.erase(other_segments[i]);

    return 0;
}

int
remove_annotations(vector<ScafAnn> &ann, uint segment)
{
    vector<ScafAnn> tmp;

    for (uint j = 0; j < ann.size(); j++) {
        if (ann[j].segment != segment)
            tmp.push_back(ann[j]);
    }
    ann.clear();
    for (uint j = 0; j < tmp.size(); j++)
        ann.push_back(tmp[j]);

    return 0;
}
