#!/usr/bin/perl
#
# Copyright 2014-2016, Julian Catchen <jcatchen@illinois.edu>
#
# This file is part of Chromonomer.
#
# Chromonomer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Chromonomer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Chromonomer.  If not, see <http://www.gnu.org/licenses/>.
#

use strict;
use Bio::SeqIO;

use constant true    => 1;
use constant false   => 0;
use constant unknown => -1;
use constant forw => 0;
use constant reve => 1;

my $debug      = 0;
my $in_path    = "";
my $agp_path   = "";
my $map_path   = "";
my $out_path   = "";
my $fasta_path = "";

my (%lgs, %map, %agp, %fasta, %mapped_scaffolds, %rev_map, %components);

parse_command_line();

load_map_file($map_path, \%map, \%rev_map, \%components);

load_lgs($in_path, \%lgs, \%mapped_scaffolds);

load_agp_file($agp_path, \%agp, \%mapped_scaffolds, \%rev_map);

load_fasta($fasta_path, \%fasta, \%mapped_scaffolds, \%rev_map, $out_path);

assemble_fasta(\%lgs, \%agp, \%fasta, \%map, \%components, $out_path);

sub assemble_fasta {
    my ($lgs, $agp, $fasta, $map, $comps, $out_path) = @_;

    my ($agp_out, $agp_fh, $lg, $num_scaf, $scaf, $cmp, $seq, $path, $seqio, $lg_seq, $cur_bp, $end_bp, $tot_len, $cnt);

    $agp_out = $out_path . "xma_washu_4.4.2-jhp_1.0.flat.agp";
    open($agp_fh, ">$agp_out") or die("Unable to open AGP output file: $agp_out, $!\n");

    $path    = $out_path . "xma_washu_4.4.2-jhp_1.0.fa";
    $seqio   = Bio::SeqIO->new(-file => ">$path", -format => "fasta");

    my $genome_len = 0;

    foreach $lg (keys %{$lgs}) {
	print STDERR "Assembling LG$lg\n";

	$lg_seq  = "";
	$cur_bp  = 1;
	$tot_len = 0;
	$cnt     = 1;

	$num_scaf = scalar(@{$lgs->{$lg}});

	for (my $i = 0; $i < $num_scaf; $i++) {
	    $scaf = $lgs->{$lg}->[$i];

	    print STDERR "  $scaf->{'scaffold'}\n";

	    if (defined($comps->{$scaf->{'scaffold'}})) {
		#
		# Grab the original components of this scaffold and determine what subset of the sequence
		# we need to pull out of the old scaffold sequence.
		#
		my @parts;
		foreach $cmp (@{$agp->{$map->{$scaf->{'scaffold'}}}}) {
		    push(@parts, $cmp) if (defined($comps->{$scaf->{'scaffold'}}->{$cmp->{'part'}}));
		}
		@parts = sort {$a->{'start'} <=> $b->{'start'}} @parts;

		# foreach my $p (@parts) {
		#     print STDERR "      ", $p->{'scaf'}, ": ", $p->{'start'}, " - ", $p->{'end'}, "bp; Type: ", $p->{'type'}, "\n";
		# }

		die ("Unable to find FASTA sequence of " . $scaf->{'scaffold'}) if (!defined($fasta->{$map->{$scaf->{'scaffold'}}}));

		$seq      = $fasta->{$map->{$scaf->{'scaffold'}}}->subseq($parts[0]->{'start'}, $parts[-1]->{'end'});
		$end_bp   = $cur_bp + length($seq) - 1;
		$tot_len += length($seq);

		print $agp_fh 
		    "LG", $lg, "\t",
		    $cur_bp, "\t",
		    $end_bp, "\t",
		    $cnt, "\t",
		    "G", "\t",
		    $parts[0]->{'scaf'}, "\t",
		    $parts[0]->{'start'}, "\t",
		    $parts[-1]->{'end'}, "\t",
		    ($scaf->{'ori'} == reve ? "-" : "+"), "\n";

		print STDERR "    Subsequence taken from ", $parts[0]->{'scaf'}, ": ", $parts[0]->{'start'}, " - ", $parts[-1]->{'end'}, "bp\n";

	    } else {
		#
		# Scaffold was not broken, we will output the whole sequence.
		#
		$seq      = $fasta->{$scaf->{'scaffold'}}->seq();
		$end_bp   = $cur_bp + length($seq) - 1;
		$tot_len += length($seq);

		print $agp_fh 
		    "LG", $lg, "\t",
		    $cur_bp, "\t",
		    $end_bp, "\t",
		    $cnt, "\t",
		    "G", "\t",
		    $scaf->{'scaffold'}, "\t",
		    "1", "\t",
		    length($seq), "\t",
		    ($scaf->{'ori'} == reve ? "-" : "+"), "\n";
	    }

	    if ($scaf->{'ori'} == reve) {
		$seq = reverse_complement($seq);
	    }
	    $lg_seq .= $seq;
	    $cur_bp  = $end_bp + 1;
	    $cnt++;

	    if ($i < ($num_scaf - 1)) {
		#
		# Add a 100bp length of Ns, representing a gap of unknown size.
		#
		$seq      = "N" x 100;
		$end_bp   = $cur_bp + length($seq) - 1;
		$tot_len += length($seq);

		print $agp_fh 
		    "LG", $lg, "\t",
		    $cur_bp, "\t",
		    $end_bp, "\t",
		    $cnt, "\t",
		    "U", "\t",
		    "100", "\t",
		    "scaffold", "\t",
		    "yes", "\t",
		    "map", "\n";

		$lg_seq .= $seq;
		$cur_bp  = $end_bp + 1;
		$cnt++;
	    }
	}

	print STDERR "Calculated length: $tot_len, observed length: ", length($lg_seq), "\n";
	$lg_seq = Bio::Seq->new(-id => "group" . $lg, -seq => $lg_seq);
	$seqio->write_seq($lg_seq);
	$genome_len += $tot_len;
    }

    $seqio->close();
    close($agp_fh);

    print STDERR "Total genome length: $genome_len\n";
}

sub reverse_complement {
    my ($s) = @_;

    my $base;
    my $seq = "";
    my @dna = split(//, $s);

    while (scalar(@dna) > 0) {
	$base = pop(@dna);

	if ($base eq "A" || $base eq "a") {
	    $base = "T";
	} elsif ($base eq "C" || $base eq "c") {
	    $base = "G";
	} elsif ($base eq "G" || $base eq "g") {
	    $base = "C";
	} elsif ($base eq "T" || $base eq "t") {
	    $base = "A";
	}

	$seq .= $base;
    }

    return $seq;
}

sub load_fasta {
    my ($fasta_path, $fasta, $mapped, $rev_map, $out_path) = @_;

    my ($seq, $id, $unord_path);

    $unord_path = $out_path . "unplaced_scaffolds.fa";

    my $contigs   = Bio::SeqIO->new(-file => "$fasta_path", -format => "fasta");
    my $unordered = Bio::SeqIO->new(-file => ">$unord_path", -format => "fasta");
    my $uncnt     = 0;
    my $unlen     = 0;

    while ($seq = $contigs->next_seq()) {

	($id) = ($seq->id() =~ /^gi\|\d+\|gb\|(.+)\|$/);

	#
	# Check if this scaffold has been incorporated into the map, either whole or broken 
	# by the mapping algorithm. If it has not been inconrporated, immediately output it to
	# the unordered scaffolds FASTA file. Otherwise, store it to parse into fragments later.
	#
	if (defined($mapped->{$id}) || defined($rev_map->{$id})) {
	    $fasta->{$id} = $seq;
	} else {
	    $uncnt++;
	    $unlen += $seq->length();
	    $unordered->write_seq($seq);
	}
    }
    print STDERR "Loaded ", scalar(keys %{$fasta}), " FASTA sequences; wrote $uncnt unplaced scaffolds to '$unord_path', total length $unlen.\n";
}

sub load_map_file {
    my ($map_path, $map, $rev_map, $comps) = @_;

    my ($fh, @parts, $line, $href, $scaf, $s);

    open($fh, "<" . $map_path) or die("Unable to open map file '$map_path' $!\n");

    my %dict;

    while ($line = <$fh>) {
        chomp $line;
        @parts = split(/\t/, $line);

	if (!defined($dict{$parts[0]})) {
	    $dict{$parts[0]} = {};
	}
	$dict{$parts[0]}->{$parts[3]}++;

	if (!defined($comps->{$parts[0]})) {
	    $comps->{$parts[0]} = {};
	}
	#
	# Record the component parts this scaffold is composed of.
	#
	$comps->{$parts[3]}->{$parts[7]}++;
    }

    print STDERR "Loaded mapping information for ", scalar(keys %{$comps}), " scaffolds.\n";

    foreach $scaf (keys %dict) {
	my @frags;

	next if ($scaf =~ /JHP/);

	# print STDERR $scaf, "\n";
	foreach $s (keys %{$dict{$scaf}}) {
	    if (defined($dict{$s})) {
		push(@frags, keys %{$dict{$s}});
	    } else {
		push(@frags, $s);
	    }
	}

	$rev_map->{$scaf} = [];
	foreach $s (@frags) {
	    push(@{$rev_map->{$scaf}}, $s);
	    $map->{$s} = $scaf;
	}
    }

    close($fh);
}

sub load_agp_file {
    my ($agp_path, $agp, $mapped, $rev_map) = @_;

    my ($fh, $out_fh, @parts, $line, $href, $unord_path);

    $unord_path = $out_path . "unplaced_scaffolds.agp";
    open($out_fh, ">" . $unord_path) or die("Unable to open unplaced scaffolds AGP file, '$!'\n");

    #
    # Load an AGP file defining gaps in the assembly.
    #  Format: object     object_beg object_end part_number component_type component_id component_beg component_end  orientation
    #  Gaps:   object     object_beg object_end part_number N              gap_length   gap_type      linkage        -empty-
    #          JH556661.1 1          70736 1 W AGAJ01000001.1 1        70736    +
    #          JH556661.1 70737      70782 2 N 46             fragment yes
    #
    open($fh, "<" . $agp_path) or die("Unable to open AGP file '$agp_path' $!\n");

    while ($line = <$fh>) {
        chomp $line;
        @parts = split(/\t/, $line);

	next if (substr($parts[0], 0, 1) eq "#");

	$href = {};
	$href->{'scaf'}  = $parts[0];
	$href->{'start'} = $parts[1];
	$href->{'end'}   = $parts[2];
	$href->{'part'}  = $parts[3];
	$href->{'type'}  = $parts[4];
	$href->{'len'}   = $href->{'type'} eq "N" ? $parts[5] : $parts[7] - $parts[6] + 1;

	push(@{$agp->{$href->{'scaf'}}}, $href);

	#
	# If this component was not incorporated into the map, immediately output it.
	#
	if (!defined($mapped->{$href->{'scaf'}}) && !defined($rev_map->{$href->{'scaf'}})) {
	    print $out_fh $line, "\n";
	}
    }

    print STDERR "Loaded component information for ", scalar(keys %{$agp}), " scaffolds.\n";

    close($fh);
    close($out_fh);
}

sub load_lgs {
    my ($in_path, $lgs, $mapped_scaffolds) = @_;

    my (@parts, @files, $file, $lg, $line, $href, $cnt, $lg_fh);

    build_file_list(\@files);

    foreach $file (@files) {
	#
	# Load a file of the markers for a particular linkage group, ordered by JoinMap
	#
	open($lg_fh, "<" . $in_path . "/" . $file) 
	    or die("Unable to open LG mapping file '$in_path/$file' $!\n");

	$cnt  = 0;
	($lg) = ($file =~ /lg(\d+)_scaffold\.tsv/);
	$lgs->{$lg} = [];

	while ($line = <$lg_fh>) {
	    chomp $line;
	    @parts = split(/\t/, $line);
	    $href = {};
	    $href->{'master'}   = $parts[0];
	    $href->{'scaffold'} = $parts[1];
	    $href->{'ori'}      = $parts[2] eq "+" ? forw : reve;
	    $href->{'start'}    = $parts[3];
	    $href->{'end'}      = $parts[4];
	    if (scalar(@parts) >= 6) {
		($href->{'old_scaf'}) = ($parts[5] =~ /\[(.+)\]/);
	    }

	    $mapped_scaffolds->{$href->{'scaffold'}}++;

	    push(@{$lgs->{$lg}}, $href);
	}

	print STDERR "Loaded ", scalar(@{$lgs->{$lg}}), " mapped scaffolds from '$file'.\n";
	close($lg_fh);
    }
}

sub build_file_list {
    my ($files) = @_;

    my (@ls, $line, $prefix, $wl_keys, $wl_fh);
    my %whitelist;

    @ls = `ls -1 $in_path/lg*_scaffold.tsv 2> /dev/null`;

    if (scalar(@ls) == 0) {
	print STDERR "Unable to locate any input files to process within '$in_path'\n";
	usage();
    }

    foreach $line (@ls) {
	chomp $line;

	($prefix) = ($line =~ /$in_path\/(lg\d+_scaffold\.tsv)/);

	push(@{$files}, $prefix) if (length($prefix) > 0);
    }
}

sub parse_command_line {
    while (@ARGV) {
	$_ = shift @ARGV;
	if    ($_ =~ /^-p$/) { $in_path    = shift @ARGV; }
	elsif ($_ =~ /^-m$/) { $map_path   = shift @ARGV; }
	elsif ($_ =~ /^-a$/) { $agp_path   = shift @ARGV; }
	elsif ($_ =~ /^-f$/) { $fasta_path = shift @ARGV; }
	elsif ($_ =~ /^-o$/) { $out_path   = shift @ARGV; }
	elsif ($_ =~ /^-d$/) { $debug++; }
	elsif ($_ =~ /^-h$/) { usage(); }
	else {
	    print STDERR "Unknown command line option: '$_'\n";
	    usage();
	}
    }

    if (length($agp_path) == 0) {
	print STDERR "You must specify the AGP file defining assembly gaps.\n";
	usage();
    }

    if (length($map_path) == 0) {
	print STDERR "You must specify the scaffold mapping file.\n";
	usage();
    }

    if (length($fasta_path) == 0) {
	print STDERR "You must specify the file containing unmapped FASTA sequences..\n";
	usage();
    }

    if (length($out_path) == 0 || !-e $out_path) {
	print STDERR "You must specify a directory to output scaffold assembly FASTAs.\n";
	usage();
    }
    $out_path .= "/" if (substr($out_path, -1) ne "/");
}

sub usage {
    print STDERR <<EOQ; 
gen_fasta.pl -f fasta -p path -m map -a agp -o path [-d] [-h]
    f: path to FASTA file containing unassembled scaffolds.
    p: path to directory containing LG map files.
    m: path to scaffold map file (output by order_contigs.pl).
    a: path to the AGP formatted file defining assembly gaps (modified by order_contigs.pl).
    o: directory to output assembled FASTA files.
    h: display this help message.
    d: turn on debug output.

EOQ

exit(0);
}
